// ================================================================
// $Id: basmath.h,v 1.10 2006/09/29 09:12:39 xkrivanj Exp $
//
// basmath.h
//     Basic mathematical definitions
//
// Class: CLimits
//
// REPLACEMENT_STRING
//
// Initial coding by Tomas Kopal, rewritten by Vlastimil Havran, 2000.

#ifndef __BASMATH_H__
#define __BASMATH_H__

#include "configh.h"

// Standard headers
#include <cmath>
#ifdef __UNIX__
//include <values.h>
#endif // __UNIX__

// forward declarations

//-----------------------------------------------------------------------------
// Miscellaneous helper macros
//-----------------------------------------------------------------------------

// The definition of some constants
#ifndef M_PI
#define M_PI        3.14159265358979323846
#endif
#ifndef M_PI_2
#define M_PI_2	    1.57079632679489661923
#endif
#ifndef M_PI_4
#define M_PI_4	    0.78539816339744830962
#endif
#ifndef M_1_PI
#define M_1_PI	    0.31830988618379067154
#endif
#ifndef M_2_PI
#define M_2_PI	    0.63661977236758134308
#endif
#ifndef M_LN2
#define M_LN2      0.693147180559945309417
#endif
#ifndef M_1_LN2
#define M_1_LN2    1.4426950408889634073599246810019
#endif
#ifndef M_SQRT2
#define M_SQRT2    1.4142135623730950488016887242097
#endif
#ifndef M_1_SQRT2
#define M_1_SQRT2  0.70710678118654752440084436210485
#endif

#ifndef MAXFLOAT
#define MAXFLOAT    ((float)3.40282347e+38)
#endif

// -------------------------------------------------------------------
/// Limits.
/**
  This class encapsulates all the concessions to floating-point
      error made by ray tracers.
  */
class CLimits
{
public:
  /// The number used to reject too-close intersections.
  static float Threshold;

  /// The number used to reject too-close intersections for planar objects.
  static float ThresholdForPlanars;

  /// This is a "small" number.  
  /** 
    Less than this number is assumed to be 0, when such things matter.
    The default value is 0.1.
  */
  static float Small;

  /// This is an impractically "large" number.
  /**
    Used for intersection parameters out to infinity (e.g. the span resulting from
    an FindAllIntersections operation on a plane).
    The default value is 100000.
  */
  static float Infinity;

  /// A parameter to set reasonable overlap value on the boundary of triangles etc. 
  /** 
      It should be set similar to the Small constant above. It corresponds to 
      the physical size of the scene and specifies maximum overlap from the 
      exact geometry to avoid holes in the polygonal scenes.
  */
  static float BoundaryEpsilon;
};

// -------------------------------------------------------------------
// Support functions
// -------------------------------------------------------------------

// -------------------------------------------------------------------
// Basic Math Functions

/// Return factorial of 'n' as a double precision floating point number.
double factorial(int n);

/// Returns square of the variable
template<class T>
T
sqr(T a)
{
  return a * a;
}

inline float
sqr(float a)
{
  return a * a;
}

/// Returns signum of the variable
inline int
signum(float a, float epsilon = CLimits::Small) 
{
  if (a > epsilon)
    return 1; 
  else
    if (a < -epsilon)
      return -1; 
  return 0;
}

// N-th power of base in integer
inline int
intintpow(int base, int exponent) {
  int res = 1;
  for(int i=0;i<exponent;i++)
    res *= base;
  return res;
}

// --------------------------------------------------------------------------
//  pow2ge
// --------------------------------------------------------------------------
/// Find a power of 2 greater than or equal to a given number.
inline int pow2ge(int a)
{
  int r = 1;
  while (r < a)
    r <<=1;
  return r;
}

// --------------------------------------------------------------------------
//  pow4ge
// --------------------------------------------------------------------------
/// Find a power of 4 greater than or equal to a given number.
inline int pow4ge(int a)
{
  int r = 1;
  while (r < a)
    r <<=2;
  return r;
}

/// Returns absolute value of a given type for floating point values
template<class T>
T
Abs(const T &x)
{
  return (x < 0) ? -x : x; 
}

/// Swaps two values.
template<class T>
inline void 
Swap(T &x, T &y)
{
  T temp = x;
  x = y;
  y = temp;
}

/// Returns nonzero if min <= candidate <= max.
template<class T>
inline int 
InRange(T min, T max, T candidate)
{
  return (candidate >= min) && (candidate <= max);
}

/// Returns maximum of two values.
template<class T>
inline const T&
Max(const T &x, const T &y)
{
  return (x < y) ? y : x;
}

/// Returns minimum of two values
template<class T>
inline const T&
Min(const T &x, const T &y)
{
  return (x > y) ? y : x;
}

/// Returns if two values are equal or not (by epsilon).
inline int
EpsilonEqual(const float &x, const float &y)
{
  return Abs(x-y) < CLimits::Small;
}

/// Returns if a value is zero (+/- epsilon)
inline int
EpsilonEqual(const float &x)
{
  return Abs(x) < CLimits::Small;
}

/// Returns if two values are equal or not by a given epsilon.
inline int
EpsilonEqual(const float &x, const float &y, const float &epsilon)
{
  return Abs(x-y) < epsilon;
}

#if (defined(__linux__) && defined(__i386__)) || defined(WIN32)

#define FAST_INT 1
#endif
#define _doublemagicroundeps (.5-1.4e-11)
//almost .5f = .5f - 1e^(number of exp bit)
inline int Round2Int(double val)
{
#ifdef FAST_INT
#define _doublemagic double (6755399441055744.0)
  //2^52 * 1.5,  uses limited precision to floor
  val = val + _doublemagic;
  return ((long*)&val)[0];
#else
  return int (val+_doublemagicroundeps);
#endif
}

inline int Float2Int(double val)
{
#ifdef FAST_INT
  return (val<0) ?  Round2Int(val+_doublemagicroundeps) :
    Round2Int(val-_doublemagicroundeps);
#else
  return (int)val;
#endif
}

inline int Floor2Int(double val)
{
#ifdef FAST_INT
  return Round2Int(val - _doublemagicroundeps);
#else
  return (int)floorf(val);
#endif
}

inline int Ceil2Int(double val)
{
#ifdef FAST_INT
  return Round2Int(val + _doublemagicroundeps);
#else
  return (int)ceilf(val);
#endif
}

/// Find M and N such that M*N \approx numRays and N/M \approx NoverM
inline void
DivideRect(const int numRays, const float NoverM, int &M, int &N)
{
  if (numRays==1) {
    M = N  = 1;
  }
  else {
    M = (int)(sqrt((float)numRays/NoverM));
     if (M < 1) M = 1;
     N = (int)((float)numRays/(float)M);
  }
}

// <JK>  There's not log2() in MS Visual C++ .NET 2003
/// Return binary logarithm of x, i.e. the logarithm base is 2.
#if (defined(_MSC_VER) || defined(__INTEL_COMPILER))
template<class T>
T log2(const T x)
{
  return log(x)*M_1_LN2;
}
#endif

inline int round(float co) 
  //! function rounds float number to integer
{
   if( fabs( co - (int)(co) ) <  0.5 )
	 return((int)floor(co) );
   if( fabs( co - (int)(co) ) >= 0.5 )
     return((int)ceil(co) );
   return(0);
}// round ---------------------------------------------------------

#endif // __BASMATH_H__

