// ===================================================================
// $Id: $
//
// lightenv.cpp
//
//     The declaration of the main routines for OpenGL viewer application
//
// Initial coding by Jiri Filip and Vlastimil Havran, 2008.
// Rewritten by Vlastimil Havran, 2013 using OpenGL shaders

//#include <GL/glew.h>
#include "pgr.h"
#include <GL/freeglut.h>
//#include <GL/glut.h>
#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <vector>
#include <cassert>
#include <cstring>
#include <sys/stat.h>

#include "configg.h"
#include "bconfig.h"
#include "basstr.h"
#include "brdfmodel.h"
#include "colorfs.h"
#include "AuxRND.h"
#include "AuxFuncs.h"
#include "TAlloc.h"
#include "readpng.h"
#include "OBJloader.h"
#include "ibtfbase.h"
#include "hdrloader.h"
#include "lightenv.h"
#include "pgr.h"
#include "timer.h"
#include "tgaimg.h"
#include "vrmlsup.h"
#include "render.h"
#include "TexturedMesh.h"
#include "CleanMesh.h"
#include "glenvmap.h"

//--------------------------------
// The use of SBRDF library, all BRDF models are included via this header
#include "sbrdflib.h"

inline float
sqr(float h) { return h*h; }

//inline double sqr(double h) { return h*h; }

//== list of lights and colours of environment map =========================

// 0 .. rendering image
// 1 .. computing albedo, variance, etc for BRDF model
// 2 .. computing mean value, variance, etc for EM (full sphere)
// 3 .. computing mean value, variance, etc for EM (north hemisphere only)
// 4 .. compare the two images
int whatToDo = 0;

// the default number of samples to be used by the program when started,
// it holds defaultCntSamples <= maxCntSamplesPerPixel
int defaultCntSamplesPerPixel = 200; // 100

// the number of samples per pixel use to render the image for a current pixels
// it must hold that cntSamplesPerPixel <= maxCntSamplesPerPixel
int cntSamplesPerPixel = defaultCntSamplesPerPixel;

// the number of samples for computing reference solution in mode 1000
int defaultCntSamplesPerPixelREF = 200000;

// the alpha blending coefficient for MIS (=multiple importance sampling)
// Usage: alpha=0.0 ... use only BRDF, alpha = 1.0 - use only EM sampling
float alphaMIS = 0.5;

// if the reference image is read and available
bool readReferenceImageFlag = false;
bool referenceImageAvailable = false;
// It is meant atlas texture, not the output file
char referenceImageFilename[2000]={"defaultrefimageatlas.hdr"};
char diffImageFilename[2000] = {"diffimage.hdr"};
char diffvisImageFilename[2000] = {"diffimagevis.hdr"};
CHDRImage *referenceImage = 0;
float upperBoundVisualization = 1.0f; // which value should be taken as maximum for visualization for RMSE

// Standard, process keys, this can be switched off for offline rendering
int disableInteractivity = 0;

// Normalizing albedo of BRDF
static float normalizeTotalAlbedoCoeff = 1.0f;
bool saveAlsoNormalizedAlbedo = false;
// Saving also gray image on the object
bool saveAlsoGrayToneImage = false;
// If to record the number to a filename
bool recordNumberFlag = false;

// these two below are internal variables
bool suffixAlbedoNormalized = false; // keep it false
bool suffixGrayTone = false; // keep it false
int currTexID = 0; // current ID for texturing

// rendering algorithm given currently
int modeCompute=0;
// When crrModeComputeIndex is set to -1, it uses a single method set above
// in 'modeCompute' variable
int currModeComputeIndex = 0;
bool printEstimateFlag = true;

// to save the values from the image
int visModeVariable = 0;
// use 0 .. ratio (evalBRDFcntST/evalBRDFcntST+evalEMcntST),
// use 1 .. value 'evalST' clipped to values minValueVIS and maxValueVIS - for debugging of whatever
int visualizationMode = 0;

// if to make some visualization how many samples are taken from EM and
// how many from BRDF
int pseudoColorTechnique=6; // setting it to 0 switches off the visualization - it does save luminance only
// the prefix to save output files
char FILEPREFIX[100]={"data/\0"};

// the prefix might be given by computation
char FILEPREFIXVAR[100]={""};

// This applies to computeMode6 and computeMode 7
// either 1 .. compute minimum, 2 .. 50/50 percent, 3 .. take a single sampling 
int modeNewSamples = 1;

// rotateEM after loading by X degrees left or right, before it is processed just after loading
float rotateEM = 0.f;

#define M_PI 3.14159265358979323846
//#define FLOATMAX ((float)(1e30))

CEnvironmentMap *debugImage=0;

int drawNeedles = 0;  // toggle additional normals drawing

// when to finish rendering of images
const int cntFinishRendering = 2;
bool saveAndExit = false;
int actualOBJ = 0;          // specifies actual 3D OBJ model
int actualEnvMap = 2;       // specifies the environment map to be used for rendering
// This is hack only!!!!!!!!!!!!
int envMapToReadHACK = 2; // as standard stpeters
int NoOfEnvMapsToReadHACK = 3;
// End of hack

bool setConstEMflag = false; // to set it constant after reading from file, if this is actualEnvMap index
bool setConstNorthHemEMflag = false; // to set it constant after reading from file the north hemisphere
bool semConstSouthHemEMflag = false; // to set it constant after reading from file the north hemisphere
bool setCosineNorthHemEMflag = false; // set north hemisphere of EM map to cosine(theta) with respect to the northpole

// Switch if to randomized precomputed lights, so avoid correlation from EM sampling if set true
bool randomizePrecomputedLightsFlag = true;

int grabbingSequence = 0;      // grabs screen for every loop run
int rotateEnvironment = 0;     //  - rotate environment illumination map
int rotateObject = 0;          //  - rotate sphere or OBJ model for each cycle   

int renderAxis = 0;            // toggle rendering Axes
int renderWire = 0;            // toggle wireframe rendering only
int renderLights = 0;          // render visualizaton of lights ON/OFF
int renderEnvMap = 1;          // render environment map

int screenShotNumber = -1;      // initial/actual screenshot number
int NoOfScreenShots = -1;      // maximal screenshot number to be stored before exiting

int indexRenderedImage = 0;    // no. of rendered images
const bool grabHDRflag = true; // true ... saving in HDR image (in EXR format), false ... in LDR image

bool savedSingleLightPositions = false;
bool allowEnvManipulationFlag = true;

int maxBTFmaterials = 0;
TAngles *TA = 0;           // object for direction handling and BTF interpolation

char modestring[30];  // string for light mode indication
//== visualisation/GUI ==========================================================

const float perspectiveAngle=64.f;           // definition of viewing verspective angle
int oldwinx, oldwiny;                        // variables determining mouse movement
int mb[3] = {0, 0, 0};                       // mouse buttons: left, middle, right

// initial colour of 2D interface, if no environment map used
float backgroundColor[4] = {1.0, 1.0, 1.0, 1.0};

// initial camera position is defined here
const float viewerDistanceInit = 15;
const float viewerAngleInit = 0.4; // radians
const float cameraat[3] = {0, 0, 0}; // looking at point
float viewerDistance = 0;
float viewerAngle = 0;
float camerapos[3] = {0, 0, 0};
float cameraup[3] = {0, 0, 0}; // up vector

// Point light position for a single light mode
const float lposInit[4] = {500.f,200.f,500.f, 1.f};   // MOVIE initial light position - from left
float lpos[4], lposBackup[4];

//== string for version and extensions printing ============================
char *auxString = 0;

const int NoOfEnvMaps = 3; // the number of environment maps to process
char envMapsFilename[NoOfEnvMaps][1000]; // the files for environment maps
int *envImageRows = 0; // the real size of the environment map image in file
int *envImageCols = 0;
int *envRows = 0; // the texture size of the environment map image as used by Opengl
int *envCols = 0;
// Environment maps in HDR
CEnvironmentMap *environMap= 0;

// initial setting of shift of environment maps in the video and for rendering
float envPhiShift = 0; // current shift of phi of the environment map

const int phiStepsPerCircle = 72; // how much to rotate lights
const float environMapRadius = 1000.f; // the geometry is far away
const float lightsRadius = environMapRadius * 0.9f; // this is for visualization of directional lights

// BRDF model used in the rendering context
BRDF *brdf = 0;

int windowID;
const int widthBargraph = 30;
const int borderBargraph = 5;

// Forward declaration
void Set3DObjectsNames(bool allocateFlag);

float
lightsPhiShift() {
  return envPhiShift;
}

void printUsage();

// The render mode 0..BTF, 1..Phong/Phongvar, 2..BRDFdataFromFile, 3...MatusikBRDFdata
int modeBRDF = -1;

// Only for debugging purposes
int debugPosX, debugPosY;

// For debugging only
int indexImageBTFatlas = -1;
bool saveImageBTFatlas = true;

//== variables for BTF rendering ===========================================
//initial BTF rendering size
//int Npow = 8; // 2^8 = 256  ...  2^11 = 2048
int Npow = 9; // 512x512
//the size of texture used for rendering
int BTFrows=0;
int BTFcols=0;

int tileSize = 32;
float objPhiShift = 0.f;
int ncolour = 2;
// for the object and other data
static GLuint BTFidTexture[5];
unsigned char *BTFimage=0;
float *BTFimageF=0; // rendered image as texture on the object = BTFimageB[0]
// 0 ... rendered image = BTFimageF,
// 1 ... visualization image,
// 2 ... gray
// 3 ... graynormalizedA
// 4 ... normalizedA
float *BTFimageB[5]; 

int printCount = 1;

//== OBJ models rendering ===========================================
const int NoOfObjects = 1; //
const int verticesPerObject = 150000; // maximum number of vertices per geometric object
TObjectData *DATA[NoOfObjects];// class for OBJ 3D object files loading
char *OBJpaths[NoOfObjects];
char *OBJnames[NoOfObjects];
float OBJscales[NoOfObjects];
float initOBJrot[NoOfObjects][3];
float initOBJtrans[NoOfObjects][3]; 
float OBJrot[NoOfObjects][3]; 
float OBJtrans[NoOfObjects][3];
// The data read from file
float ***V[NoOfObjects],***Vbackup[NoOfObjects];
float ***T[NoOfObjects],***Tbackup[NoOfObjects];
float ***B[NoOfObjects],***Bbackup[NoOfObjects];
float ***N[NoOfObjects],***Nbackup[NoOfObjects];
float ***UV[NoOfObjects],***UVbackup[NoOfObjects];

// the global timer
CTimer timer;

char databaseName[100];

// smooth texturing + simple BRDF model over it is possible as well
int textureSize = 2048;
float ***textureImage;

// for comparison image writing - side by side original and model images
int compareScreenShots = 0;

// ---------------------------------------------------------------------------------
// Particular mesh to be rendered
CTexturedMesh mesh[NoOfObjects];
// Particular mesh for environment map
CEnvironmentMapGL meshenvmap[NoOfEnvMaps];
bool useFrameBuffer = false;
// ----------------------------------------------------
// Output framebuffer
int win_fbo_width=800, win_fbo_height=600;    // Frame buffer object
GLuint fbo=-1; // frame buffer object
GLuint fbodepth=-1; // Depth buffer
GLuint fbotexture=-1; // color buffer
glm::mat4 viewMatrix = glm::mat4(1.0f);
glm::mat4 projectionMatrix = glm::mat4(1.0f);

void
initProjectionMatrix(int width, int height)
{
  // glm::mat4 perspective(float fovy, float aspect, float zNear,float zFar);
  projectionMatrix = glm::perspective(perspectiveAngle, (float)width/(float)height,
				      0.1f, environMapRadius * 10.f);
  return;
}

void
initViewMatrices(int width, int height)
{
  // camera is rotated in YZ plane only
  camerapos[0] = viewerDistance * cos(viewerAngle);
  camerapos[1] = 0;
  camerapos[2] = viewerDistance * sin(viewerAngle);
  cameraup[0] = sin(viewerAngle);
  cameraup[1] = 0;
  cameraup[2] = cos(viewerAngle);
  cout << "Camerapos " << camerapos[0] << " " << camerapos[1] << " " << camerapos[2] << endl;
  cout << "Cameraup " << cameraup[0] << " " << cameraup[1] << " " << cameraup[2] << endl;
 
  // viewing matrix
  // glm::lookAt(camera position, camera target, up); 
  viewMatrix = glm::lookAt( glm::vec3(camerapos[0], camerapos[1], camerapos[2]),
			    glm::vec3(cameraat[0], cameraat[1], cameraat[2]),
			    glm::vec3(cameraup[0], cameraup[1], cameraup[2]) );
  initProjectionMatrix(width, height);
  return;
}

void
createCameraposFromObjectPosition()
{
  initViewMatrices(win_fbo_width, win_fbo_height);  
  if (TA)
    TA->setViewPos(camerapos);
}

//#####################################################################################
//### FUNCTIONS #######################################################################
//#####################################################################################

///////////////////////////////////////////////////////////////////////////////
// check FBO completeness
///////////////////////////////////////////////////////////////////////////////
bool checkFramebufferStatus()
{
    // check FBO status
    GLenum status = glCheckFramebufferStatus(GL_FRAMEBUFFER);
    switch(status)
    {
    case GL_FRAMEBUFFER_COMPLETE:
        std::cout << "Framebuffer complete." << std::endl;
        return true;

    case GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT:
        std::cout << "[ERROR] Framebuffer incomplete: Attachment is NOT complete." << std::endl;
        return false;

    case GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT:
        std::cout << "[ERROR] Framebuffer incomplete: No image is attached to FBO." << std::endl;
        return false;
	/*
    case GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS:
        std::cout << "[ERROR] Framebuffer incomplete: Attached images have different dimensions." << std::endl;
        return false;

    case GL_FRAMEBUFFER_INCOMPLETE_FORMATS:
        std::cout << "[ERROR] Framebuffer incomplete: Color attached images have different internal formats."
	          << std::endl;
        return false;
	*/
    case GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER:
        std::cout << "[ERROR] Framebuffer incomplete: Draw buffer." << std::endl;
        return false;

    case GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER:
        std::cout << "[ERROR] Framebuffer incomplete: Read buffer." << std::endl;
        return false;

    case GL_FRAMEBUFFER_UNSUPPORTED:
        std::cout << "[ERROR] Framebuffer incomplete: Unsupported by FBO implementation." << std::endl;
        return false;

    default:
        std::cout << "[ERROR] Framebuffer incomplete: Unknown error." << std::endl;
        return false;
    }
}

void
deallocations()
{
  glDeleteFramebuffers(1, &fbo);
  glDeleteRenderbuffers(1, &fbodepth);
  glDeleteTextures(1, &fbotexture);
  
  if (auxString)
    freecmem1(auxString, 0, 1000);

  delete TA;
  TA = 0;
  for (int i = 0; i < 5; i++) {
    freemem1(BTFimageB[i], 0, 2048*2048*4);
  }
  if(envImageRows)
    freeimem1(envImageRows, 0, NoOfEnvMaps);
  if(envImageCols)
    freeimem1(envImageCols, 0, NoOfEnvMaps);
  if(envRows)
    freeimem1(envRows, 0, NoOfEnvMaps);
  if(envCols)
    freeimem1(envCols, 0, NoOfEnvMaps);
  //  if(envCorrCoeff)
  //    freemem2(envCorrCoeff, 0, NoOfEnvMaps, 0, NoOfObjects);
  //  if(envCorrMat)
  //    freemem1(envCorrMat, 0, maxBTFmaterials-1);  
  
  for(int i=0;i<NoOfObjects;i++) {
    delete DATA[i];
    if(OBJnames[i])
      freecmem1(OBJnames[i], 0,1000);
    if(OBJpaths[i])
      freecmem1(OBJpaths[i], 0,1000);
    if(V[i])
      freemem3(V[i], 1,verticesPerObject, 0,2, 0,2); 
    if(Vbackup[i])
      freemem3(Vbackup[i], 1,verticesPerObject, 0,2, 0,2); 
    if(T[i])
      freemem3(T[i], 1,verticesPerObject, 0,2, 0,2); 
    if(Tbackup[i])
      freemem3(Tbackup[i], 1,verticesPerObject, 0,2, 0,2); 
    if(N[i])
      freemem3(N[i], 1,verticesPerObject, 0,2, 0,2); 
    if(Nbackup[i])
      freemem3(Nbackup[i], 1,verticesPerObject, 0,2, 0,2); 
    if(B[i])
      freemem3(B[i], 1,verticesPerObject, 0,2, 0,2); 
    if(Bbackup[i])
      freemem3(Bbackup[i], 1,verticesPerObject, 0,2, 0,2); 
    if(UV[i])
      freemem3(UV[i], 1,verticesPerObject, 0,2, 0,1); 
    if(UVbackup[i])
      freemem3(UVbackup[i], 1,verticesPerObject, 0,2, 0,1); 
  } // for
  
}//--- deallocations ------------------------------------------------------------


void 
setTexParams()
{
  // setting texturing params. The same for all texturesl 
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
  if(1) {
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  }
  else {
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
  }
}//-- setTexParams -------------------------------------------------------------

void 
changeSize(int w, int h) 
{
  // Prevent a divide by zero, when window is too short (you cant make a window of zero width).
  win_fbo_width = w;
  win_fbo_height = h;
  // Set the viewport to be the entire window
  glViewport(0, 0, w, h);
  initProjectionMatrix(w, h);

  return;
}//--- changeSize -----------------------------------------------------------

void 
rotateEnvironmentLights(float angle)
{
  // ... rotation of whole environment lights by angle "lightsPhiShift"
  for(int em=0; em<NoOfEnvMaps ; em++) {
    environMap[em].rotateEnvironmentLights(angle);
  }
  return;
}//--- rotateEnvironmentLights ---------------------------------------------

void 
transform3DPoint(const glm::mat4 &m, float *v)
{
  glm::vec4 ww(v[0], v[1], v[2], 1.0f);

  glm::vec4 res = m * ww;
  v[0] = res[0] / res[3];
  v[1] = res[1] / res[3];
  v[2] = res[2] / res[3];
  
  return;
}//-- transform3DPoint -------------------------------------------------------

void
// Make transformation of the object accordingly, use backup data
transformOBJmodelGeometry(int actualObj, const glm::mat4 &modelMatrix)
{
  for(int f=1;f<=DATA[actualObj]->noOfFaces();f++) { // for faces
    for(int k=0;k<3;k++) { // for vertices
      // Model rotation
      // make three points before rotation over the initial data
      for(int j=0;j<3;j++) { // for xyz
	V[actualOBJ][f][k][j] = Vbackup[actualOBJ][f][k][j];
	N[actualOBJ][f][k][j] = Vbackup[actualOBJ][f][k][j] + Nbackup[actualOBJ][f][k][j];
	T[actualOBJ][f][k][j] = Vbackup[actualOBJ][f][k][j] + Tbackup[actualOBJ][f][k][j];
      }
      transform3DPoint(modelMatrix, V[actualOBJ][f][k]);
      transform3DPoint(modelMatrix, N[actualOBJ][f][k]);
      transform3DPoint(modelMatrix, T[actualOBJ][f][k]);
      for(int j=0;j<3;j++) { // for xyz
	N[actualOBJ][f][k][j] -= V[actualOBJ][f][k][j];
	T[actualOBJ][f][k][j] -= V[actualOBJ][f][k][j];
      }
      vec_norm(N[actualOBJ][f][k]);
      vec_norm(T[actualOBJ][f][k]);
      vec_crossprod(N[actualOBJ][f][k], T[actualOBJ][f][k], B[actualOBJ][f][k]);
#if 0
      cout << "V " << k << " =  "; 
      for(int j=0;j<3;j++) { // for xyz
	cout << V[actualOBJ][f][k][j] << " ";
      }
      cout << endl;
      cout << "N = ";
      for(int j=0;j<3;j++) { // for xyz
	cout << N[actualOBJ][f][k][j] << " ";
      }
      cout << endl;
#endif
    } // for k
  } // for f
  
  return;
}//--- transformOBJModelGeometry -----------------------------------------------

void
makeFilename(bool atlas, bool useOrigData, bool visflag,
	     int screenShotNumber, int pass, char *filename)
{
  strcpy(auxString, FILEPREFIX);
  
  char aux[1000];
  if (strlen(FILEPREFIXVAR) > 0) {
    // we start with some precomputed prefix
    strcat(auxString, FILEPREFIXVAR);
  }  
  if (useOrigData)
    sprintf(aux,"%s_%s_%04do",
	    OBJnames[actualOBJ], databaseName, (int)pow(2, Npow));
  else {
    if (compareScreenShots==1)
      sprintf(aux,"compare_%s_%s_%04dm",
	      OBJnames[actualOBJ], databaseName, (int)pow(2, Npow));
    else
      sprintf(aux,"%s_%s_%04dm", OBJnames[actualOBJ],
	      databaseName, (int)pow(2, Npow));
    strcat(auxString, aux);
  }
  // ---------------------------------------------
  if (IsEnvMapLighting()) {
    sprintf(aux,"_%s", environMap[actualEnvMap].getName().c_str());
    strcat(auxString,aux);
  }
  if (atlas) {
    strcpy(aux, "_atlas");
    strcat(auxString, aux);
  }
  if (visflag) {
    sprintf(aux, "_vis%d", pseudoColorTechnique);
    strcat(auxString, aux);
  }
  if (visModeVariable) {
    sprintf(aux, "_V%02d", visModeVariable);
    strcat(auxString, aux);
  }
  if (pass > -1) {
    sprintf(aux, "_p%1d", pass);    
    strcat(auxString, aux);
  }  
  sprintf(aux, "_M%03d", modeCompute);
  strcat(auxString,aux);
  if ((grabbingSequence)&&(recordNumberFlag)) {
    sprintf(aux, "_%04d", screenShotNumber);
    strcat(auxString,aux);
  }
  
  // copy result and exit
  strcpy(filename, auxString);
}

void
CopyImageToHDRClass(CHDRImage *img, float *imgBTF, bool visflag = false)
{
  assert(img);
	
  if (visflag)
    img->allocate(BTFcols+widthBargraph+borderBargraph, BTFrows, true);
  else
    img->allocate(BTFcols, BTFrows, true);
  for (int y = 0; y < BTFrows; y++) {
    for (int x = 0; x < BTFcols; x++) {
      int i = (BTFrows*y+x)*4;
      img->setRGB_XY(x, y, imgBTF[i], imgBTF[i+1], imgBTF[i+2]);
    } // for x
    if (visflag) {
      // copy the bargraph values to the right
      float RGB[3];
      float value = (float)(BTFrows-y)/BTFrows; // get the color
      GetPseudoColor(pseudoColorTechnique, value, 0.0, 1.0, RGB);
      for (int x = BTFcols; x < BTFcols+borderBargraph; x++) {
	img->setRGB_XY(x, y, 0, 0, 0); // make it black
      }
      for (int x = BTFcols+borderBargraph; x < BTFcols + borderBargraph + widthBargraph; x++) {
	img->setRGB_XY(x, y, RGB[0], RGB[1], RGB[2]);	
      }
    }
  } // for y
  return;
}

int 
printAtlas(float *imgBTF, int indexImage, bool visflag, int pass)
{
  if (indexImage < 0)
    return 0;
  
  // saving rendered BTF map to HDR
  char fileName[1000];

  makeFilename(true, false, visflag, indexImage, pass, fileName);
  // save as HDR file
#ifdef USE_HDR  
  strcat(fileName,".hdr");
#else
  // EXR is standard here
  strcat(fileName, FILEEXT);  
#endif
  
  CHDRImage *img = new CHDRImage();
  CopyImageToHDRClass(img, imgBTF, visflag);
    
  img->saveHDR(fileName);
  delete img;
  printf("BTF atlas saved to the file %s\n", fileName);  
  return 0;
}//--- printAtlas --------------------------------------------------------

void
UpdateFPS(float timeFrameUser, float timeFrameReal)
{
  char ch;
  char ch2;
  static int printCount = 0;
  switch (printCount) {
    case 0:
      ch = '/'; ch2 = ' '; printCount++; break;
    case 1:
      ch = '-'; ch2 = '-'; printCount++; break;
    case 2:
      ch = '\\'; ch2 = ' '; printCount++; break;
    case 3:
      ch = '|'; ch2 = ' '; printCount=0; break;
  }  
  // print the time
  char *buf = new char [2000];
  assert(buf);
  string strmessage;
  brdf->getString(strmessage);
  sprintf(buf, "EnvRenderer MC=%d %c%c  WxH=%dx%d, %3.2f/%3.2f FPS obj=%s %s",
	  modeCompute,
	  ch, ch2, win_fbo_width, win_fbo_height,
	  (1.0f / timeFrameUser), (1.0f/ timeFrameReal),
	  OBJpaths[actualOBJ], strmessage.c_str());
  glutSetWindowTitle(buf);
  glFlush();
  glutPostRedisplay();  
  delete [] buf;
}//--- UpdateFPS ------------------------------------------------------------

void
GetTonemappingMult(float &normF)
{
  // This provides some normalization coefficient for visualization
  // during rendering that needs to get the value to the lightness
  // suitable for showing the image on the screen

  // physically based correct for any material
  if (IsEnvMapLighting()) {
    normF = 5.f; // all rendering modes with environment map
  }
  else
    normF = 5.f; // single point light

  //  cout << "normF = " << normF << endl;
}//--- GetTonemappingF -------------------------------------------------------------

void
DrawNeedles(int drawNeedles, int imin, int imax)
{
  if (drawNeedles == 0)
    return;
  int cnt = imax - imin;
  float scale = 0.03f;
  float *vert = new float[3*2*cnt];
  CCleanMesh meshneedles;
  for (int i2 = imin; i2 < imax; i2++) {
    const int i = i2-imin;
    for (int j = 0; j < 3 ; j++) { // three vertices per triangle
      float scale = 10.0f;
      // The first is always the vertex given by triangle
      vert[i*3+0] = V[actualOBJ][i2][j][0];
      vert[i*3+1] = V[actualOBJ][i2][j][1];
      vert[i*3+2] = V[actualOBJ][i2][j][2];
      if (drawNeedles==1) { // normals
	// second vertex
	vert[i*3+3] = V[actualOBJ][i2][j][0] + N[actualOBJ][i2][j][0] * scale;
	vert[i*3+4] = V[actualOBJ][i2][j][1] + N[actualOBJ][i2][j][1] * scale;
	vert[i*3+5] = V[actualOBJ][i2][j][2] + N[actualOBJ][i2][j][2] * scale;
      }
      if(drawNeedles==2) { // tangents
	// second vertex
	vert[i*3+3] = V[actualOBJ][i2][j][0] + T[actualOBJ][i2][j][0] * scale;
	vert[i*3+4] = V[actualOBJ][i2][j][1] + T[actualOBJ][i2][j][1] * scale;
	vert[i*3+5] = V[actualOBJ][i2][j][2] + T[actualOBJ][i2][j][2] * scale;
      }
      if(drawNeedles==3) { // bitangents
	vert[i*3+3] = V[actualOBJ][i2][j][0] + B[actualOBJ][i2][j][0] * scale;
	vert[i*3+4] = V[actualOBJ][i2][j][1] + B[actualOBJ][i2][j][1] * scale;
	vert[i*3+5] = V[actualOBJ][i2][j][2] + B[actualOBJ][i2][j][2] * scale;
      }
      if(drawNeedles==4) { // direction to the single light
	float light_pos[3], illu_vis[3];
	TA->getIlluPos(light_pos);
	vec_diff(light_pos, V[actualOBJ][i2][j], illu_vis);
	vert[i*3+3] = V[actualOBJ][i2][j][0] + illu_vis[0] * scale;
	vert[i*3+4] = V[actualOBJ][i2][j][1] + illu_vis[1] * scale;
	vert[i*3+5] = V[actualOBJ][i2][j][2] + illu_vis[2] * scale;
      }
      if(drawNeedles==5) { // direction to the viewer
	float cam_pos[3], illu_vis[3];
	TA->getViewPos(cam_pos);
	vec_diff(cam_pos, V[actualOBJ][i2][j], illu_vis);
	vert[i*3+3] = V[actualOBJ][i2][j][0] + illu_vis[0] * scale;
	vert[i*3+4] = V[actualOBJ][i2][j][1] + illu_vis[1] * scale;
	vert[i*3+5] = V[actualOBJ][i2][j][2] + illu_vis[2] * scale;	
      }
    } // for j
  } // for i
  // Move the model somewhere by model matrix
  glm::mat4 envmapMatrixEM = glm::mat4(1.0f);
  glm::mat4 PVMmatrix = projectionMatrix * viewMatrix * envmapMatrixEM;
  
  // Draw the lines
  glLineWidth(2.0f);
  meshneedles.draw(PVMmatrix);
    
  return;
}//--- DrawNeedles -----------------------------------------------------------------

// Here we prepare the data to the texture
void 
renderOBJECT(const glm::mat4 &modelMatrix, TObjectData *OBJ, int indexRenderedImage)
{
#if 0
  cout << "modelMatrix " << modelMatrix[0][0] << " " << modelMatrix[0][1] << " " << modelMatrix[0][2]
       << " " << modelMatrix[0][3] << "\n";
  cout << "            " << modelMatrix[1][0] << " " << modelMatrix[1][1] << " " << modelMatrix[1][2]
       << " " << modelMatrix[1][3] << "\n";
  cout << "            " << modelMatrix[2][0] << " " << modelMatrix[2][1] << " " << modelMatrix[2][2]
       << " " << modelMatrix[2][3] << "\n";
  cout << "            " << modelMatrix[3][0] << " " << modelMatrix[3][1] << " " << modelMatrix[3][2]
       << " " << modelMatrix[3][3] << "\n";
#endif
  long double sumLum = 0;
  int cntPixelsRendered = 0;
  float valCheck = -100;
  // If to compute the estimate for rendering time
  static int cntEstimatedPixels = 2952;// this should be put here by hand from the console
  // seconds - do the evaluation after 10 seconds of rendering
  int intervalToPrintEstimate = 10;
  int firstEstimateOfTime = 1;

  static int cntFrame = 0;
  static int modeComputeSave = 0;
  if (cntFrame == 0) {
    // we store the method to be use
    modeComputeSave = modeCompute;
    modeCompute = 0; // point light source to be computed fast
  }
  int localModeCompute = modeCompute;

  // Make transformation of the object accordingly to OpenGL visualization, use backup data
  transformOBJmodelGeometry(actualOBJ, modelMatrix);  
  
  if (maxCntSamplesPerPixel < cntSamplesPerPixel) {
    cout << "ERROR: maxCntSamplesPerPixel < cntSamplesPerPixel cannot happen\n" << endl;
    abort();
  }
  if (modeCompute > 0) {
    cout << "================================================================"
	 << endl << flush;
    cout << "RENDERING in MC: " << modeCompute << " cntSamplesPerPixel = "
	 << cntSamplesPerPixel << endl;
    cout << flush;
  }
  timer.Restart();
#define MARK 250.f
  // Here we zero the image
  //cout << "Zeroing BTFimageFvis" << endl;
  for(int i=0;i<BTFcols*BTFrows*4;i+=4)
  {
    BTFimageF[i] = 0.f; BTFimageB[1][i] = 0.f;
    BTFimageF[i+1] = 0.f; BTFimageB[1][i+1] = 0.f;
    // for some reason it has to be here VH 1/September/2013
    BTFimageF[i+2] = MARK; BTFimageB[1][i+2] = 0.f;
    BTFimageF[i+3] = 0.f;  BTFimageB[1][i+3] = 0.f;
  }
  valCheck = BTFimageF[0+2]; 
  float weight0, weight1, weight2;
  float den, pom1[3],pom2[3],pom3[3];
  // interpolated vertices, tangents, normals and binormals
  float Vint[3], Tint[3], Nint[3], Bint[3]; 
  int UVr[3][2]; // rasterized UV coordinates
  float uv[2];   // UV coordinate to be interpolated
  int minX,maxX,minY,maxY;

  // reset counters for MIS evaluation in smarter techniques
  float t_v, p_v;
  float lastBTFvalue[3];

  // ... rotation of the OBJ vertices and TBN vectors by angle 'objectShift'

  // Here is the lightness coefficient computation to reach acceptable range
  // visible on the display, otherwise it does not play role in the computation
  // and it is used as a simple multiplier
  float normF;
  GetTonemappingMult(normF);

  // A single point light for the whole model
  if (!IsEnvMapLighting())
    TA->setIlluPos(lpos);
  
  //printf("                triangle processing from OBJ ");fflush(stdout);
  // for all OBJ triangles, i.e. faces
  // OBJ model's faces rendering 
  // for all faces
  for (int i=1; i<= OBJ->noOfFaces(); i++) {
    // if (envLighting)
    //   cout << "+" << fflush;
#if 0
    if (saveAndExit) {
      printf(" %",i);
      fflush(stdout);
    }
#endif
    // triangle rasterization.............
    
    // compute rasterized UV coords and boundary rectangle around triangle
    minX =  1000000;
    maxX = -1000000;
    minY =  1000000;
    maxY = -1000000;
    for(int k=0;k<3;k++) { // for 3 vertices of a triangle
      assert(UV[actualOBJ][i][k][0]<=1.f);
      assert(UV[actualOBJ][i][k][0]>=0.f);
      assert(UV[actualOBJ][i][k][1]<=1.f);
      assert(UV[actualOBJ][i][k][1]>=0.f);
      UVr[k][0] = (int)(UV[actualOBJ][i][k][0]*(float)BTFrows);
      UVr[k][1] = (int)(UV[actualOBJ][i][k][1]*(float)BTFcols);
      if (UVr[k][0]<minX) minX = UVr[k][0];
      if (UVr[k][1]<minY) minY = UVr[k][1];
      if (UVr[k][0]>maxX) maxX = UVr[k][0];
      if (UVr[k][1]>maxY) maxY = UVr[k][1];
    }

    // for all pixels in triangle, i.e. face
    for(int irow=minY;irow<=maxY;irow++) {
      for(int jcol=minX;jcol<=maxX;jcol++) {
	// 1D texture index computation
	int BTFindex = irow*BTFcols*4 + jcol*4;
	if (BTFindex >= BTFrows*BTFcols*4)
	  continue;
	if (BTFindex < 0)
	  continue;
	assert(BTFindex < BTFrows*BTFcols*4);
#if 0
	if (BTFimageF[BTFindex+2] != valCheck)
	  continue; // the pixel has already been filled in by neighboring triangle
#endif	    

#if 0
	// debugging
	bool bord1=false;
	if ((irow == 300) && (jcol == 500)) {
	  BTFimageF[BTFindex] = 255.f; 
	  BTFimageF[BTFindex+1] = 255.f;
	  BTFimageF[BTFindex+2] = 0.f;
	  BTFimageF[BTFindex+3] = 0.f;
	  //continue;
	  bord1 = true;
	}
#endif

	uv[0] = (float)jcol/(float)BTFcols;
	uv[1] = (float)irow/(float)BTFrows;
	
	// barycenric weights computation:
	//  b0 =  (x2 - x1) * (y3 - y1) - (x3 - x1) * (y2 - y1)
	//  b1 = ((x2 - x0) * (y3 - y0) - (x3 - x0) * (y2 - y0)) / b0
	//  b2 = ((x3 - x0) * (y1 - y0) - (x1 - x0) * (y3 - y0)) / b0
	//  b3 = ((x1 - x0) * (y2 - y0) - (x2 - x0) * (y1 - y0)) / b0 
	// Then:if b1, b2, and b3 are all > 0, (x0,y0) is strictly inside the triangle; 
	//      if bi = 0 and the other two coordinates are positive, 
	//                   (x0,y0) lies on the edge opposite (xi,yi); 
	//      if bi and bj = 0, (x0,y0) lies on (xk,yk); 
	//      if bi < 0, (x0,y0) lies outside the edge opposite (xi,yi); 
	//      if all three coordinates are negative, something else is wrong. 
	// This method does not depend on the cyclic order of the vertices.
	float t1,t2,t3,t4,t5,t6; 
	t1 = uv[1] - UV[actualOBJ][i][0][1];
	t2 = uv[0] - UV[actualOBJ][i][0][0];
	t3 = UV[actualOBJ][i][1][0]-UV[actualOBJ][i][0][0];
	t4 = UV[actualOBJ][i][2][0]-UV[actualOBJ][i][0][0];
	t5 = UV[actualOBJ][i][1][1]-UV[actualOBJ][i][0][1];
	t6 = UV[actualOBJ][i][2][1]-UV[actualOBJ][i][0][1];
	den = t3 * t6 - t4 * t5;
	if (den == 0)
	  den = 2.e-7;
	den = 1.0f/den;
	weight1 = (t6*t2 - t1*t4) * den;
	weight2 = (t1*t3 - t5*t2) * den;
	weight0 = 1.f - weight1 - weight2;
	
	// if we are outside of the triangle ->  skip the following processing
	float thresh = -0.1;
	if(weight0<thresh || weight1<thresh || weight2<thresh) {
	  if(BTFimageF[BTFindex+2] == valCheck) {
	    // to avoid artifacts, copy last rendered value
	    BTFimageF[BTFindex]   = lastBTFvalue[0];
	    BTFimageF[BTFindex+1] = lastBTFvalue[1];
	    BTFimageF[BTFindex+2] = lastBTFvalue[2];
	    BTFimageF[BTFindex+3] = 0.f; // alpha
	  }
	  continue;
	}
	
	// v, T, N, B interpolation using barycentric weights
	for(int j=0;j<3;j++) { // for u,v
	  //cout << V[actualOBJ][i][0][j] << " ";

	  // vertex
	  Vint[j] = weight0*V[actualOBJ][i][0][j] + 
	    weight1*V[actualOBJ][i][1][j] + weight2*V[actualOBJ][i][2][j];
	  // tangent
	  Tint[j] = weight0*T[actualOBJ][i][0][j] + weight1*T[actualOBJ][i][1][j] +
	    weight2*T[actualOBJ][i][2][j];
	  // normal
	  Nint[j] = weight0*N[actualOBJ][i][0][j] + weight1*N[actualOBJ][i][1][j] +
	    weight2*N[actualOBJ][i][2][j];
	  // bitangent
	  Bint[j] = weight0*B[actualOBJ][i][0][j] + weight1*B[actualOBJ][i][1][j] +
	    weight2*B[actualOBJ][i][2][j];
	}
	//cout << "end vertex" << endl;
	// This has to be done for vector, DO NOT REMOVE IT (VH)!
	vec_norm(Tint);
	vec_norm(Nint);
	vec_norm(Bint);	
	// viewing angles computations per vertex, results are in degrees
	TA->computeViewAngles(Vint, Tint, Bint, Nint, &t_v, &p_v);
	
	// correcting precision error in the rendering from database
	if (p_v < 0.00002)
	  p_v = 0.0002;
	// the triangle is not visible ->  just copy the pixel from
	// the last computation and go on to the next texel
	if (t_v > M_PI/2.f) {
#if 1
	  // go to the next pixel
	  if (BTFimageF[BTFindex+2] == valCheck) {	    
	    // either last rendered value to avoid artifacts for theta_v close to M_PI/2
	    // or black, if we are interested only in that we do render
	    BTFimageF[BTFindex]   = lastBTFvalue[0];
	    BTFimageF[BTFindex+1] = lastBTFvalue[1];
	    BTFimageF[BTFindex+2] = lastBTFvalue[2];
	    BTFimageF[BTFindex+3] = 0.f;
	  }
	  continue;
#else
	  t_v = M_PI - t_v; // render it as it were visible from the opposite direction
#endif	  
	}
#ifdef DEBUG_RESTRICT_RENDERING_TO_CENTER_IMAGE
	// To speed up debugging only for some setting of large number of samples
	// pickups only such parts of model, where the normal is towards the viewing vector
	if (t_v > THRESHOLD_MAX_ANGLE) {
	  BTFimageF[BTFindex] = BTFimageF[BTFindex+1] =
	    BTFimageF[BTFindex+2] = BTFimageF[BTFindex+3] = 0.f;
	  continue;
	}
#endif	

	// checking again, if t_v is all right
	if (t_v >= (M_PI/2.f-1e-6))
	  t_v = M_PI/2.f-1e-6;
	assert(t_v <= M_PI/2.f);
	
	// illumination computation
	for(int isp = 0; isp<=3; isp++)
	  BTFimageF[BTFindex+isp] = 0.f;
	debugPosX = jcol;
	debugPosY = irow;
#ifdef DEBUG_SINGLE_PIXEL
	const int jcolTest = 100;
	const int irowTest = 100;
	if ((jcol != jcolTest)||
	    (irow != irowTest)) { // set it black and continue
	  for (int isp=0; isp<3;isp++)
	    BTFimageF[BTFindex+isp] = 0;
	  continue;
	}
#endif	
	// for texturing
	const int irow2 = irow % REPEATX;
	const int jcol2 = jcol % REPEATY;
	
	// Here we fill in the value of a texture by computed value
	// given 3D point, normal, tangent, bitangent vectors
	switch(localModeCompute) {
	case 0: {
	  // single point light
	  computeMode0(Vint, Tint, Nint, Bint, 
		       t_v, p_v, irow2, jcol2, BTFindex, normF);
	  break;
	}
	case 1:  {
	  // the same set of directional lights for the whole image
	  computeMode1(Vint, Tint, Nint, Bint, 
		       t_v, p_v, irow2, jcol2, BTFindex, normF);
	  break;
	}
	case 2: {
	  // each pixel unique (different) set of directional lights from EM
	  // computed for each pixel separately
	  computeMode2(Vint, Tint, Nint, Bint, 
		       t_v, p_v, irow2, jcol2, BTFindex, normF);
	  break;
	}
	case 3: {// random sampling according to BRDF*cos(theta)
	  computeMode3(Vint, Tint, Nint, Bint, 
		       t_v, p_v, irow2, jcol2, BTFindex, normF);
	  break;
	}
	case 4: {// random sampling according BRDF*cos(theta)
	  // and EM sampling together, combined by MIS (=multiple importance sampling)
	  computeMode4(Vint, Tint, Nint, Bint, 
		       t_v, p_v, irow2, jcol2, BTFindex, normF);
	  break;
	}
	case 5: {// random sampling according BRDF*cos(theta)
	  // and EM sampling together, combined by MIS (=multiple importance sampling)
	  computeMode5(Vint, Tint, Nint, Bint, 
		       t_v, p_v, irow2, jcol2, BTFindex, normF);
	  break;
	}
        } // switch

	// copy the last rendered pixel
	for (int isp=0; isp<=2; isp++) { 
	  BTFimageF[BTFindex+isp];
	  lastBTFvalue[isp] = BTFimageF[BTFindex+isp];
	}
	BTFimageF[BTFindex+3] = 0.0f; // setting alpha channel value       	
	// luminance statistics
	float lum = luminance(&(BTFimageF[BTFindex]));
	if ((!isnan_local(lum)) && (lum >= 0.0f)) {
	  sumLum += (long double)lum;
	  cntPixelsRendered++;
	}
	// printing time estimate is allowed ?
	if ((printEstimateFlag) && (cntPixelsRendered % 10 == 0)) {
	  static float lastRealTime = 0;
	  float userTime = timer.RealTime();
	  int pixelsToCompute = cntEstimatedPixels-cntPixelsRendered;
	  if (pixelsToCompute <= 0) pixelsToCompute = 1;
	  // compute estimated time in minutes
	  float timeEstimated = userTime/(float)cntPixelsRendered*(float)pixelsToCompute;
	  if (userTime - lastRealTime > intervalToPrintEstimate) { // here in seconds
	    printf("\r                                                ");
	    printf("\rEtime [min.] %3.2f                                 \n", timeEstimated/60);
	    fflush(stdout);
	    lastRealTime = userTime;
	    // now we will be printing estimate in different interval
	    firstEstimateOfTime = 0; 
	  }
	  if (!firstEstimateOfTime) {
	    if (timeEstimated > 60) { // seconds
	      intervalToPrintEstimate = 30; // print out estimate each 30 seconds
	    }
	    if (timeEstimated > 600) { // seconds
	      intervalToPrintEstimate = 120; // print out estimate each 1 minute
	    }
	    if (timeEstimated > 3600) { // seconds
	      intervalToPrintEstimate = 360; // print out estimate each 6 minutes
	    }
	    if (timeEstimated > 18000) { // seconds
	      intervalToPrintEstimate = 1800; // print out estimate each 30 minutes
	    }	    
	  } // the interval is not changed
	} // estimating time to computed

	if (pseudoColorTechnique) {
	  if (visualizationMode == 0) {
	    // the value to make visualization - but avoid it during visualization pass
	    const float value = (float)evalBRDFcntST/(float)(evalBRDFcntST + evalEMcntST+1e-12);
	    //cout << value << " ";
	    // cout << value << " " << endl;
	    GetPseudoColor(pseudoColorTechnique, value, 0.0f, 1.0f, // store pseudocolor to texture
			   &(BTFimageB[1][BTFindex])); // prior rendering
	  } else {
	    float vv = (evalValueVIS - minValueVIS)/(maxValueVIS - minValueVIS);
	    //cout << vv << " V";
	    if (vv < 0.0f) vv = 0; if (vv > 1.0f) vv = 1.0f;
	    GetPseudoColor(pseudoColorTechnique, vv, 0.0f, 1.0f, // store pseudocolor to texture
			   &(BTFimageB[1][BTFindex])); // prior rendering	    
	  }	  
	} // if
#if 0
	goto finish_label;
#endif	
      } // for jcol
    } // for irow
  } // for all faces
  // for debugging only the first pixel
finish_label:
  if (printEstimateFlag) {
    // this is the estimate for the next frame, how many pixels will be rendered
    cntEstimatedPixels = cntPixelsRendered;
    intervalToPrintEstimate = 10; // initial interval after to make estimate
    firstEstimateOfTime = 1; // estimating again for the next method
    printf("\r                                                           \r"); // clear the line
  }

  if (modeCompute > 0) {
    cout << "MC: " << modeCompute << " avgLum = " << sumLum/(long double)cntPixelsRendered
	 << " cntPixelsRendered=" << cntPixelsRendered << endl;
  }
  if (modeCompute > 0) {
    cout << "evalBRDFcnt[%]= " << 100.f*(float)evalBRDFcnt/(float)(evalBRDFcnt+evalEMcnt+1e-12)
	 << " stats. renderedPix= " << cntPixelsRendered
	 << " samplesPerPix= " << (float)(evalBRDFcnt+evalEMcnt)/(long double)cntPixelsRendered << endl;
    if (failBRDFcnt) 
      cout << "failBRDFcnt = " << failBRDFcnt << " evalBRDFcount= " << evalBRDFcnt
	   << " r[%] = " << (float)100.f*failBRDFcnt/(float)(evalBRDFcnt+1) << " ";
    if (failEMcnt)
      cout << "failEMcnt = " << failEMcnt << " evalEMcount= " << evalEMcnt
	   << " r[%] = " << (float)100.f*failEMcnt/(float)(evalEMcnt+1) << " ";
    if (failBRDFcnt+failEMcnt > 0) cout << endl;
  }
  failBRDFcnt = 0;
  evalBRDFcnt = 0;
  failEMcnt = 0;
  evalEMcnt = 0;
  
  // compute the time for single frame
  timer.Stop();
  float userTime = timer.UserTime();
  float realTime = timer.RealTime();
  UpdateFPS(userTime, realTime);
  if (modeCompute > 0)
    cout << "RealTime = " << realTime << endl;
  
  if (saveAndExit)
    printf("Time: user - %f, real - %f\n", userTime, realTime);
  
  // Rendering of the first frame is always done by point light
  if (cntFrame == 0) {
    modeCompute = modeComputeSave;
  }  
  cntFrame++; // go to the next frame
  // rendering progress printing
  fflush(stdout);
  return;
}//--- renderOBJECT --------------------------------------------------------

void 
grabBufferLDR(int screenShotNumber, int useOrigData, bool visflag, int pass, bool fbFlag)
{
  if (screenShotNumber < 0)
    return;

  unsigned char *buffer = ballocation1(0, 4*win_fbo_width*win_fbo_height-1);

  if (fbFlag) {
    // from framebuffer
    glReadBuffer(GL_COLOR_ATTACHMENT0);
    glReadPixels(0, 0, win_fbo_width, win_fbo_height, GL_RGBA, GL_FLOAT, buffer);
  }
  else {
    // from screen buffer
    glReadBuffer(GL_BACK);
    glReadPixels(0, 0, win_fbo_width, win_fbo_height, GL_RGBA, GL_FLOAT, buffer);
  }

  char filename[200];  
  makeFilename(false, false, visflag, screenShotNumber, pass, filename);
  strcat(auxString,".png");

  if (compareScreenShots==0) {
    int writeStatus = writePNGfile1D(auxString, buffer,3, win_fbo_height, win_fbo_width, 1);
    printf("   Writing framebuffer to '%s' file with signal %d      "
	   "                                                                                        ",
	   auxString,writeStatus);
    fflush(stdout);
  }
  else {
    // reading original rendered image
    char origName[1000], aux2[1000];
    makeFilename(false, true, visflag, screenShotNumber, pass, aux2);
    strcat(aux2,".png");
    char originalRenderingsPath[1000];
    sprintf(origName,"./%s%s", originalRenderingsPath, aux2);
    //          win_fbo_width = win_fbo_height = 1000;
    unsigned char *origBuffer = ballocation1(0, 4*win_fbo_width*win_fbo_height-1);
    int readStatus = readPNGfile(origName, origBuffer, 3, win_fbo_width, win_fbo_height, 1);
    printf("Original image '%s' (%dx%d) read with status: %d\n",
	   origName, win_fbo_height, win_fbo_width, readStatus);
    if (readStatus!=1)
      printf("Sorry the original image for comparison is not available!!!\n");

    // copying original and model images side by side
    unsigned char *doubleBuffer = ballocation1(0, 4*2*win_fbo_width*win_fbo_height-1);
    for(int irow=0;irow<win_fbo_height;irow++)
      for(int jcol=0;jcol<win_fbo_width;jcol++)
	for(int isp=0;isp<=2;isp++) {
	  int indexSingle = 4*(irow*win_fbo_width+jcol)+isp;
	  int indexSingleV = 4*((win_fbo_height-irow)*win_fbo_width+jcol)+isp;
	  int indexDoubleL = 4*(irow*2*win_fbo_width+jcol)+isp;
	  int indexDoubleR = 4*(irow*2*win_fbo_width+win_fbo_width+jcol)+isp;
	  doubleBuffer[indexDoubleR] = buffer[indexSingle];
	  if (readStatus==1)// original image was readed correctly
	    doubleBuffer[indexDoubleL] = origBuffer[indexSingleV];
	}
    // saving comparison image
    int writeStatus = writePNGfile1D(auxString, doubleBuffer,3, win_fbo_height, 2*win_fbo_width, 1);
    printf("   Writing comparison image to '%s' file with signal %d"
	   "                                    ",
	   auxString,writeStatus); fflush(stdout);

    freebmem1(doubleBuffer, 0,4*2*win_fbo_width*win_fbo_height-1);
    freebmem1(origBuffer, 0,4*win_fbo_width*win_fbo_height-1);
  }
  freebmem1(buffer, 0,4*win_fbo_width*win_fbo_height-1);

  return;
}//--- grabBufferLDR -------------------------------------------------------------

void 
grabBufferHDR(int screenShotNumber, int useOrigData, bool visflag, int pass, int bufferFlag)
{
  if (screenShotNumber < 0)
    return;
  char filename[200];  
  makeFilename(false, false, visflag, screenShotNumber, pass, filename);

  //  glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
  float *buffer = new float[win_fbo_width * win_fbo_height * 4];
  if (bufferFlag) {
    // read from frame buffer
    glBindRenderbuffer(GL_RENDERBUFFER, fbo);
    glClampColor(GL_CLAMP_READ_COLOR, GL_FALSE);
    glReadBuffer(GL_COLOR_ATTACHMENT0);
    glClampColor(GL_CLAMP_READ_COLOR, GL_FALSE);
    glReadPixels(0, 0, win_fbo_width, win_fbo_height, GL_RGBA, GL_FLOAT, buffer);
    CHECK_GL_ERROR();
    glBindRenderbuffer(GL_RENDERBUFFER, 0);
    CHECK_GL_ERROR();
    glBindFramebuffer(GL_FRAMEBUFFER, 0);
    CHECK_GL_ERROR();
  }
  else {
    // from screen buffer
    glBindRenderbuffer(GL_RENDERBUFFER, 0);
    CHECK_GL_ERROR();
    glReadBuffer(GL_BACK);
    CHECK_GL_ERROR();
    glReadPixels(0, 0, win_fbo_width, win_fbo_height, GL_RGBA, GL_FLOAT, buffer);
    CHECK_GL_ERROR();
    //    glReadBuffer(GL_BACK);
    // glReadPixels(0, 0, win_fbo_width, win_fbo_height, GL_RGBA, GL_FLOAT, buffer);
  }
#if 0
  // read data from framebuffer of OpenGL API
  glReadBuffer(GL_BACK);
  glReadPixels(0, 0, win_fbo_width, win_fbo_height, GL_RGBA, GL_FLOAT, buffer);
#endif
  if (suffixAlbedoNormalized) {
    if (suffixGrayTone)
      strcat(filename, ".GA");// gray and normalized albedo
    else
      strcat(filename, ".A");// normalized albedo      
  }
  else {
    if (suffixGrayTone)
      strcat(filename, ".G");// gray
  }
  
  // ------------------------------------------------------------------------
  strcat(filename, FILEEXT);	   
  if (compareScreenShots==0) {
    // writing a single file
    CHDRImage hdri;
    float win_fbo_width_new = win_fbo_width;
    if (visflag)
      win_fbo_width_new += widthBargraph + borderBargraph;
    hdri.allocate(win_fbo_width_new, win_fbo_height, true);
    // Four channels
    hdri.copyDataFromArray(win_fbo_width_new, win_fbo_height, 4, buffer, win_fbo_width, 0, 0, true);

    // copy the bargraph values to the right of the image
    if (visflag) {
      for (int y = 0; y < win_fbo_height; y++) {
	float RGB[3];	
	float value = (float)(win_fbo_height-y)/win_fbo_height; // get the color
	GetPseudoColor(pseudoColorTechnique, value, 0.0, 1.0, RGB);
	for (int x = win_fbo_width; x < win_fbo_width+borderBargraph; x++) {
	  hdri.setRGB_XY(x, y, 0, 0, 0); // make it black
	}
	for (int x = win_fbo_width+borderBargraph; x < win_fbo_width+borderBargraph+widthBargraph; x++) {
	  hdri.setRGB_XY(x, y, RGB[0], RGB[1], RGB[2]);	
	} // for x
      } // for y
    } // if to do vis
    
    bool writeStatus = hdri.saveHDR(filename);
    printf("Writing framebuffer to '%s' file with signal %d "
	   "                                                                         ",
	   filename, writeStatus ? 1:0); fflush(stdout);
    CHDRImage *hdri2=0;
    delete []buffer;
    return;
  }

  // Making possibly double picture reading original rendered image
  char origName[1000]; char aux2[1000];
  makeFilename(false, true, visflag, screenShotNumber, pass, aux2);
  strcat(aux2,".hdr");
  char originalRenderingsPath[1000];
  sprintf(origName,"./%s%s", originalRenderingsPath, aux2);
  CHDRImage hdrOrig;
  bool readStatus = hdrOrig.loadHDR(aux2);
  printf("Original image '%s' (%dx%d) read with status: %d\n",
	 origName, win_fbo_width, win_fbo_height, readStatus ? 1:0);
  if (!readStatus)
    printf("Sorry the original image for comparison is not available!!!\n");

  CHDRImage hdrnew; // combined image
  int offsetx = 0;
  if (readStatus) {
    float *bufOrig = hdrOrig.getBuf();
    offsetx = win_fbo_width;
    hdrnew.allocate(win_fbo_width*2, win_fbo_height);
    hdrnew.copyDataFromArray(win_fbo_width, win_fbo_height, 3, bufOrig, win_fbo_width, 0, 0);
  }
  else
    hdrnew.allocate(win_fbo_width, win_fbo_height);
  // here we copy the data read from the OpenGL
  hdrnew.copyDataFromArray(win_fbo_width, win_fbo_height, 3, buffer, win_fbo_width, offsetx, 0);
  // and here we save the whole image
  hdrnew.saveHDR(filename);

  delete []buffer;
  return;
}//--- grabBufferHDR -------------------------------------------------------------

void 
grabBuffer(int screenShotNumber, int useOrigData, bool visflag, int pass, int bufferFlag)
{
  if (grabHDRflag)
    grabBufferHDR(screenShotNumber, useOrigData, visflag, pass, bufferFlag); // HDR format EXR/HDR
  else
    grabBufferLDR(screenShotNumber, useOrigData, visflag, pass, bufferFlag); // PNG format
  return;
}

void
renderEnvMapLights(int cnt = 0)
{
  if (cnt == 0) {
    cnt = cntSamplesPerPixel;
  }
  cnt = maxCntSamplesPerPixel;
  
  // environment lights visualization
  if (renderLights) {
    float *lightsvertices = new float[3*cnt];
    bool flag = environMap[actualEnvMap].getModeRandomLights();
    environMap[actualEnvMap].setModeRandomLights(false);
    for (int i = 0; i < cnt; i++) {
      lightsvertices[0+i*3] = lightsRadius * environMap[actualEnvMap].getDirectionPrecomputedLight(i, 0);
      lightsvertices[1+i*3] = lightsRadius * environMap[actualEnvMap].getDirectionPrecomputedLight(i, 1);
      lightsvertices[2+i*3] = lightsRadius * environMap[actualEnvMap].getDirectionPrecomputedLight(i, 2);
      //      cout << i << " " << lightsvertices[0+i*3] << " " << lightsvertices[1+i*3] << " "
      //	   << lightsvertices[2+i*3] << endl;
    }
    CCleanMesh meshlights;
    meshlights.init(cnt, GL_POINTS, lightsvertices);
    environMap[actualEnvMap].setModeRandomLights(flag);

    // Move the model somewhere by model matrix
    glm::mat4 envmapMatrixEM = glm::mat4(1.0f);
    // Environment lights are rotated really as data, so the rotation matrix again were wrong
    //envmapMatrixEM = glm::rotate(envmapMatrixEM, (float)((envPhiShift) * 180.f/M_PI), glm::vec3(0, 0, 1));
    glm::mat4 PVMmatrix = projectionMatrix * viewMatrix * envmapMatrixEM;

    // Draw the mesh
    glPointSize(4.0f);
    meshlights.setColor(1.f, 0.f, 0.f); // red
    meshlights.draw(PVMmatrix);
  }

  if (renderAxis) {
    // Move the model somewhere by model matrix
    glm::mat4 modelMatrixEM = glm::mat4(1.0f);
    glm::mat4 PVMmatrix = projectionMatrix * viewMatrix * modelMatrixEM;

    // render coordinate axes
    int cnt = 1000;
    // X-axis
    float *verticesX = new float[3*cnt];
    float sizeAxis = 10.f;
    for (int i = 0; i < cnt; i++) {
      verticesX[0+i*3] = sizeAxis/(float)cnt * (float)i + 0.01f;
      verticesX[1+i*3] = 0.f;
      verticesX[2+i*3] = 0.f;
    } // for i
    CCleanMesh meshX;
    meshX.init(cnt, GL_POINTS, verticesX);
    // Draw the mesh
    glPointSize(1.5f);
    meshX.setColor(1.f, 0.f, 0.f); // red for X-axis
    meshX.draw(PVMmatrix);
    delete[]verticesX;

    // Y-axis
    float *verticesY = new float[3*cnt];
    for (int i = 0; i < cnt; i++) {
      verticesY[0+i*3] = 0.f;
      verticesY[1+i*3] = sizeAxis/(float)cnt * (float)i + 0.01f;
      verticesY[2+i*3] = 0.f;
    } // for i
    CCleanMesh meshY;
    meshY.init(cnt, GL_POINTS, verticesY);
    // Draw the mesh
    glPointSize(1.5f);
    meshY.setColor(0.f, 1.f, 0.f); // green for Y-axis
    meshY.draw(PVMmatrix);
    delete[]verticesY;

    // Z-axis
    float *verticesZ = new float[3*cnt];
    for (int i = 0; i < cnt; i++) {
      verticesZ[0+i*3] = 0.f;
      verticesZ[1+i*3] = 0.f;
      verticesZ[2+i*3] = sizeAxis/(float)cnt * (float)i + 0.01f;
    } // for i
    CCleanMesh meshZ;
    meshZ.init(cnt, GL_POINTS, verticesZ);
    // Draw the mesh
    glPointSize(1.5f);
    meshZ.setColor(0.f, 0.f, 1.f); // blue for Z-axis
    meshZ.draw(PVMmatrix);
    delete[]verticesZ;
  }  
  return;
}

void
renderEnvMapOpenGL()
{
  if (renderEnvMap) {  
    // Move the model somewhere by model matrix
    glm::mat4 envmapMatrixEM = glm::mat4(1.0f);
    envmapMatrixEM = glm::rotate(envmapMatrixEM, (float)((envPhiShift) * 180.f/M_PI), glm::vec3(0, 0, 1));
    glm::mat4 PVMmatrix = projectionMatrix * viewMatrix * envmapMatrixEM;

    // Draw the mesh
    meshenvmap[actualEnvMap].draw(PVMmatrix, envmapMatrixEM);
  }
  return;
}

void 
renderSceneSimple(glm::mat4 PVMmatrix, bool useFrameBuffer, bool saveFlag, int considerExit)
{
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  CHECK_GL_ERROR();
  // Attach always render buffer only
  glBindRenderbuffer(GL_RENDERBUFFER, 0);
  //cout << "DISPLAY" << endl << flush;
  glViewport(0, 0, win_fbo_width, win_fbo_height);
  initProjectionMatrix(win_fbo_width, win_fbo_height);
  CHECK_GL_ERROR();

  // Draw the mesh with texture
  //=== scene objects rendering ===================================
  glEnable(GL_DEPTH_TEST);
  CHECK_GL_ERROR();
  glFrontFace(GL_CW); // GL_CW and GL_CCW
  CHECK_GL_ERROR();
  glDisable(GL_CULL_FACE);
  CHECK_GL_ERROR();
  //glCullFace(GL_FRONT);

  //mesh[actualOBJ].detachTextureData();
  mesh[actualOBJ].setTextureData(BTFimageF, BTFcols, BTFrows);
  //mesh[actualOBJ].setTextureData(0, BTFcols, BTFrows);
  mesh[actualOBJ].draw(PVMmatrix);
  CHECK_GL_ERROR();
  
  // render also environment map
  glDisable(GL_CULL_FACE);
  CHECK_GL_ERROR();
  glCullFace(GL_FRONT);
  CHECK_GL_ERROR();

  renderEnvMapOpenGL();
  CHECK_GL_ERROR();
  renderEnvMapLights();
  CHECK_GL_ERROR();

  // DrawNeedles
  if (drawNeedles)
    DrawNeedles(drawNeedles, 1, mesh[actualOBJ].getCntVertices()/3); 
  
  if ((grabbingSequence)&&(saveFlag)) {
    bool useOrigData = false;
    if (useFrameBuffer) {
      // Save it to the image
      glBindFramebuffer(GL_FRAMEBUFFER, fbo);
      CHECK_GL_ERROR();
      //grabBuffer(int screenShotNumber, int useOrigData,
      //            bool visflag, int pass, int bufferFlag)
      // HDR format from screen buffer
      grabBuffer(screenShotNumber++, useOrigData, 0, -1, 0); 
      CHECK_GL_ERROR();
    }
    else
      glBindRenderbuffer(GL_RENDERBUFFER, 0);
    printf("  Frame: %04d saved", screenShotNumber-1);fflush(stdout);
    // finish rendering
    if (screenShotNumber == NoOfScreenShots) {
      printf(" Video sequence rendering of %d frames finished.\n",NoOfScreenShots);
      deallocations();
      exit(0);
    }

    // scene change for the next frame
    if (rotateObject) {
      OBJrot[actualOBJ][2] += 2.f*M_PI/(float)phiStepsPerCircle;
      if (OBJrot[actualOBJ][2] >= 2*M_PI)
	OBJrot[actualOBJ][2] -= 2*M_PI;      
    }
    if (rotateEnvironment) {
      envPhiShift += 2.f*M_PI/(float)phiStepsPerCircle;   // environment map shift
      if (envPhiShift >= 2*M_PI)
	envPhiShift -= 2*M_PI;      
      
      rotateEnvironmentLights(2.f*M_PI/(float)phiStepsPerCircle); // corresponding env. lights shift
    }
  }

  if (saveAndExit && saveFlag) {
    // saves the image and exits
    bool useOrigData = false;
    // grabBuffer(int screenShotNumber, int useOrigData, bool visflag, int pass, int bufferFlag)
    grabBuffer(screenShotNumber++, useOrigData, 0, -1, 0); // HDR format from screen buffer

    if ((indexRenderedImage == cntFinishRendering)&&considerExit) {
      glutLeaveMainLoop();
      glutExit();
      glutDestroyWindow(windowID);
      deallocations();
      exit(0);
    }
  }
  glutSwapBuffers();
  //  glFlush();
  glutPostRedisplay();
  
  return;
}

void
renderToFrameBuffer(glm::mat4 PVMmatrix, bool useFrameBuffer, bool saveFlag,
		    bool visFlag, bool considerExit)
{
  // render to frame buffer only and save, without showing to screen
  // Return to onscreen rendering for now
  bool howToGrabBuffer = false; // HDR
  if (useFrameBuffer) {
    glBindRenderbuffer(GL_RENDERBUFFER, 0);
    //CHECK_GL_ERROR();	  
    glBindFramebuffer(GL_FRAMEBUFFER, fbo);
    //CHECK_GL_ERROR();	  
    howToGrabBuffer = true; // HDR
  }
  else {
    glBindFramebuffer(GL_FRAMEBUFFER, 0);
    CHECK_GL_ERROR();	  
    glBindRenderbuffer(GL_RENDERBUFFER, 0);
    CHECK_GL_ERROR();	  
    //howToGrabBuffer = false; // LDR
  }
  const int lastPass = 3;
  for (int pass = 0; pass <= lastPass; pass++) {
    //glCullFace(GL_FRONT);
    //cout << "pass = " << pass << endl;

    //cout << "DISPLAY" << endl << flush;
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    CHECK_GL_ERROR();
    glViewport(0, 0, win_fbo_width, win_fbo_height);
    initProjectionMatrix(win_fbo_width, win_fbo_height);
    CHECK_GL_ERROR();

    // Draw the mesh with texture
    //=== scene objects rendering ===================================
    glEnable(GL_DEPTH_TEST);
    glFrontFace(GL_CCW); // GL_CW and GL_CCW
    glDisable(GL_CULL_FACE);
    //=== modelview setting =========================================
    // now we can bind texture which was filled in above
    // Ordinary rendering with prepared floating texture
    //mesh[actualOBJ].detachTextureData();
    if (visFlag) {
      // with visualization
      CHECK_GL_ERROR();
      glBindTexture(GL_TEXTURE_2D, BTFidTexture[1]); // visualization texture
      CHECK_GL_ERROR();
      setTexParams();
      CHECK_GL_ERROR();
      // load the data
      glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA32F, BTFrows, BTFcols,
		   0, GL_RGBA, GL_FLOAT, BTFimageB[1]);
      CHECK_GL_ERROR();
      glActiveTexture(GL_TEXTURE0 + 0);
      CHECK_GL_ERROR();
      glBindTexture(GL_TEXTURE_2D, BTFidTexture[1]); // visualization
      CHECK_GL_ERROR();
      //cout << "H 1 BTFimageB[1]" << endl;
      mesh[actualOBJ].setTextureData(BTFimageB[1], BTFcols, BTFrows);
    }
    else {
      // no visualization
      //CHECK_GL_ERROR();
      //glBindTexture(GL_TEXTURE_2D, 0);
      CHECK_GL_ERROR();
      //cout << "currTexID = " << currTexID << endl;
      glBindTexture(GL_TEXTURE_2D, BTFidTexture[currTexID]); // data give by index, data directly
      CHECK_GL_ERROR();
      setTexParams();
      CHECK_GL_ERROR();
      glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA32F, BTFrows, BTFcols,
		   0, GL_RGBA, GL_FLOAT, BTFimageB[currTexID]);
      CHECK_GL_ERROR();
      glActiveTexture(GL_TEXTURE0 + 0);
      CHECK_GL_ERROR();
      glBindTexture(GL_TEXTURE_2D, BTFidTexture[currTexID]); // data give by index
      CHECK_GL_ERROR();
      // cout << "H " << currTexID << " BTFimageB[currTexID] " << BTFimageB[currTexID] << endl;
      mesh[actualOBJ].setTextureData(BTFimageB[currTexID], BTFcols, BTFrows);
    }

    //mesh[actualOBJ].setTextureData(0, BTFcols, BTFrows);
    mesh[actualOBJ].draw(PVMmatrix);
    
    // render also environment map
    if (IsEnvMapLighting()) {
      glDisable(GL_CULL_FACE);
      glCullFace(GL_BACK);
      renderEnvMapOpenGL();
      // Possibly the position of lights for environment map
      renderEnvMapLights();
      CHECK_GL_ERROR();
    }
    
    if ((grabbingSequence)&&(pass=lastPass)&&(saveFlag)) {
      // only in the second pass
      bool gb = false;
      // grabBuffer(int screenShotNumber, int useOrigData, bool visflag, int pass, int bufferFlag)
      grabBuffer(screenShotNumber++, gb, visFlag, -1, howToGrabBuffer); // HDR format from frame buffer
      
      printf("  Frame: %04d saved", screenShotNumber-1);fflush(stdout);
      // finish rendering
      if (screenShotNumber == NoOfScreenShots) {
	printf(" Video sequence rendering of %d frames finished.\n",NoOfScreenShots);
	deallocations();
	exit(0);
      }

      // scene change for the next frame
      if (rotateObject && (pass == lastPass)) {
	OBJrot[actualOBJ][2] += 2.f*M_PI/(float)phiStepsPerCircle;   // sphere rotation
	if (OBJrot[actualOBJ][2] >= 2*M_PI)
	  OBJrot[actualOBJ][2] -= 2*M_PI;      
      }
      if (rotateEnvironment) {
	envPhiShift += 2.f*M_PI/(float)phiStepsPerCircle;   // environment map shift
	if (envPhiShift >= 2*M_PI)
	  envPhiShift -= 2*M_PI;      
	
	rotateEnvironmentLights(2.f*M_PI/(float)phiStepsPerCircle); // corresponding env. lights shift
      }
    }

    if (saveAndExit && (pass==lastPass)&&(saveFlag)) {
      // saves the image and exits
      bool useOrigData = false;
      // grabBuffer(int screenShotNumber, int useOrigData, bool visflag, int pass, int bufferFlag)
      grabBuffer(screenShotNumber++, useOrigData, visFlag, -1, howToGrabBuffer); // HDR format from frame buffer
      if ((indexRenderedImage == cntFinishRendering)&&(considerExit)) {
	glutLeaveMainLoop();
	glutExit();
	glutDestroyWindow(windowID);
	deallocations();
	exit(0);
      }
    }
    glFlush();
    glutSwapBuffers();
    glutPostRedisplay();
  } // for pass    

  // Disable rendering with texture
  glDisable(GL_TEXTURE_2D);    

  // Return to onscreen rendering for now
  glBindFramebuffer(GL_FRAMEBUFFER,0);
  return;  
}

void
renderToScreen(glm::mat4 PVMmatrix, bool saveFlag, bool visFlag, bool considerExit)
{
  // Return to onscreen rendering for now
  //cout << "Render to screen" << endl;
  glBindFramebuffer(GL_FRAMEBUFFER, 0);
  CHECK_GL_ERROR();
  glBindRenderbuffer(GL_FRAMEBUFFER, 0);
  CHECK_GL_ERROR();
  
  for (int pass = 0; pass < 3; pass++) {      
    //=== scene objects rendering ===================================
    // prepare 2D texture
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    CHECK_GL_ERROR();
    //CHECK_GL_ERROR();
    glViewport(0, 0, win_fbo_width, win_fbo_height);
    CHECK_GL_ERROR();
    initProjectionMatrix(win_fbo_width, win_fbo_height);
    CHECK_GL_ERROR();
    if (visFlag) {
      glDisable(GL_TEXTURE_2D);
      glEnable(GL_TEXTURE_2D);    
      glBindTexture(GL_TEXTURE_2D, BTFidTexture[1]); // visualization
      setTexParams();
      glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA32F, BTFrows, BTFcols,
		   0, GL_RGBA, GL_FLOAT, BTFimageB[1]);
      CHECK_GL_ERROR();
    }
    else {
      glDisable(GL_TEXTURE_2D);
      glEnable(GL_TEXTURE_2D);
      glBindTexture(GL_TEXTURE_2D, BTFidTexture[0]); // standard texture
      setTexParams();
      glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA32F, BTFrows, BTFcols,
		   0, GL_RGBA, GL_FLOAT, BTFimageF);
      CHECK_GL_ERROR();
    }

    //cout << "DISPLAY" << endl << flush;

    // Draw the mesh with texture
    //=== scene objects rendering ===================================
    glEnable(GL_DEPTH_TEST);
    CHECK_GL_ERROR();
    glFrontFace(GL_CCW); // GL_CW and GL_CCW
    CHECK_GL_ERROR();
    glDisable(GL_CULL_FACE);
    //glCullFace(GL_FRONT);
    
    // Move the model somewhere by model matrix
    //mesh[actualOBJ].detachTextureData();
    if (visFlag)
      mesh[actualOBJ].setTextureData(BTFimageB[1], BTFcols, BTFrows);
    else
      mesh[actualOBJ].setTextureData(BTFimageF, BTFcols, BTFrows);
    //mesh[actualOBJ].setTextureData(0, BTFcols, BTFrows);
    mesh[actualOBJ].draw(PVMmatrix);
    
    // render also environment map
    if (IsEnvMapLighting()) {
      glDisable(GL_CULL_FACE);
      glCullFace(GL_BACK);
      renderEnvMapOpenGL();
      // Possibly the position of lights for environment map
      renderEnvMapLights();
      CHECK_GL_ERROR();
    }
        
    if ((grabbingSequence)&&(pass==2)&&(saveFlag)) {
      // only in the second pass
      bool useOrigData = false;
      // saves the image and exits
      bool origData = false;
      // grabBuffer(int screenShotNumber, int useOrigData, bool visflag, int pass, int bufferFlag)
      grabBuffer(screenShotNumber++, useOrigData, visFlag, -1, false); // From screen buffer
      
      printf("  Frame: %04d saved", screenShotNumber-1);fflush(stdout);
      // finish rendering
      if (screenShotNumber == NoOfScreenShots) {
	printf(" Video sequence rendering of %d frames finished.\n",NoOfScreenShots);
	deallocations();
	exit(0);
      }

      // scene change for the next frame
      if (rotateObject && (pass == 2)) {
	OBJrot[actualOBJ][2] += 2.f*M_PI/(float)phiStepsPerCircle;   // sphere rotation
	if (OBJrot[actualOBJ][2] >= 2*M_PI)
	  OBJrot[actualOBJ][2] -= 2*M_PI;      
      }
      if (rotateEnvironment) {
	envPhiShift += 2.f*M_PI/(float)phiStepsPerCircle;   // environment map shift
	if (envPhiShift >= 2*M_PI)
	  envPhiShift -= 2*M_PI;      
	
	rotateEnvironmentLights(2.f*M_PI/(float)phiStepsPerCircle); // corresponding env. lights shift
      }
    }

    if (saveAndExit && (pass==2) && (saveFlag)) {
      // saves the image and exits
      bool useOrigData = false;
      // grabBuffer(int screenShotNumber, int useOrigData, bool visflag, int pass, int bufferFlag)
      grabBuffer(screenShotNumber++, useOrigData, visFlag, -1, false); // From screen buffe
      if ((indexRenderedImage == cntFinishRendering)&&(considerExit)) {
	glutLeaveMainLoop();
	glutExit();
	glutDestroyWindow(windowID);
	deallocations();
	exit(0);
      }
    }
    glFlush();
    glutSwapBuffers();
    glutPostRedisplay();
  } // for pass
    
  glBindTexture(GL_TEXTURE_2D, 0);
  return;  
}//--- renderToScreen ------------------------------------------------

void
copyImage(float *dest, float *src)
{
  for (int i =0; i < 2048*2048*4; i++) {
    dest[i] = src[i];
  }
}
void
multiplyImage(float *dest, float value)
{
  // cout << "multiplyying by " << value << endl;
  for (int i =0; i < 2048*2048*4; i++) {
    dest[i] *= value;
  }
}
void
makeGray(float *src)
{
  for (int i =0; i < 2048*2048*4; i += 4) {
    float lum = luminance(src+i);  // compute luminance
    src[i+0] = src[i+1] = src[i+2] = lum; // set it to all 3 channels
    src[i+3] = 0.f; // alpha channel
  }
} 

// Main rendering function for the whole image
void 
renderScene() 
{
  // For testing
  if ((maxModeCompute) && (currModeComputeIndex > -1)) {
    modeCompute = modeComputeArray[currModeComputeIndex];
    SetRandomSeed(1); // reinitializing Mersenne Twister to some seed to make it reproducible    
    currModeComputeIndex++; // for the next image
    if (modeCompute == -1) { // -1 is the stop mark
      deallocations();
      exit(0);
    }
  }

  // Move the model somewhere by model matrix
  //  glm::mat4 modelMatrix = glm::translate(glm::mat4(1.0f), glm::vec3(0.0f, 0.0f, 1.5f));
  // Move the model somewhere by model matrix
  glm::mat4 modelMatrix = glm::translate(glm::mat4(1.0f),
					 glm::vec3(OBJtrans[actualOBJ][0],
						   OBJtrans[actualOBJ][1],
						   OBJtrans[actualOBJ][2]));
  modelMatrix = glm::rotate(modelMatrix, (float)(OBJrot[actualOBJ][2] * 180.f/M_PI),
			    glm::vec3(0, 0, 1)); // rotation matrix
  modelMatrix = glm::rotate(modelMatrix, (float)(OBJrot[actualOBJ][1] * 180.f/M_PI),
			    glm::vec3(0, 1, 0)); // rotation matrix
  modelMatrix = glm::rotate(modelMatrix, (float)(OBJrot[actualOBJ][0] * 180.f/M_PI),
			    glm::vec3(1, 0, 0)); // rotation matrix	       
  glm::mat4 PVMmatrix = projectionMatrix * viewMatrix * modelMatrix;
  
  // without visualization render the texture and prepare possibly the visualization, without OpenGL
  renderOBJECT(modelMatrix, DATA[actualOBJ], indexRenderedImage);
  
  if (modeCompute == 0) { // interactive rendering
    renderSceneSimple(PVMmatrix, false, false, false);
    //cout << "A" << endl;
    return;
  }

  // save BTF atlas texture with the data computed to a file
  if (saveImageBTFatlas) {
    // only once for all rendered images
    printAtlas(BTFimageB[1], indexImageBTFatlas, true, -1); // visualization
    printAtlas(BTFimageF, indexImageBTFatlas, false, -1); // texture rendered
    if (referenceImageAvailable) {
      assert(referenceImage);
      CHDRImage *imgAtlas = new CHDRImage();
      // This is the rendered image created below
      CopyImageToHDRClass(imgAtlas, BTFimageF, false); // no visualization
      // Reference HDR atlas is if file 'referenceImage'
      static int counter = 0;
      counter++;
      if (counter < 2)
	referenceImage->CompareTo(*imgAtlas, false); // this prints RMSE only
      else
	referenceImage->CompareTo(*imgAtlas, true); // this prints RMSE only
      CHDRImage * diffImage = referenceImage->Substruct(*imgAtlas, 0, false, false); // this creates difference image
      diffImage->saveHDR(diffImageFilename);
       // Visualization
      CHDRImage *diffvisImage = new CHDRImage();; // it is populated below
      CopyImageToHDRClass(diffvisImage, BTFimageF, true); // with visualization bargraph
      // and this populates it with color difference
      //cout << "1 Image diffimagevis size = " << diffvisImage->getWidth() << " " << diffvisImage->getHeight() << endl;
      diffImage->setsPseudocolor(*diffvisImage, pseudoColorTechnique, 0, upperBoundVisualization);
      diffvisImage->saveHDR(diffvisImageFilename); // saving to a file
      
      //cout << "Image difffimage size = " << diffImage->getWidth() << " " << diffImage->getHeight() << endl;
      //cout << "2 Image diffimagevis size = " << diffvisImage->getWidth() << " " << diffvisImage->getHeight() << endl;
      delete diffImage;
      delete diffvisImage;
      delete imgAtlas;
    } // reference image for atlas was available and we made comparision
    indexImageBTFatlas++;
  }
  
  bool saveFlag = true;
  bool visFlag = true; // or true - if we want to watch visualization
  bool considerExit = false;
  // First we save standard rendering
  // parameters: (saveFlag, bool visFlag)
  if (visFlag) {
    // We first possibly save visualization
    //cout << "V1 " << endl << flush;
    currTexID = 0; // this is reserved for this purpose
    renderToFrameBuffer(PVMmatrix, true, saveFlag, true, considerExit);
    if (saveFlag) // visualization file with the same index
      screenShotNumber--;
    visFlag = false;
  }
  // Then we save rendering itself - this always increases number
  if (visModeVariable > 0)
    cout << "Avoiding rendering due to saving variable index = " << visModeVariable << endl;
  else {
    renderToFrameBuffer(PVMmatrix, true, saveFlag, false, considerExit);
    
    if (saveAlsoNormalizedAlbedo) {
      // parameters: (saveFlag, bool visFlag)
      // We multiply the texture itself by 1/albedo
      //    cout << "V2 " << endl << flush;
      copyImage(BTFimageB[2], BTFimageF);
      currTexID = 2; // this is reserved for this purpose
      multiplyImage(BTFimageB[2], normalizeTotalAlbedoCoeff);
      suffixAlbedoNormalized = true;
      screenShotNumber--;
      currTexID = 2; // this is reserved for this purpose
      renderToFrameBuffer(PVMmatrix, true, saveFlag, false, considerExit);
      if (saveAlsoGrayToneImage) {
	// both normalized albedo and saving gray
	//cout << "V3 " << endl << flush;
	copyImage(BTFimageB[3], BTFimageB[2]);
	makeGray(BTFimageB[3]);
	suffixGrayTone = true;
	screenShotNumber--;
	currTexID = 3; // this is reserved for this purpose
	renderToFrameBuffer(PVMmatrix, true, saveFlag, false, considerExit);    
	suffixGrayTone = false;      
      }
      suffixAlbedoNormalized = false;
      currTexID = 0; // this is reserved for this purpose
      glBindTexture(GL_TEXTURE_2D, 0);      
    }
    if (saveAlsoGrayToneImage) {
      // parameters: (saveFlag, bool visFlag)
      // We multiply the texture itself by 1/albedo
      //    cout << "V4 " << endl << flush;
      copyImage(BTFimageB[4], BTFimageF); // make backup
      makeGray(BTFimageB[4]);
      suffixGrayTone = true;
      screenShotNumber--;
      currTexID = 4; // this is reserved for this purpose
      renderToFrameBuffer(PVMmatrix, true, saveFlag, false, considerExit);    
      suffixGrayTone = false;
      currTexID = 0; // this is reserved for this purpose
      glBindTexture(GL_TEXTURE_2D, 0);      
    }
    saveFlag = false; // not saving when rendering to the screen
  }

  bool showRealImage = true;
  if (showRealImage) {
    // at the end of rendering show the image itself, possibly save visualization if not done in framebuffer
    if (visFlag) {
      // save possibly visualization if not done by framebuffer
      renderToScreen(PVMmatrix, saveFlag, true, considerExit);
      if (saveFlag) // visualization file with the same index
	screenShotNumber--;
    }
    // Only to screen - bool saveFlag, bool visFlag
    renderToScreen(PVMmatrix, saveFlag, false, considerExit); // image itself
  }
  else { // show only visualization at the end to the screen!!!!
    // save image if not done in frame buffer
    renderToScreen(PVMmatrix, saveFlag, false, considerExit);
    if (saveFlag) // visualization file with the same index
      screenShotNumber--;
    // visualization of the function shown on screen
    renderToScreen(PVMmatrix, saveFlag, true, considerExit); 
  }

  indexRenderedImage++;
  if (saveAndExit && (indexRenderedImage==2)) {
    //cout << "A1" << endl << flush;
    glutLeaveMainLoop();
    //cout << "C1" << endl << flush;
    glutDestroyWindow(windowID);
    //  cout << "D1" << endl << flush;
    deallocations();
    return;
  }
  
  return;  
}//--- renderScene ------------------------------------------------

void
printHelp()
{
  printf(" ________________________________________________________________________\n"
	 " |HELP on individual keys:                                              |\n"
	 " |Environment lighting mode:                                            |\n"
	 " |   e|E - loop between available environment maps(i.e.lighting schemes)|\n"
	 " |   p|P|{|} - set|max|set,increase,decrease number of environment lights|\n"
	 " |One light mode:                                                       |\n"
	 " |   d - reinit light position                                          |\n"
	 " |   c - print camera translation and rotation                          |\n"
	 " |   v - use light backgrond                                            |\n"
	 " |   b - use dark backgroud                                             |\n"
	 " |   l - toggle render lights                                           |\n"
	 " |   k - toggle render environment map (debugging)                      |\n"
	 " |General commands:                                                     |\n"
	 " |   0,1,2,3 - change rendering mode                                    |\n"
	 " |   q|w - rotate viewer left | right                                   |\n"
	 " |   Q|W - rotate viewer to upwards | downwards                         |\n"
	 " |   a|s - rotate object left | right (yaw)                             |\n"
	 " |   A|S - rotate object (roll)                                         |\n"
	 " |   z|x - rotate environment map to left |right                        |\n"
	 " |   Z|X - rotate object (pitch)                                        |\n"
	 " |   i|I - loop between available BTF materials                         |\n"
	 " |   o|O - loop between available OBJ 3D models (forward|backward)      |\n"
	 " |   r - reinit camera position                                         |\n"
	 " |   [|] - decrease/increase texture size twice                         |\n"
	 " |   +|- - zoom in, zoom out                                            |\n"
	 " |   n - draw vertex normals, tangents and binormals                    |\n"
	 " |   u - toggle render axes                                             |\n"
	 " |   y - save texture mapping atlas                                     |\n"
	 " |   f - print OpenGL version and available extensions                  |\n"
	 " |   g - grab a screen-shot                                             |\n"
	 " |   j - toggle between fill anf wire-frame display of polygons         |\n"
	 " |   h - this help                                                      |\n"
	 " |  Esc - quit the renderer                                             |\n"
	 " |______________________________________________________________________|\n\n");
  return;
}//--- printHelp -------------------------------------------------------

void
RestoreSingleLightPosition()
{
  savedSingleLightPositions = false;

  // restore position of the light
  for (int i = 0; i < 4; i++)
    lpos[i] = lposBackup[i];
  printf("Restoring single light : x: %7.2f y: %7.2f z: %7.2f\n",
	 lpos[0],lpos[1],lpos[2]);
  printf("   One light mode:             "
	 "                                                        ");
}

void
SaveSingleLightPosition(int mode)
{
  if (savedSingleLightPositions)
    return;
  savedSingleLightPositions = true;
  // from a single point light to environment lights

  // save position of a single light
  for (int i = 0; i < 4; i++)
    lposBackup[i] = lpos[i];
  // Set the environment map
  cntSamplesPerPixel = defaultCntSamplesPerPixel;
	 
  printf("Saving single light : x: %7.2f y: %7.2f z: %7.2f\n",
	 lpos[0], lpos[1], lpos[2]);
  printf("   Environment light mode: envmap=%d lights=%d               "
	 "                                                       ",
	 actualEnvMap, cntSamplesPerPixel);fflush(stdout);
  // makeAlmostWhiteEnvColor(actualEnvMap, almostWhiteWeightColorCoeff[actualEnvMap]);
  return;
}

void 
processKeys(unsigned char key, int x, int y) 
{
  if (disableInteractivity)
    // switch off processing keys
    return;
 
   switch (key) {
   case 27:         // finish the renderer
     printf("\n\n Quiting the renderer\n\n");
     deallocations();
     exit(0);
     break;
   case 'c':        // print camera translation and rotation
     printf("Object: rotation    (hpr): %3.2f %3.2f %3.2f\n",
	    OBJrot[actualOBJ][0], OBJrot[actualOBJ][1], OBJrot[actualOBJ][2]);
     printf("Camera:  translation (xyz): %3.2f %3.2f %3.2f\n",
	    camerapos[0], camerapos[1], camerapos[2]);
     break;
   case 'r':        // reinit camera/environment map position
     // The position and rotation of the object is reset
     viewerDistance = viewerDistanceInit;
     viewerAngle = viewerAngleInit;
     envPhiShift = 0;
     rotateEnvironmentLights(envPhiShift);
     OBJrot[actualOBJ][0] = initOBJrot[actualOBJ][0];
     OBJrot[actualOBJ][1] = initOBJrot[actualOBJ][1];
     OBJrot[actualOBJ][2] = initOBJrot[actualOBJ][2];
     OBJtrans[actualOBJ][0] = initOBJtrans[actualOBJ][0];
     OBJtrans[actualOBJ][1] = initOBJtrans[actualOBJ][1];
     OBJtrans[actualOBJ][2] = initOBJtrans[actualOBJ][2];
     initViewMatrices(win_fbo_width, win_fbo_height);  

     // The position of the viewer
     createCameraposFromObjectPosition();
     if (TA)
       TA->setViewPos(camerapos);
     printf("   Reseting camera position                             ");
     fflush(stdout);
     glutPostRedisplay();
     break;
   case 'd':        // reinit light position
     if (!IsEnvMapLighting()) {
       lpos[0] = lposInit[0];
       lpos[1] = lposInit[1];
       lpos[2] = lposInit[2];
       lpos[3] = lposInit[3];
       if (TA)
	 TA->setIlluPos(lpos);
       printf("   Reseting light position                             ");
       fflush(stdout);
       glutPostRedisplay();
     }
     break;
   case 'e':        // loop between change environment-maps
     if (allowEnvManipulationFlag || IsEnvMapLighting()) {
       actualEnvMap++;
       actualEnvMap %= NoOfEnvMaps;
       rotateEnvironmentLights(envPhiShift);
       createCameraposFromObjectPosition();
       glutPostRedisplay();
       printf("   Environment map toggle: envmap=%d   "
	      "                                                                   ",
	      actualEnvMap);fflush(stdout);
     }
     break;
   case 'E':        // loop between change environment-maps
     if (allowEnvManipulationFlag || IsEnvMapLighting()) {
       actualEnvMap--;
       if (actualEnvMap == -1)
	 actualEnvMap = NoOfEnvMaps-1;
       rotateEnvironmentLights(envPhiShift);
       createCameraposFromObjectPosition();
       glutPostRedisplay();
       printf("   Environment map toggle: envmap=%d                "
	      "                                                      ",
	      actualEnvMap);fflush(stdout);
     }
     break;
   case 'o':        // loop between OBJ models
     actualOBJ++;
     actualOBJ %= NoOfObjects;
     glutPostRedisplay();
     printf("   OBJ model: %s                                     "
	    "                          ", OBJpaths[actualOBJ]);
     fflush(stdout);
     break;
   case 'O':        // loop between OBJ models
     actualOBJ--;
     if (actualOBJ < 0)
       actualOBJ = NoOfObjects-1;
     glutPostRedisplay();
     printf("   OBJ model: %s                                      "
	    "                         ", OBJpaths[actualOBJ]);
     fflush(stdout);
     break;
   case 'f': {       // printout OpenGL Driver version and available extensions
     char *version = (char *)glGetString(GL_VERSION);
     printf("OpenGL version: %s\n",version);
     char *extension = (char *)glGetString(GL_EXTENSIONS);
     printf("OpenGL extensions: %s\n",extension);
     break;
   }
   case 'g': {       // grab a screenshot
     bool gb = false;
     grabBufferHDR(screenShotNumber, gb, false, -1, false); // visualization - from screen buffer
     grabBufferHDR(screenShotNumber, gb, true, -1, false); // visualization - from screen buffer
     screenShotNumber++;
     break;
   }
   case 'j':        // switch to wireframe model
     if (renderWire == 1) { // to line segments
       for (int i = 0; i < NoOfObjects; i++) {
	 mesh[i].setRenderMode(GL_TRIANGLES); // does not work as we would need to change the geometry
	 mesh[i].toggleLinesTriangles(0);
       }
       renderWire = 0;
     }
     else { // back to triangles
       for (int i = 0; i < NoOfObjects; i++) {
	 mesh[i].setRenderMode(GL_LINES);
	 mesh[i].toggleLinesTriangles(1);
       }
       renderWire = 1;
     }
     glutPostRedisplay();
     break;
   case 'n':        // draw needles for normals, tangents, bitangents, direction to light, direction to viewer
     drawNeedles++;
     drawNeedles %= 6;
     switch (drawNeedles) {
     case 0: printf("   drawNeedles=0 .. nothing                                     ");
       break;
     case 1: printf("   drawNeedles=1 .. normal                                      ");
       break;
     case 2: printf("   drawNeedles=2 .. tangent                                     ");
       break;
     case 3: printf("   drawNeedles=3 .. bitangent                                   ");
       break;
     case 4: printf("   drawNeedles=4 .. directon to light                           ");
       break;
     case 5: printf("   drawNeedles=5 .. direction to viewer                         ");
       break;
     }
     glutPostRedisplay();
     break;
   case 'b':       // black background colour
     cout << "Black background color" << endl;
     backgroundColor[0] = 0.1;
     backgroundColor[1] = 0.1;
     backgroundColor[2] = 0.1;
     backgroundColor[3] = 0.0;
     glClearColor(backgroundColor[0], backgroundColor[1],
		  backgroundColor[2], backgroundColor[3]);
     glutPostRedisplay();        
     break;
   case 'v':       // white background colour
     cout << "White background color" << endl;
     backgroundColor[0] = 0.9;
     backgroundColor[1] = 0.9;
     backgroundColor[2] = 0.9;
     backgroundColor[3] = 0.0;
     glClearColor(backgroundColor[0], backgroundColor[1],
		  backgroundColor[2], backgroundColor[3]);
     glutPostRedisplay();
     break;
   case 'l':
     if (renderLights == 0) {
       renderLights = 1; printf("Render lights ON\n");
     }
     else {
       renderLights = 0; printf("Render lights OFF\n");
     }
     fflush(stdout);
     break;     
   case 'u':
     if (renderAxis == 0) {
       renderAxis = 1; printf("Render axis ON\n");
     }
     else {
       renderAxis = 0; printf("Render axis OFF\n");       
     }
     fflush(stdout);
     break;          
   case 'k':
     if (renderEnvMap == 0) {
       renderEnvMap = 1; printf("Render environment map ON\n");
     }
     else {
       renderEnvMap = 0; printf("Render environment map OFF\n");       
     }
     fflush(stdout);
     break;          
   case '+': {      // zoom in
     if (fabs(viewerDistance) > 0.001f) {
       viewerDistance /= 1.1;
       createCameraposFromObjectPosition();
       printf("  Zooming in: viewerDistance = %3.2f\n", viewerDistance);
       fflush(stdout);
       glutPostRedisplay();
     }
     break;
   }
   case '-': {       // zoom out
     if (fabs(viewerDistance) < 2000.f) {
       viewerDistance *= 1.1;
       createCameraposFromObjectPosition();
       printf("  Zooming out: viewerDistance = %3.2f\n", viewerDistance);
       fflush(stdout);
       glutPostRedisplay();
     }
     break;
   }
   case 'q': // rotate viewer left
     if (allowEnvManipulationFlag || IsEnvMapLighting()) {
       OBJrot[actualOBJ][2] += 2.f*M_PI/(float)phiStepsPerCircle;
       if (OBJrot[actualOBJ][2] >= 2.f * M_PI)
	 OBJrot[actualOBJ][2] -= 2.f * M_PI;
       // rotate environment map left
       envPhiShift += 2.f*M_PI/(float)phiStepsPerCircle;
       if (envPhiShift >= 2.f * M_PI)
	 envPhiShift -= 2.f * M_PI;

       rotateEnvironmentLights(envPhiShift);
       printf("q  Envir. Phi Shift: %3.0f,  Lights Phi Shift: %3.0f, Object Phi Shift: %3.0f"
	      "     ", (180.f/M_PI) * envPhiShift, (180.f/M_PI)*lightsPhiShift(), 
	      (180.f/M_PI) * OBJrot[actualOBJ][1]);
       fflush(stdout);
     }
     createCameraposFromObjectPosition();
     glutPostRedisplay();
     break;
   case 'w': // rotate viewer right
     if (allowEnvManipulationFlag || IsEnvMapLighting()) {
       OBJrot[actualOBJ][2] -= 2.f*M_PI/(float)phiStepsPerCircle;
       if (OBJrot[actualOBJ][2] < 0)
	 OBJrot[actualOBJ][2] += 2.f*M_PI;
       // rotate environment map left       
       envPhiShift -= 2.f*M_PI/(float)phiStepsPerCircle;
       if (envPhiShift < 0.f)
	 envPhiShift += 2.f*M_PI;

       rotateEnvironmentLights(envPhiShift);
       printf("w  Envir. Phi Shift: %3.0f,  Lights Phi Shift: %3.0f, Object Phi Shift: %3.0f"
	      "        ", (180.f/M_PI) * envPhiShift, (180.f/M_PI) * lightsPhiShift(), 
	      (180.f/M_PI) * OBJrot[actualOBJ][1]);
       fflush(stdout);
     }
     createCameraposFromObjectPosition();
     glutPostRedisplay();
     break;
   case 'Q': // rotate viewer upwards
     viewerAngle += 2.f*M_PI/(float)phiStepsPerCircle;
     if (viewerAngle >= 2.f*M_PI) viewerAngle -= 2.f*M_PI;
     printf("   Object Phi Shift: %3.0f                               "
	    "                              ",
	    (180.f/M_PI) * viewerAngle);fflush(stdout);
     createCameraposFromObjectPosition();
     glutPostRedisplay();
     break;
   case 'W': // rotate viewer downwards
     viewerAngle -= 2.f*M_PI/(float)phiStepsPerCircle;
     if (viewerAngle < 0) viewerAngle += 2.f*M_PI;
     printf("   Viewer Phi Shift: %3.0f                               "
 	    "                              ",
	    (180.f/M_PI) * viewerAngle);fflush(stdout);
     createCameraposFromObjectPosition();
     glutPostRedisplay();
     break;
   case 'a':         // rotate object left
     OBJrot[actualOBJ][2] += 2.f*M_PI/(float)phiStepsPerCircle;
     if (OBJrot[actualOBJ][2] > 2.f*M_PI)
       OBJrot[actualOBJ][2] -= 2.f*M_PI;
     printf("   Object Phi Shift: %3.0f                            "
	    "                                 ",
	    (180.f/M_PI)*OBJrot[actualOBJ][2]);fflush(stdout);
     createCameraposFromObjectPosition();
     glutPostRedisplay();
     break;     
   case 's':         // rotate object right
     OBJrot[actualOBJ][2] -= 2.f*M_PI/(float)phiStepsPerCircle;
     if (OBJrot[actualOBJ][2] < 0)
       OBJrot[actualOBJ][2] += 2.f*M_PI;
     printf("   Object Phi Shift: %3.0f                           "
	    "                                  ",
	    (180.f/M_PI)*OBJrot[actualOBJ][2]);fflush(stdout);
     createCameraposFromObjectPosition();
     glutPostRedisplay();
     break;
   case 'A':         // rotate object upwards
     OBJrot[actualOBJ][0] += 2.f*M_PI/(float)phiStepsPerCircle;
     if (OBJrot[actualOBJ][0] > 2.f*M_PI)
       OBJrot[actualOBJ][0] -= 2.f*M_PI;
     printf("   Object Y-theta shift: %3.0f                            "
	    "                                 ",
	    (180.f/M_PI)*OBJrot[actualOBJ][0]);fflush(stdout);
     createCameraposFromObjectPosition();
     glutPostRedisplay();
     break;     
   case 'S':         // rotate object downwards
     OBJrot[actualOBJ][0] -= 2.f*M_PI/(float)phiStepsPerCircle;
     if (OBJrot[actualOBJ][0] < 0)
       OBJrot[actualOBJ][0] += 2.f*M_PI;
     printf("   Object Y-theta shift: %3.0f                           "
	    "                                  ",
	    (180.f/M_PI)*OBJrot[actualOBJ][0]);fflush(stdout);
     createCameraposFromObjectPosition();
     glutPostRedisplay();
     break;
   case 'z': // rotate environment map left
     if (allowEnvManipulationFlag) {
       envPhiShift += 2.f*M_PI/(float)phiStepsPerCircle;
       if (envPhiShift >= 2.f*M_PI)
	 envPhiShift -= 2.f*M_PI;

       rotateEnvironmentLights(envPhiShift);
       printf("   Envir. Phi Shift: %3.0f,  Lights Phi Shift: %3.0f             "
	      "                    ",
	      (180.f/M_PI) * envPhiShift,
	      (180.f/M_PI) * lightsPhiShift()); fflush(stdout);
     }
     createCameraposFromObjectPosition();
     glutPostRedisplay();
     break;
   case 'x': // rotate environment map right
     if (allowEnvManipulationFlag) {
       envPhiShift -= 2.f*M_PI/(float)phiStepsPerCircle;
       if (envPhiShift <= 0.f)
	 envPhiShift += 2.f*M_PI;
       
       rotateEnvironmentLights(envPhiShift);
       printf("   Envir. Phi Shift: %3.0f,  Lights Phi Shift: %3.0f      "
	      "                           ",
	      (180.f/M_PI) * envPhiShift,
	      (180.f/M_PI) * lightsPhiShift());fflush(stdout);
     }
     createCameraposFromObjectPosition();
     glutPostRedisplay();
     break;
   case 'Z':         // rotate object clockwise
     OBJrot[actualOBJ][1] += 2.f*M_PI/(float)phiStepsPerCircle;
     if (OBJrot[actualOBJ][1] > 2.f*M_PI)
       OBJrot[actualOBJ][1] -= 2.f*M_PI;
     printf("   Object X-angle shift: %3.0f                            "
	    "                                 ",
	    (180.f/M_PI)*OBJrot[actualOBJ][1]);fflush(stdout);
     createCameraposFromObjectPosition();
     glutPostRedisplay();
     break;     
   case 'X':         // rotate object anti-clockwise
     OBJrot[actualOBJ][1] -= 2.f*M_PI/(float)phiStepsPerCircle;
     if (OBJrot[actualOBJ][1] < 0)
       OBJrot[actualOBJ][1] += 2.f*M_PI;
     printf("   Object X-angle shift: %3.0f                           "
	    "                                  ",
	    (180.f/M_PI)*OBJrot[actualOBJ][1]);fflush(stdout);
     createCameraposFromObjectPosition();
     glutPostRedisplay();
     break;
   case 'p': // set number of light sources for environment lighting to default
     if (IsEnvMapLighting()) {
       cntSamplesPerPixel = defaultCntSamplesPerPixel;
       printf("   Envmap #lights set to %d                        ",
	      cntSamplesPerPixel);
       fflush(stdout);
       glutPostRedisplay();
     }
     break;       
   case 'P': // set the number of light sources for environment lighting to max
     if (IsEnvMapLighting()) {
       cntSamplesPerPixel = maxCntSamplesPerPixel;
       printf("   Envmap #lights set to %d                        ",
	      cntSamplesPerPixel);
       fflush(stdout);
       glutPostRedisplay();
     }
     break;
   case '{': { // decrease number of light sources for environment lighting
     //if (IsEnvMapLighting()) {
       cntSamplesPerPixel = (int)((float)cntSamplesPerPixel / 1.4f);
       if (cntSamplesPerPixel < 2) cntSamplesPerPixel = 2; 
       printf("   Envmap #lights decreased to %d                        ",
	      cntSamplesPerPixel);
       fflush(stdout);
       glutPostRedisplay();
       //}
     }
     break;
   case '}': {         // increase number of light sources for environment lighting
     //if (IsEnvMapLighting()) {
       int oldNumber = cntSamplesPerPixel;
       cntSamplesPerPixel = (int)((float)cntSamplesPerPixel * 1.3f);
       if (oldNumber == cntSamplesPerPixel)
	 cntSamplesPerPixel += 2;	  
       if (cntSamplesPerPixel > maxCntSamplesPerPixel)
	 cntSamplesPerPixel = maxCntSamplesPerPixel;
       printf("   Envmap #lights increased to %d                        ",
	      cntSamplesPerPixel);
       fflush(stdout);
       glutPostRedisplay();
     //}
     }
     break; 	
   case '[':         // decrease BTF texture size
     if (Npow>6) {
       Npow--;
       BTFrows = (int)pow(2, Npow);
       BTFcols = (int)pow(2, Npow);
       printf("   Texture size : [%d x %d]                                                     ",
	      BTFrows,BTFcols);
       fflush(stdout);
       glutPostRedisplay();
     }
     break;
   case ']':         // increase BTF texture size
     if (Npow<11) {
       Npow++;
       BTFrows = (int)pow(2, Npow);
       BTFcols = (int)pow(2, Npow);
       printf("   Texture size : [%d x %d]                                                     ",
	      BTFrows,BTFcols);
       fflush(stdout);
       glutPostRedisplay();
     }
     break;
   case 'y': {
     // For debugging only, saving atlas texture to HDR file
     indexImageBTFatlas++;
     printAtlas(BTFimageB[0], indexImageBTFatlas, true, 0); // visualization
     printAtlas(BTFimageF, indexImageBTFatlas, false, 0);
     saveImageBTFatlas = true;
     break;
   }	
   case 'h':         // help print out
     printHelp();
     break;
   case '0':
     modeCompute = 0;
     SetRandomSeed(1); // reinitializing Mersenne Twister to some seed to make it reproducible    
     RestoreSingleLightPosition();     
     printf("Setting modeCompute = 0 ... single point light\n"); fflush(stdout);
     glutPostRedisplay();
     break;
   case '1': {
     modeCompute = 1;
     SetRandomSeed(1); // reinitializing Mersenne Twister to some seed to make it reproducible    
     SaveSingleLightPosition(modeCompute);
     printf("Setting modeCompute = 1 ... deterministic samples from EM\n");fflush(stdout);
     glutPostRedisplay();
     break;
   }
   case '2':
     modeCompute = 2;
     SetRandomSeed(1); // reinitializing Mersenne Twister to some seed to make it reproducible    
     SaveSingleLightPosition(modeCompute);
     printf("Setting modeCompute = 2 ... random samples from EM\n");fflush(stdout);
     glutPostRedisplay();
     break;
   case '3':
     modeCompute = 3;
     SetRandomSeed(1); // reinitializing Mersenne Twister to some seed to make it reproducible    
     SaveSingleLightPosition(modeCompute);
     printf("Setting modeCompute = 3 ... sampling according to BRDF*cos(theta)\n");
     fflush(stdout);
     glutPostRedisplay();
     break;     
   case '4':
     modeCompute = 4;
     SetRandomSeed(1); // reinitializing Mersenne Twister to some seed to make it reproducible    
     SaveSingleLightPosition(modeCompute);
     printf("Setting modeCompute = 4 ... MIS from BRDF*cos(theta) a EM\n");fflush(stdout);
     glutPostRedisplay();
     break;
   case '5':
     modeCompute = 5;
     SetRandomSeed(1); // reinitializing Mersenne Twister to some seed to make it reproducible    
     SaveSingleLightPosition(modeCompute);
     printf("Setting modeCompute = 5 ... MIS from BRDF*cos(theta) a EM\n");fflush(stdout);
     glutPostRedisplay();
     break;     
   default: {
     printf("No function implemented for %c",key);fflush(stdout);
   }
   } // switch key

   //  freebmem1(buffer, 0,4*win_fbo_width*win_fbo_height-1);
   return;
}//--- processKeys ------------------------------------------------

void 
processMouse(int button, int state, int winx, int winy)
{
  mb[0] = mb[1] = mb[2];
  if (button == GLUT_LEFT_BUTTON && state == GLUT_DOWN)
    mb[0] = 1;
  if (button == GLUT_LEFT_BUTTON && state == GLUT_UP)
    mb[0] = 0;

  if (button == GLUT_MIDDLE_BUTTON && state == GLUT_DOWN)
    mb[1] = 1;
  if (button == GLUT_MIDDLE_BUTTON && state == GLUT_UP)
    mb[1] = 0;

  if (button == GLUT_RIGHT_BUTTON && state == GLUT_DOWN)
    mb[2] = 1;
  if (button == GLUT_RIGHT_BUTTON && state == GLUT_UP)
    mb[2] = 0;


  // redraw the scene if any mouse-button was pressed
  if (mb[0]==0 && mb[1]==0 && mb[2]==0) {
    glutSwapBuffers();
    glutPostRedisplay();
  }
  oldwinx = winx;
  oldwiny = winy;
  return;
}//--- processMouse -----------------------------------------------------

void 
motion(int winx, int winy)
{
  int i;
  float rotmat[16];
  float rotmatinv[16];
  //cout << "mb = " << mb[0] << " " << mb[1] << endl;

  if (mb[0]) { // rotate object or light
    // move a point light if no environment map
    if (!IsEnvMapLighting()) {
      //cout << "1 Light rotate" << endl;
      lpos[1] += 2.f* (winx - oldwinx);
      lpos[2] += -2.f* (winy - oldwiny);
      TA->setIlluPos(lpos);
    }
  }
  if (mb[1]) {
    // rotate object
    //cout << "1 Object rotate" << endl;
    OBJrot[actualOBJ][1] += float(winx - oldwinx) * 0.002f;
    OBJrot[actualOBJ][0] += float(winy - oldwiny) * 0.002f;
  }
  if (mb[2]) {
    // rotate object
    //cout << "2 Object rotate" << endl;
    OBJrot[actualOBJ][2] += float(winx - oldwinx) * 0.002f;
    OBJrot[actualOBJ][0] += float(winy - oldwiny) * 0.002f;
  }
  
  oldwinx = winx;
  oldwiny = winy;

  return;
}//--- motion --------------------------------------------------------------------

void
GenerateOPENGLtextures()
{
  char tt[200];
  // Generate OpenGL textures
  for(int em=0;em<NoOfEnvMaps;em++) {    
    // allocate GPU textures
    bool emHDRtexture = true; // true ... HDR texture, false... LDR texture 8-bit
    if (!environMap[em].getTypeHDR())
      emHDRtexture = false; // LDR texture, as input image was also 8-bit
    cout << "EM: width = " << envImageCols[em] << " height = " << envImageRows[em] << endl; 
    // allocation of PNG env. image
    //  corresponding 2^n texture for OpenGL + copy data from PNG image in BYTE format
    float *environmentGLtexture = new float[4*envCols[em]*envRows[em]]; 
    assert(environmentGLtexture);    
    // copy the data to environmentGLtexture[em][...]
    int index = 0;
    float RGB[3];
    int ww = environMap[em].getWidth();
    float multIntensityGL = 1.0f;

    // for visualization of the background by picture of environment map
    // HDR and LDR
    if (emHDRtexture) {
      // HDR content will be used during rendering
      multIntensityGL = 10.f/environMap[em].getTotalPower();
    }
    else {
      // LDR content
      if (environMap[em].getTypeHDR()) // LDR content from HDR input IMAGE
	multIntensityGL = 255.f/environMap[em].getTotalPower();
      else
	multIntensityGL = 0.001f;// LDR content from LDR image
    }
      
    // Fill in the texture to be used by OpenGL
    for (int irow=0;irow<envRows[em];irow++) {
      for (int jcol=0;jcol<envCols[em];jcol++) {
	if (irow < envImageRows[em]) {
	  if (jcol < envImageCols[em]) {
	    environMap[em].getRGB_XY(jcol, irow, RGB);
	    for (int isp = 0; isp < 3; isp++) {
	      float val = RGB[isp]*multIntensityGL;
	      if ((!emHDRtexture) && (val > 255.f)) val = 255.f; // possibly clipping for LDR
		environmentGLtexture[index+isp] = val;
	    }
	    index += 3;
	  }
	  else {
	    if ( (jcol >= envImageCols[em]) &&
		 (jcol < envImageCols[em]+5) ) { 
	      environMap[em].getRGB_XY(envImageCols[em]-1, irow, RGB);
	      for (int isp = 0; isp < 3; isp++) {
		float val = RGB[isp]*multIntensityGL;
		if ((!emHDRtexture) && (val > 255.f)) val = 255.f; // possibly clipping for LDR
		environmentGLtexture[index+isp] = val;
	      }
	      index += 3;
	    }
	    else {
	      if ( (jcol > envCols[em]-5) &&
		   (jcol < envCols[em]) ) {
		environMap[em].getRGB_XY(0, irow, RGB);
		for (int isp = 0; isp < 3; isp++) {
		  float val = RGB[isp]*multIntensityGL;
		  if ((!emHDRtexture)&&(val > 255.f)) val = 255.f; // possibly clipping for LDR
		  environmentGLtexture[index+isp] = val;
		}
		index += 3;
	      }
	      else { // zero value to all channels		
		environmentGLtexture[index++] = 0.f;
		environmentGLtexture[index++] = 0.f;
		environmentGLtexture[index++] = 0.f;
	      }
	    }
	  }
        }
	// setting alpha channel in texture
	environmentGLtexture[index++] = 0.f;	    
      } // for jcol
    } // for irow
    // radius, number of triangles, how to draw
    meshenvmap[em].init(environMapRadius, 100, GL_TRIANGLES); 
    meshenvmap[em].setTextureData(environmentGLtexture, envImageCols[em], envImageRows[em]);
    // debugging only
    if (0) {
      sprintf(tt,"0a_%d.hdr", em);
      environMap[em].saveDebugHDR(tt, 0, 1.0);
      sprintf(tt,"0b_%d.hdr", em);
      environMap[em].saveDebugHDR(tt, 1, 1.0);
      sprintf(tt,"0c_%d.hdr", em);
      environMap[em].saveDebugHDR(tt, 2, 1.0);
    }
    //debugImage.SaveToFile("dd.tga");
    // env. images + env. textures deallocation
    delete []environmentGLtexture;
  } // for em    
}

void
SetEnvironmentMapsNames()
{
  if (NoOfEnvMaps == 0)
    return;
#if 1
  if (semConstSouthHemEMflag) {
    cout << "SETTING BLACK REFERENCE IMAGES AS WELL" << endl;
  }

  // Hack to avoid crash and specify the ENV maps correctly by index, only one is possible to work
  int indexhack = 2;
  if (envMapToReadHACK == 0) {
    strcpy(envMapsFilename[indexhack],"ENV/cityhall_land.hdr");
    if (readReferenceImageFlag) {
      if (semConstSouthHemEMflag)
        strcpy(referenceImageFilename, "refs/bcityhallatlasref.hdr");
      else
        strcpy(referenceImageFilename, "refs/cityhallatlasref.hdr");      
    }
    cout << "Hack = ENV/cityhall_land.hdr" << endl;
  }
  if (envMapToReadHACK == 1) {
    strcpy(envMapsFilename[indexhack],"ENV/gracerot140.hdr"); //rotated by 140 degrees
    if (readReferenceImageFlag) {
      if (semConstSouthHemEMflag)
        strcpy(referenceImageFilename, "refs/bgraceatlasref.hdr");
      else
        strcpy(referenceImageFilename, "refs/graceatlasref.hdr");
    }
    cout << "Hack = ENV/gracerot140.hdr" << endl;
  }
  if (envMapToReadHACK == 2) {
    strcpy(envMapsFilename[indexhack],"ENV/stpetersrot85.hdr"); //rotated by 85 degrees
    if (readReferenceImageFlag) {
      if (semConstSouthHemEMflag)
        strcpy(referenceImageFilename, "refs/bstpetersatlasref.hdr");
      else
        strcpy(referenceImageFilename, "refs/stpetersatlasref.hdr");      
    }
    cout << "Hack = ENV/stpetersrot85.hdr" << endl;
  }
  if ((envMapToReadHACK >= NoOfEnvMapsToReadHACK)||(envMapToReadHACK < 0)) {
    cout << "Imposible value for " << envMapToReadHACK << " .. exiting" << endl;
    abort();
  }
#else
    strcpy(envMapsFilename[indexhack],"ENV/stpetersrot85.hdr"); //rotated by 85 degrees
#endif
  //------------------------------------------------------------------------------
  //strcpy(envMapsFilename[0],"ENV/cafe_lalo.hdr");
  //strcpy(envMapsFilename[0],"ENV/stpeters.hdr"); //rotated by 85 degrees
  //strcpy(envMapsFilename[0],"ENV/dtest.png");
  strcpy(envMapsFilename[0],"ENV/cafe_lalo.hdr");
  //strcpy(envMapsFilename[0],"ENV/btest.png");
  //strcpy(envMapsFilename[0],"ENV/salon_land.hdr");
  strcpy(envMapsFilename[1],"ENV/studio_land.hdr");
  //strcpy(envMapsFilename[1],"ENV/stpeters.hdr");
  //strcpy(envMapsFilename[1],"ENV/cityhall_land.hdr");
  //strcpy(envMapsFilename[2],"ENV/stpeters.hdr");
  //strcpy(envMapsFilename[2],"ENV/grace.hdr"); //rotated by 140 degrees
  //strcpy(envMapsFilename[0],"ENV/grace110.hdr");
  //strcpy(envMapsFilename[0],"ENV/ennis.hdr");
  //strcpy(envMapsFilename[0],"ENV/doge2.hdr");
  //strcpy(envMapsFilename[0],"ENV/grassplain.png");
  //strcpy(envMapsFilename[0],"ENV/stpeters.hdr");
  //strcpy(envMapsFilename[1],"ENV/grace.hdr");
  //strcpy(envMapsFilename[2],"ENV/ennis.hdr");
  //strcpy(envMapsFilename[1],"ENV/grassplain.png");
  
  cout << "index=2 filename = " << envMapsFilename[2] << endl;
  return;
}

void
LoadEnvironmentMaps(int supportOpenGL = 1)
{
  if (NoOfEnvMaps == 0)
    return;
  //=== reading environment image =================================
  printf("ENVIRONMENT MAPS PROCESSING\n");

  // allocate corresponding number of texture handlers
  envImageRows = iallocation1(0, NoOfEnvMaps);
  envImageCols = iallocation1(0, NoOfEnvMaps);
  envRows = iallocation1(0, NoOfEnvMaps);
  envCols = iallocation1(0, NoOfEnvMaps);
  environMap = new CEnvironmentMap[NoOfEnvMaps];
  int envPlanes[NoOfEnvMaps];
  bool readStatus;

  // reads sizes of range maps
  for(int em=0;em<NoOfEnvMaps;em++) {
    std::string fileextension;
    ExtractFilenameExtension(std::string(envMapsFilename[em]), fileextension);
    if (fileextension.compare("png") == 0) {
      // Read PNG file as environment map
      readStatus = environMap[em].loadPNG((const char*)(envMapsFilename[em]));	
      if (!readStatus) {
	printf("ERROR: environment map file '%s' cannot be read, exiting ...\n",
	       envMapsFilename[em]);
	deallocations();
	exit(-1);
      }
      environMap[em].savePNG("a.png");
      char aux[100]; strcpy(aux, "a"); strcat(aux, FILEEXT);
      environMap[em].saveDebugHDR(aux, 0, 1.f);
    }
    else {
      // Read HDR file as environment map
      readStatus = environMap[em].loadHDR((const char*)(envMapsFilename[em]));
      if (rotateEM != 0.f) {
	// rotate EM by given angle given in degrees
	int degrees = rotateEM;
	int pixelOffset = (int)((float)degrees/(360.f) * (float)environMap[em].getWidth());
	environMap[em].horizontallyShift(pixelOffset);
	if (1) {
	  char aux[1000];
	  sprintf(aux,"%s_r%03d.exr", environMap[em].getFileName().c_str(),degrees);
	  environMap[em].saveHDR(aux); // save as new file
	}
      }

      // set whole EM map to constant
      if (setConstEMflag) {// &&em == actualEM)
	environMap[em].setConstant(1.0f,1.0f,1.0f);
	cout << "Setting constant half hemisphere" << endl;
      }
      else {
	// set north hemisphere of EM map to constant (theta=0 .. northpole) and
	// set south hemisphere to almost zero
	if (setConstNorthHemEMflag) { // &&em == actualEM)	
	  environMap[em].setFunctionConstNorthHemisphere(1.0f,1.0f,1.0f);
	  cout << "Setting constant north half hemisphere" << endl;
	}
	else {
	  // set north hemisphere of EM map to consine(theta) with respect to the northpole
	  // and south hemisphere set to almost zero value
	  if (setCosineNorthHemEMflag) { // &&em == actualEM)
	    environMap[em].setFunctionCosineThetaNorthHemisphere(1.0f,1.0f,1.0f);
	    cout << "Setting cosine north half hemisphere" << endl;
	  }
	}
      }
      // Optionally, set south hemisphere to almost zero value
      if (semConstSouthHemEMflag) { // &&em == actualEM)
	const float almostZero = 1e-20;
	environMap[em].setFunctionConstSouthHemisphere(almostZero, almostZero, almostZero);
	cout << "Setting constant north half hemisphere" << endl;
      }
      
#if 0
      // For debugging we can overwrite the EM by another function here
      //environMap[em].setConstant(1.0f,1.0f,1.0f); // constant
      //environMap[em].setFunctionConstNorthHemisphere(1.0f,1.0f,1.0f); // constant
      //environMap[em].setFunctionCosineThetaNorthHemisphere(1.0f,1.0f,1.0f); // cos(theta)
      //environMap[em].setFunctionCosineTheta(1.0f,1.0f,1.0f); // cos(theta)
      //environMap[em].setFunction2(1.0f,1.0f,1.0f); // sin(theta)
      //environMap[em].setFunction3(1.0f,1.0f,1.0f); // some more complex function
#endif      
      if (!readStatus) {
	printf("ERROR: Envmap %s cannot be read\n", (const char*)(envMapsFilename[em]));
	deallocations();
	exit(-1);
      }
    }

    //environMap[em].saveTGA("aaa.tga", 0, 100);
    
    // compute the luminance and CDF of an image from luminance
    // for importance sampling
    environMap[em].computeLuminance();
    // let us normalize the power of the environment map here!
    environMap[em].normalizePower();
    // and compute cumulative density function used for sampling
    environMap[em].computeCDF();
    // compute total variance
    //environMap[em].getTotalVarianceSphere();
#if 0
    // Debug test of the algorithm for variance of hemisphere
    for (int i = 0; i < 1000;i++) {
      CVector3D result;
      CVector2D rndvalue;
      rndvalue.xx = RandomValue();
      rndvalue.yy = RandomValue();
      CVector3D rnddir = RND::sph_const(rndvalue);
      float XYZ[3];
      XYZ[0] = rnddir.x;
      XYZ[1] = rnddir.y;
      XYZ[2] = rnddir.z;
      float var =  environMap[em].getVarianceHemisphere(XYZ);
      // cout << "XYZ = " << XYZ[0] << " " << XYZ[1] << " " << XYZ[2] << endl;
    }
#endif
    
    // determine size of OpenGL textures
    readStatus = 0; // OK
    envPlanes[em] = 3; // 3 planes
    envImageRows[em] = environMap[em].getHeight();
    envImageCols[em] = environMap[em].getWidth();
    //	CTGAImage dd;
    //	dd.copyImageDataFromEM(environMap[em], 0, 100);
    //	dd.SaveToFile("00.tga");
    
    // specifies closest 2^n texture size to actual env. map
    if ( (envImageRows[em] > 512) || (envImageCols[em] > 512) ) {
      if ( (envImageRows[em] > 1024) || (envImageCols[em] > 1024) ) {
	envCols[em] = 2048; // maximum texture size
	envRows[em] = 2048; // limited to 2048, 1024 or 512
      }
      else {
	envCols[em] = 1024;
	envRows[em] = 1024;
      }
    }
    else {
      envCols[em] = 512;
      envRows[em] = 512;
    }
    printf("RS: %d Planes: %d Img. size: [%d,%d] Tex. size: [%d,%d]\n",
	   readStatus, envPlanes[em], envImageRows[em],
	   envImageCols[em], envRows[em], envCols[em]);
    
    assert(envPlanes[em] == 3); // RGB either in HDR or PNG image    
  } // for em    

//#define GENERATE_OPENGL_FIRST
#ifdef GENERATE_OPENGL_FIRST
  if (supportOpenGL) {
    GenerateOPENGLtextures();
  }
#endif

  char text[1000];
  int number;
  float fnumber;
  FILE *fp = 0;
  envPhiShift = 0;

  //DeadCode001(); // possibly read corrections for BTF materials
  
  // specifiing files with lights coord & colours corresponding to env. maps
  char envFiles[NoOfEnvMaps][1000];
  const float multRadius = 1.0f;
  { // usePrecomputedLights = false
    for(int em=0;em<NoOfEnvMaps;em++) {
      // precompute possibly more lights than will be needed.
      environMap[em].precomputeLights(maxCntSamplesPerPixel);
      // environMap[em].rotate(initEnvShifts[em]);
    }
  } // compute the data on the fly
  
  printf("ENVIRONMENT LIGHTS AND COLOURS LOADING/PRECOMPUTING FINISHED\n");

#if 0
  // This is only for development
  environMap[0].Test_Random();
#endif    
  
#if 0
  // Debugging visualization
  char tt[200];
  // This is only for debugging environment map sampling
  for (int iEM = 0; iEM < NoOfEnvMaps; iEM++) {
    sprintf(tt,"0_%d.wrl", iEM);
    // Visualization to VRML files
    VRMLStreamOpen(tt);
    VRMLStreamSaveAxes(0, 0, 0, 1.f, 0.001f);
    //environMap[iEM].MakeMarkRGB(100, 100, EE_ENVMAPSamples, 10, 100.f);
    //environMap[iEM].saveTGA("0.tga", 0, 1.0);
    //exit(0);
    float sumlum = 0.f;
    // the access to light sources cannot be randomized
    bool flag = environMap[iEM].getModeRandomLights();
    environMap[iEM].setModeRandomLights(false);
    // the access to light sources cannot be randomized
    for(int i=0; i<maxCntSamplesPerPixel; i++) {
#ifdef XYZCOORSYSTEM
      float x = environMap[iEM].getDirectionPrecomputedLight(i, 0);
      float y = environMap[iEM].getDirectionPrecomputedLight(i, 1);
      float z = environMap[iEM].getDirectionPrecomputedLight(i, 2);
#else
      float x = environMap[iEM].getDirectionPrecomputedLight(i, 0);
      float y = environMap[iEM].getDirectionPrecomputedLight(i, 2);
      float z = environMap[iEM].getDirectionPrecomputedLight(i, 1);      
#endif      
      float RGB[3]; 
      RGB[0] = environMap[iEM].getColorPrecomputedLight(i, 0);
      RGB[1] = environMap[iEM].getColorPrecomputedLight(i, 1);
      RGB[2] = environMap[iEM].getColorPrecomputedLight(i, 2);
      float lum = luminance(RGB);
      float prob = environMap[iEM].getProbPrecomputedLight(i);
      VRMLStreamPutPoint(x, y, z, 255.0, 255.0, 255.0, 0.002f);
      float denom = sqrt(x*x + y*y + z*z);
      float phi = atan2(y, x);
      if (phi < 0.f) phi += 2.0*M_PI;
      float theta = acos(z/denom);
      //cout << "x = " << x << "y = " << y << " z = " << z << endl;
      //cout << "phi = " << phi << " theta = " << theta << endl;
      //cout << " lum = " << lum << " prob = " << prob << " lum/prob = "
      //   << lum/prob << endl;
      int mx = phi/(2.f*M_PI)*environMap[iEM].getWidth();
      int my = theta/(M_PI)*environMap[iEM].getHeight();
      environMap[iEM].makeMarkRGB(mx, my, EE_ENVMAPSamples, 10, 10000.f);    
      //environMap[iEM].makeMarkLUM(mx, my, EE_ENVMAPSamples, 10, 10000.f);    
    } // for
    VRMLStreamClose();
    // the access to light sources cannot be randomized
    environMap[iEM].setModeRandomLights(flag); 
    // saving color image
    sprintf(tt,"c_%d.hdr", iEM);
    environMap[iEM].saveDebugHDR(tt, 0, 1.0);
    // saving luminance image
    sprintf(tt,"l_%d.hdr", iEM);
    environMap[iEM].saveDebugHDR(tt, 1, 1.0);
    //exit(0);
  } // for 
#endif
  
#ifndef GENERATE_OPENGL_FIRST
  // we generate the markers later so they are visible during the rendering
  if (supportOpenGL) {
    GenerateOPENGLtextures();
  }
#endif

  // Set possibly randomization of directional lights
  for (int iEM = 0; iEM < NoOfEnvMaps; iEM++) {
    environMap[iEM].setModeRandomLights(randomizePrecomputedLightsFlag);
  }
    
  return;
} //  -------- LoadEnvironmentMaps() -----------------------------------------------

void
Set3DObjectsNames(bool allocateFlag)
{
  //=== OBJ data allocation and loading ======================================
  if (allocateFlag) {
    for(int i=0;i<NoOfObjects;i++) {
      OBJnames[i] = callocation1(0,1000);
      OBJpaths[i] = callocation1(0,1000);
    }
  }
#if 0
  strcpy(OBJpaths[0],"OBJ/test_rectangle2.obj");
  strcpy(OBJnames[0],"rectangle");
  OBJscales[0] = 7.f;
  // Only for debugging new methods
  initOBJrot[0][0]   = 0; //1.f/16.f*M_PI;
  initOBJrot[0][1]   = 0.0; // 0.f; -1.19;
  initOBJrot[0][2]   = 0;//M_PI/2.f;
  initOBJtrans[0][0] = 0.f; 
  initOBJtrans[0][1] = 0.f;
  initOBJtrans[0][2] = 0.f;
#endif
#if 0
  strcpy(OBJpaths[0],"OBJ/koule.obj");
  strcpy(OBJnames[0],"sphere");
  OBJscales[0] = 0.05f;
  initOBJrot[0][0]   = 0.f; 
  initOBJrot[0][1]   = 0.f; 
  initOBJrot[0][2]   = 0.f; 
  initOBJtrans[0][0] = 0.f; 
  initOBJtrans[0][1] = 0.f; 
  initOBJtrans[0][2] = 0.f; 
#endif
#if 1
  strcpy(OBJpaths[0],"OBJ/bunnyplane.obj");
  strcpy(OBJnames[0],"bunny");
  OBJscales[0] = 0.5f;
  initOBJrot[0][0]   = M_PI/2.f; 
  initOBJrot[0][1]   = 0;
  initOBJrot[0][2]   = M_PI/2.f; 
  initOBJtrans[0][0] = 0.f; 
  initOBJtrans[0][1] = 0.f; 
  initOBJtrans[0][2] = 0.f; 
#endif
#if 0
  strcpy(OBJpaths[0],"OBJ/teapot.obj");
  strcpy(OBJnames[0],"teapot");
  OBJscales[0] = 0.007f;
  initOBJrot[0][0]   = -0.5f; 
  initOBJrot[0][1]   = 3.14; //  6.f*M_PI/7.f;
  initOBJrot[0][2]   = 1.8f; 
  initOBJtrans[0][0] = 0.f; 
  initOBJtrans[0][1] = 0.f; 
  initOBJtrans[0][2] = 0.f; 
#endif  


  return;
}

void
Read3DObjects(bool supportOpenGL = 1)
{
  for(int io=0;io<NoOfObjects;io++) {
    DATA[io] = new TObjectData(verticesPerObject);
    assert(DATA[io]);
    DATA[io]->loadOBJfile(OBJpaths[io], 1, OBJscales[io], 0.f,0.f,0.f); // loading OBJ data
    //      DATA[io]->testOnDegeneratedPolygons();
    //      DATA[io]->testAndCorrectNormalsContinuity();
    //      DATA[io]->normalsCorrection();
    
    // precomputing per vertex tangents and binormals and normals
    DATA[io]->tangentsPerVertex();
    DATA[io]->binormalsPerVertex();
    
    // loading data from the object to buffer for faster rendering and rotation enabling
    V[io] = allocation3(1,verticesPerObject, 0,2, 0,2); 
    Vbackup[io] = allocation3(1,verticesPerObject, 0,2, 0,2); 
    T[io] = allocation3(1,verticesPerObject, 0,2, 0,2); 
    Tbackup[io] = allocation3(1,verticesPerObject, 0,2, 0,2); 
    N[io] = allocation3(1,verticesPerObject, 0,2, 0,2); 
    Nbackup[io] = allocation3(1,verticesPerObject, 0,2, 0,2); 
    B[io] = allocation3(1,verticesPerObject, 0,2, 0,2); 
    Bbackup[io] = allocation3(1,verticesPerObject, 0,2, 0,2); 
    UV[io] = allocation3(1,verticesPerObject, 0,2, 0,1); 
    UVbackup[io] = allocation3(1,verticesPerObject, 0,2, 0,1); 
    
    // 3 floats per vertex, 3 vertices per triangle
    float *verticesN = new float[3 * 3 * verticesPerObject];
    assert(verticesN);
    //float *normalsN = new float[3 * 3 * verticesPerObject];
    //assert(normalsN);
    float *UVN = new float[2 * 3 * verticesPerObject];
    assert(UVN);
    int indexVN = 0, indexUVN = 0; // indexNN = 0;
    int cntFaces = DATA[io]->noOfFaces();
    
    for(int f=1; f<= cntFaces;f++) {
      for(int k=0;k<3;k++) // for vertices
      {
	for(int j=0;j<3;j++) // for x,y,z
	{
	  verticesN[indexVN++] = Vbackup[io][f][k][j] = V[io][f][k][j] = DATA[io]->getV(f,k,j);
	  
	  Nbackup[io][f][k][j] = N[io][f][k][j] = DATA[io]->getN(f,k,j);
	  //normalsN[indexNN++] = DATA[io]->getN(f,k,j);
	  
	  Tbackup[io][f][k][j] = T[io][f][k][j] = DATA[io]->getT(f,k,j);
	  Bbackup[io][f][k][j] = B[io][f][k][j] = DATA[io]->getB(f,k,j);
	}

	// test for zero vectors
	if (isnan_local(T[io][f][k][0]) || 
	    isnan_local(T[io][f][k][0]) || 
	    isnan_local(T[io][f][k][0])) {
	  printf("NaN in Tangent at face: %d vertex: %d - [%g,%g,%g]. Exiting...\n",
		 f,k,T[io][f][k][0],T[io][f][k][1],T[io][f][k][2]);
	  deallocations();
	  exit(-1);
	}

	if (vec_len(T[io][f][k])<=0.f) {
	  printf("Zero Tangent at face: %d vertex: %d - [%g,%g,%g]. Exiting...\n",
		 f,k,T[io][f][k][0],T[io][f][k][1],T[io][f][k][2]);
	  deallocations();
	  exit(-1);
	}
	if(vec_len(N[io][f][k])<=0.f) {
	  printf("Zero Normal at face: %d vertex: %d - [%g,%g,%g]. Exiting...\n",
		 f,k,N[io][f][k][0],N[io][f][k][1],N[io][f][k][2]);
	  deallocations();
	  exit(-1);
	}
	if(vec_len(B[io][f][k])<=0.f) {
	  printf("Zero Binormal at face: %d vertex: %d - [%g,%g,%g]. Exiting...\n",
		 f,k,B[io][f][k][0],B[io][f][k][1],B[io][f][k][2]);
	  deallocations();
	  exit(-1);
	}

	for(int j=0;j<2;j++) {// for u,v
	  UVN[indexUVN++] = UVbackup[io][f][k][j] = UV[io][f][k][j] = DATA[io]->getUV(f,k,j);
	}
      }// for faces
    }

    if (supportOpenGL) {
      // Single object has been read
      mesh[io].init(DATA[io]->noOfFaces()*3, GL_TRIANGLES, verticesN, UVN, io);
    }
    delete []verticesN;
    //delete []normalsN;
    delete []UVN;
    
    //##### initial rotation of OBJ model ##############################################
    //      rotateOBJModelGeometry(i, 2, -M_PI/2.f); // turn object for 90 degrees along horiz. axis    
    //rotateOBJModelGeometry(io, 0, initOBJrot[io][0]);
    OBJrot[io][0] = initOBJrot[io][0];
    //mesh[io].rotateOBJModelGeometry(0, initOBJrot[io][0]);
    
    //rotateOBJModelGeometry(io, 1, initOBJrot[io][1]);
    OBJrot[io][1] = initOBJrot[io][1];
    //mesh[io].rotateOBJModelGeometry(1, initOBJrot[io][1]);
    
    //rotateOBJModelGeometry(io, 2, initOBJrot[io][2]);
    OBJrot[io][2] = initOBJrot[io][2];
    //mesh[io].rotateOBJModelGeometry(2, initOBJrot[io][2]);
    
    // axis +/-: 1 (zoom in/out), 2 (down/up), 0 (right/left)
    //translateOBJModelGeometry(io, 0, initOBJtrans[io][0]);
    OBJtrans[io][0] = initOBJtrans[io][0];
    //mesh[io].rotateOBJModelGeometry(0, initOBJrot[io][0]);
    
    //translateOBJModelGeometry(io, 1, initOBJtrans[io][1]);
    OBJtrans[io][1] = initOBJtrans[io][1];
    //mesh[io].rotateOBJModelGeometry(1, initOBJrot[io][1]);

    //translateOBJModelGeometry(io, 2, initOBJtrans[io][2]);    
    OBJtrans[io][2] = initOBJtrans[io][2];
    //mesh[io].rotateOBJModelGeometry(2, initOBJrot[io][2]);
  }// for objects
  return;
} // ----- Read3DObjects() --------------------------------------------------------------

void 
init(int supportOpenGL = 1)
{
  if (readReferenceImageFlag) {
    referenceImage = new CHDRImage();
    assert (referenceImage);
    if (!referenceImage->loadHDR(referenceImageFilename)) {
      cout << "Reference image = " << referenceImageFilename << " cannot be read" << endl;
      referenceImageAvailable = false;
      delete referenceImage;
      referenceImage = 0;
      abort();
    }
    else {
      referenceImageAvailable = true;
      cout << "Refimage (reference) was successfully read from " << referenceImageFilename << endl;
    }
  }
  
  if (maxCntSamplesPerPixel < cntSamplesPerPixel) {
    cout << "ERROR: maxCntSamplesPerPixel < cntSamplesPerPixel cannot happen\n" << endl;
    abort();
  }

  auxString = callocation1(0, 1000);
  assert(auxString);
  sprintf(auxString, "%sfilename.test", FILEPREFIX);
  FILE *fpp = fopen(auxString, "wb");
  if (!fpp) {
    cout << "ERROR: Filepath " << FILEPREFIX << " does not exist" << endl;
    cout << "Opening test file " << auxString << " failed" << endl;
    if (FILEPREFIX[strlen(FILEPREFIX-1)] != '/')
      strcat(FILEPREFIX,"/");
    // try to make a directory
    cout << "Trying to make a directory " << FILEPREFIX << endl;
    mkdir(FILEPREFIX,0777);
    sprintf(auxString, "%sfilename.test", FILEPREFIX);
    fpp = fopen(auxString, "wb");
    if (!fpp) {
	cout << "ERROR: Filepath " << FILEPREFIX << " does not exist" << endl;
	cout << "Opening test file " << auxString << " failed" << endl;
	cout << "Exiting" << endl;    
	exit(-1);
    } // if even second opening has failed
  }
  fclose(fpp);
  
  TA = new TAngles(); // object filling for measurement angles and interpolation handling
  assert(TA);

  // BTF texture size initialization and data array allocation
  // setting initial texture size
  BTFrows = (int)pow(2, Npow);
  assert(BTFrows);
  BTFcols = (int)pow(2, Npow);
  assert(BTFcols);
  // maximum texture size here for rendering
  for (int i = 0; i < 5; i++) {
    // maximum texture size here for visualization and other purposes
    int sizeT = 2048*2048*4;
    BTFimageB[i] = allocation1(0, sizeT);
    assert(BTFimageB[i]);
    // zero the image
    for (int j = 0; j <= sizeT; j++) BTFimageB[i][j]=0;
  }
  BTFimageF = BTFimageB[0];
  assert(BTFimageF);
      
  SetRandomSeed(1); // reinitializing Mersenne Twister

  //=== setting camera nad light positions ===============================
  assert(TA);

  // Camera position
  viewerDistance = viewerDistanceInit;
  viewerAngle = viewerAngleInit;
  createCameraposFromObjectPosition();

  // Single point light
  lposBackup[0] = lpos[0] = lposInit[0];
  lposBackup[1] = lpos[1] = lposInit[1];
  lposBackup[2] = lpos[2] = lposInit[2];
  lposBackup[3] = lpos[3] = lposInit[3];
  TA->setIlluPos(lpos);

  //=== setting material reflectance properties ==========================
  float value = 0.5;  
  float diffuse0[] = {value, value, value, 1.0f};
  value = 0.5;
  float ambient[] = {value, value, value, 1.0f}; 
  value = 0.5;
  float specular[] = {value, value, value, 1.0f}; 
  //=== test of required texturesize and No. of TUs ======================    
  GLint ntex=0;
  ntex = 2048; // I do not know
  if(ntex >= 2048)
    printf("Max. texture size: %d\n", ntex);
  else {
    printf("Insufficient max. texture size! Available: %dpx (required: 2048px)\n",ntex);
    exit(1);
  }

  if (supportOpenGL) {
    //=== set rendering options ===================================================
    glClearColor(backgroundColor[0], backgroundColor[1], backgroundColor[2], backgroundColor[3]);
    
    //=== prepare texture for BTF rendering =================================
    glGenTextures(5, BTFidTexture); // several textures
    CHECK_GL_ERROR();
  }
  LoadEnvironmentMaps(supportOpenGL);
  assert(alphaMIS >= 0.f);
  assert(alphaMIS <= 1.f);
  Read3DObjects(supportOpenGL);
  
  // for environment map/BRDF sampling initial number of samples
  cntSamplesPerPixel = defaultCntSamplesPerPixel;  
  return;
}//--- init ------------------------------------------------------


//#####################################################################################
//### MAIN ############################################################################
//#####################################################################################
void
printInitParameters()
{
  printf("________________________________________________________________________\n");
  printf("| BTFBASE - INTERACTIVE RENDERING (resolution %d x %d)\n", win_fbo_width, win_fbo_height);
  if(IsEnvMapLighting())
    printf("|  3D model: %d, env. lighting: 1, env. map: %d, no. of lights: %d \n",
	   actualOBJ, actualEnvMap, defaultCntSamplesPerPixel);
  else
    printf("|  3D model: %d, env. lighting: 0, point light pos: [%g,%g,%g]\n",
	   actualOBJ, lposInit[0], lposInit[1], lposInit[2]);
  printf("|  BTFresolution: %d x %d \n", (int)pow(2,Npow), (int)pow(2,Npow));
  printf("|_______________________________________________________________________\n\n");
  return;
}//--- printInitParameters ----------------------------------------

// This is only for development, testing correctness of underlying BRDF models
// used in the project
int
computeBRDFstats(BRDF *brdf, float &meanAlbedoO, float &meanVarO, float &stats, int &cntEvalsO)
{
  // cout << "Test_BRDF_Random started" << endl;
  // we integrate BRDF by cosine sampling over the hemisphere
  long double meanAlbedo = 0.f;
  long double meanVar = 0.f;
  long long int cntEvals = 0;
  // for different theta and phi
  const int stepPhi = 90;
  const int stepTheta = 5;
  const int irow = 0, jcol = 0;
  //  brdf->setSingleCacheSVA(1); // avoid interpolation
  //  brdf->setCntThresholdMean(1000); // do 1000 samples for one value
  int cntSamplesView = 1000; // ten milion of evaluations of BRDF to get a value
  for (;cntSamplesView--;) {
    float theta_v, phi_v;
    // we compute 1/PI sampling over hemisphere here
    float rndxx = RandomValue();
    float rndyy = RandomValue();
    float sin_theta = sqrt(1.0000 - rndxx * rndxx);
    if (sin_theta >= 1.0f) sin_theta=0.99999;
    theta_v = asin(sin_theta);
    assert(theta_v >=0);
    assert(theta_v <= M_PI/2);
    phi_v = 2.0 * M_PI * rndyy;
    float RGB[3];
    // Now let us integrate by importance sampling from the BRDF
    //cnt = 1000000;
    float albedo = brdf->getAlbedo(irow, jcol, theta_v, phi_v);
    //float var = brdf->getStats(irow, jcol, theta_v, phi_v);
    meanAlbedo += (long double)albedo;
    //meanVar += (long double)var;
    cntEvals++;
#if 0
    float multDIFF = 1.0f;    
    cout << "albedo = " << albedo << " var= "<< var << " var/albedo"
	 << var/albedo << " formulaDiffuseness="
	 << 1.0f/(1.0f+sqr(multDIFF*var/albedo)) << endl;
#endif    
    //int cntSamplesPrecomp = brdf->getCntSamplesPrecomp();
    //cout << "albedo = " << albedo << " precomputed="<< cntSamplesPrecomp << endl;
  } // for all samples;

  meanAlbedo /= (long double)cntEvals;
  meanVar /= (long double)cntEvals;

  // copy the results to the output
  meanAlbedoO = meanAlbedo;
  meanVarO = meanVar;
  cntEvalsO = cntEvals;
  // compute variance by square of total albedo
  stats = meanVar/sqr(meanAlbedo);
}

void
computeBRDFstats(BRDF *brdf)
{
  // mean albedo, meanVar0, cntEvals used.
  float meanAlbedo, meanVar, stats;
  int cntEvals; // how many evaluations of albedo were used
  computeBRDFstats(brdf, meanAlbedo, meanVar, stats, cntEvals);
  string btext;
  brdf->getString(btext);
  cout << "meanAlbedo|meanVar|charSts " <<  meanAlbedo << " " << meanVar << " " << stats 
       << " description " << btext << endl;
  return;
}

// Compute statistics from environment map
int
computeEMstats(int actualEM)
{
  const int cntSamplesSTATS = 10000000; // how many samples to take
  // The EM map was processed, do only print
  float power = environMap[actualEnvMap].getTotalPower(); // this corresponds to albedo
  // original prior possible normalization in Watts
  float powerOrig = environMap[actualEnvMap].getTotalPowerOrig(); 
  //float var = environMap[actualEnvMap].getTotalVarianceSphere(); 
  //float var = environMap[actualEnvMap].computeVarianceSphere(cntSamplesSTATS);
  float var = 0.0f; // temporary
  float statsSCV = var/sqr(power);
  cout <<  environMap[actualEnvMap].getName() << " "
       << environMap[actualEnvMap].getFileName() << " ";
#define ALSO_NORTHHEMISPHERE
#ifdef ALSO_NORTHHEMISPHERE
  float normalXYZ[3] = {0,0,1}; // the north hemisphere
  float meanHS, varHS;
  environMap[actualEnvMap].computeMeanAndVarianceHemisphere(normalXYZ, meanHS, varHS, cntSamplesSTATS);
  cout << "POWERORIG|POWER|VAR|STATS|MEANHS|VARHS|SCVHS " << powerOrig << " " << power << " "
       << var << " " << statsSCV << " " << meanHS << " " << varHS << " " << varHS/(sqr(meanHS));
#else
       << "POWERORIG|POWER|VAR|STATS|SCV " << powerOrig << " " << power << " "
       << var << " " << statsSCV << ;
#endif
  cout << endl << flush;

  return 0;
}

// Compute statistics from environment map for only north hemisphere
int
computeEMHALFstats(int actualEM)
{
  float normalXYZ[3] = {0,0,1}; // the north hemisphere
  float meanHS, varHS;
  const int cntSamplesHS = 10000000;
  environMap[actualEnvMap].computeMeanAndVarianceHemisphere(normalXYZ, meanHS, varHS, cntSamplesHS);
  cout << "MEANHS|VARHS|STATSHS " << meanHS << " " << varHS << " " << varHS/(sqr(meanHS));
  cout << endl << flush;
  return 0;
}

void
printUsage()
{
  cout << "Usage:" << endl;
  cout << "lightenv [options] [brdfmodel options]" << endl;
  cout << "[brdfmodel options] = one of below:" << endl;
  cout << "1) phong kd ks rhospec"<< endl;
  cout << "2) phong3 kdR kdG kdB ksR ksG ksB rhospec"<< endl;
  // Here it goes new models = .....
  cout << " [new BRDF model] ... " << endl;
  cout << "-varBRDF .. compute stats BRDF model" << endl;
  cout << "-varEM .. compute stats EM" << endl;
  cout << "-varEMH .. compute stats hemisphere (north) given by EM" << endl;
  cout << "-render .. do rendering (default)" << endl;
  cout << "-nointeractivity .. disables reactions to keys" << endl;
  cout << "-verbose 0|1 .. switch ON/OFF estimate of rendering time" << endl;
  cout << "-obj index .. obj by index" << endl;
  cout << "-objname filename .. set OBJ from a file" << endl;
  cout << "-em index .. EM by index" << endl;
  cout << "-emhack index .. EM specification hack by index" << endl;
  cout << "-emname filename .. set EM from a file" << endl;
  cout << "-emconst .. set whole EM const" << endl;
  cout << "-emnorthhemconst .. set north hemisphere of EM const and south to almost zero" << endl;
  cout << "-emnorthhemcosine .. set north hemisphere of EM cosine(theta) and south to almost zero" << endl;
  cout << "-emsouthhemzero .. set south hemisphere to almost zero" << endl;
  cout << "-emrotate angle .. rotated EM for 'angle' degrees after loading" << endl;
  cout << "-emrandomize 0|1 .. randomize (default) or not light sources from EM." << endl;
  cout << "-cnt samplesPerPixel .. set samples per pixel" << endl;
  cout << "-cntref samplesPerPixel .. set samples per pixel for reference method" << endl;
  cout << "-alphamis alpha .. alpha for MIS" << endl;
  cout << "-saveGray .. save also gray tone image." << endl;
  cout << "-saveNormAlbedo .. normalizedAlbedo when saving the images, save two version" << endl;
  cout << "-saveAndExit .. save and immediately exit" << endl;
  cout << "-method id .. use method/algorithm of ID for rendering" << endl;
  cout << "-atlas 0|1 .. save or not save atlas" << endl;
  cout << "-modenewsamples 1|2|3 .. how to compute number of samples based on variance" << endl;
  cout << "-pseudocolor 0|1|2|3|4|5|6 .. visualization function using pseudocolor scheme" << endl;
  cout << "-refimagefname filename.hdr .. the explicit name of the reference image, default refimage.hdr" << endl;
  cout << "-diffimagefname filename.hdr .. the explicit name of the diffimage, default diffimage.hdr" << endl;
  cout << "-diffvisimagefname filename.hdr .. the explicit name of the diffimage, default diffvisimage.hdr" << endl;
  cout << "-readrefimage 0|1 .. flag if to read the reference image or not" << endl;
  cout << "-o filepath .. where to store the generated imaged" << endl;
  cout << "-vv index .. variable to save to visualization" << endl;
  cout << "-upperboundvis value .. upper bound of RMSE Error for visualization, default 1.0" << endl;
  return;
}

// possibly decrease argc so removing processed switches
int
ProcessSwitches(int &argc, int offset, char **argv)
{
  int i = offset;
  for (; i < argc;) {
    if (strcmp(argv[i],"-varBRDF") == 0) {
      // compute average albedo, variance etc. for BRDF model and finish
      assert(i+1 < argc);
      whatToDo = 1;
      i+=1; // parsed, the next param different
      cout << "Program is going to compute albedo, variance and chars. for BRDF model only." << endl;
      continue;
    }
    if (strcmp(argv[i],"-varEM") == 0) {
      // compute average albedo, variance etc. for full EM and finish
      assert(i+1 < argc);
      whatToDo = 2;
      i+=1; // parsed, the next param different
      cout << "Program is going to compute albedo, variance and chars. for EM map" << endl;
      continue;
    }
    if (strcmp(argv[i],"-varEMH") == 0) {
      // compute average albedo, variance etc. for half of the hemisphere and finish
      assert(i+1 < argc);
      whatToDo = 3;
      i+=1; // parsed, the next param different
      cout << "Program is going to compute albedo, variance and chars. for EM map" << endl;
      continue;
    }
    if ( (strcmp(argv[i],"-diff") == 0)||
         (strcmp(argv[i],"-compare") == 0)||
         (strcmp(argv[i],"-cmp") == 0) ) {
      // it does difference between two images
      assert(i+1 < argc);
      whatToDo = 4;
      i+=1; // parsed, the next param different
      cout << "Program is going to compare two images." << endl;
      return i; // we start with this parameter as the next, we do not change argv
    }
    if (strcmp(argv[i],"-render") == 0) {
      // render the image
      assert(i+1 < argc);
      whatToDo = 0;
      i+=1; // parsed, the next param different
      cout << "Program is going to render the image." << endl;
      continue;
    }
    if (strcmp(argv[i],"-nointeractivity") == 0) {
      // variable to save the visualization
      bool k = disableInteractivity;
      assert(i+1 < argc);
      int vv = atoi(argv[i+1]);
      disableInteractivity = (vv == 0) ? false: true;
      i+=2; // parsed, the next param different
      cout << "Disabling interactivity = " << disableInteractivity << endl;
      continue;
    }
    if (strcmp(argv[i],"-verbose") == 0) {
      // if to verbosely print estimate of time
      bool k = printEstimateFlag;
      assert(i+1 < argc);
      int vv = atoi(argv[i+1]);
      printEstimateFlag = (vv == 0) ? false: true;
      i+=2; // parsed, the next param different
      cout << "Setting verbose to " << k << endl;
      continue;
    }
    if (strcmp(argv[i],"-obj") == 0) {
      // which object will be rendered from those read by given index
      int k = actualOBJ;
      assert(i+1 < argc);
      actualOBJ = atoi(argv[i+1]);
      assert(actualOBJ >= 0);
      assert(actualOBJ < NoOfObjects);
      i+=2; // parsed, the next param different
      cout << "Changing the object to be rendered from " << k
	   << " to " << actualOBJ << endl;
      continue;
    }
    if (strcmp(argv[i],"-objname") == 0) {
      // which object will be rendered by path filename
      char oldname[200]; strcpy(oldname, OBJnames[actualOBJ]);
      char oldpath[200]; strcpy(oldpath, OBJpaths[actualOBJ]);
      //  int k = actualOBJ;
      assert(i+2 < argc);
      strcpy(OBJnames[actualOBJ], argv[i+1]);
      strcpy(OBJpaths[actualOBJ], argv[i+2]);
      assert(actualOBJ > 0);
      assert(actualOBJ < 20);
      i +=3; // parsed, the next param different
      cout << "Changing the OBJ to be rendered from " << oldname << " " << oldpath
	   << " to " << OBJnames[actualOBJ] << " " << OBJpaths[actualOBJ] << endl;
      continue;
    }
    
    if (strcmp(argv[i],"-em") == 0) {
      // which environment map will be used given the index
      assert(i+1 < argc);
      int k = actualEnvMap;
      actualEnvMap = atoi(argv[i+1]);
      assert(actualEnvMap >= 0);
      assert(actualEnvMap < NoOfEnvMaps);
      i+=2; // parsed, the next param different
      cout << "Changing the EM map to be rendered from " << k
	   << " to " << actualEnvMap << endl;
      continue;
    }
    if (strcmp(argv[i],"-emhack") == 0) {
      // which environment map will be used given the index
      assert(i+1 < argc);
      int k = envMapToReadHACK;
      envMapToReadHACK = atoi(argv[i+1]);
      assert(envMapToReadHACK >= 0);
      assert(envMapToReadHACK < NoOfEnvMapsToReadHACK);
      i+=2; // parsed, the next param different
      cout << "Changing the EMHACK index to be rendered from " << k
	   << " to " << envMapToReadHACK << endl;
      continue;
    }
    if (strcmp(argv[i],"-emname") == 0) {
      // which environment map will be used given the path filename
      char oldpath[200]; strcpy(oldpath, envMapsFilename[actualEnvMap]);
      int k = actualEnvMap;
      assert(i+1 < argc);
      strcpy(envMapsFilename[actualEnvMap], argv[i+1]);
      assert(actualEnvMap >= 0);
      assert(actualEnvMap < 20);
      i+=2; // parsed, the next param different
      cout << "Changing the EM map to be rendered from " << oldpath
	   << " to " << envMapsFilename[actualEnvMap] << endl;
      continue;
    }
    if (strcmp(argv[i],"-emconst") == 0) {
      // Setting the EM to be constant
      setConstEMflag = true; setConstNorthHemEMflag = false; setCosineNorthHemEMflag = false;
      i++;
      continue;
    }
    if (strcmp(argv[i],"-emnorthhemconst") == 0) {
      // Setting the EM to be constant for north hemisphere
      setConstNorthHemEMflag = true; setConstEMflag = false; setCosineNorthHemEMflag = false;
      i++;
      continue;
    }
    if (strcmp(argv[i],"-emnorthhemcosine") == 0) {
      // Setting the EM to be constant for north hemisphere
      setCosineNorthHemEMflag = true; setConstEMflag = false; setConstNorthHemEMflag = false;
      i++;
      continue;
    }
    if (strcmp(argv[i],"-emsouthhemzero") == 0) {
      // Setting the EM to be almost zero for the south hemisphere, the rest is unchanged.
      semConstSouthHemEMflag = true;
      i++;
      continue;
    }
    if (strcmp(argv[i],"-emrotate") == 0) {
      // how much to rotate environment map by phi after reading, applied to all read EMs
      // angle is specified in degrees!
      float rotateOLD = rotateEM;
      assert(i+1 < argc);
      rotateEM = atof(argv[i+1]);
      assert(rotateEM > -360); assert(rotateEM <= 360);
      i+=2; // parsed, the next param different
      cout << "Rotating all EMs by " << rotateEM << " radians." << endl;
      continue;
    }    
    if (strcmp(argv[i],"-emrandomize") == 0) {
      // if to randomize (default) or not light sources from EM.
      bool k = randomizePrecomputedLightsFlag;
      assert(i+1 < argc);
      int vv = atoi(argv[i+1]);
      randomizePrecomputedLightsFlag = (vv == 0) ? false: true;
      i+=2; // parsed, the next param different
      cout << "Setting EM randomization from " << k << " to " << vv << endl;
      continue;
    }
    if (strcmp(argv[i],"-cnt") == 0) {
      // setting number of samples to a value
      int k = defaultCntSamplesPerPixel;
      assert(i+1 < argc);
      defaultCntSamplesPerPixel = cntSamplesPerPixel = atoi(argv[i+1]);
      assert(cntSamplesPerPixel > 0);
      assert(cntSamplesPerPixel < 10000000);
      i+=2; // parsed, the next param different
      cout << "Changing number of samples per pixel from " << k
	   << " to " << cntSamplesPerPixel << endl;
      continue;
    }
    if (strcmp(argv[i],"-cntref") == 0) {
      // setting number of samples to a value
      int k = defaultCntSamplesPerPixelREF;
      assert(i+1 < argc);
      defaultCntSamplesPerPixelREF = atoi(argv[i+1]);
      assert(defaultCntSamplesPerPixelREF > 0);
      assert(defaultCntSamplesPerPixelREF < 10000000);
      i+=2; // parsed, the next param different
      cout << "Changing number of samples per pixel for reference images from " << k
	   << " to " << defaultCntSamplesPerPixelREF << endl;
      continue;
    }
    if (strcmp(argv[i],"-alphamis") == 0) {
      // the alpha blending coefficient for MIS (=multiple importance sampling)
      // Usage: alpha=0.0 ... use only BRDF, alpha = 1.0 - use only EM sampling
      // setting blending coefficient for MIS
      float alphaOld = alphaMIS;
      assert(i+1 < argc);
      alphaMIS = atof(argv[i+1]);
      assert(alphaMIS >= 0.f); assert(alphaMIS <= 1.0f);
      i+=2; // parsed, the next param different
      cout << "Changing alpha for MIS from " << alphaOld
	   << " to " << alphaMIS << endl;
      continue;
    }
    if (strcmp(argv[i],"-saveGray") == 0) {
      // if to verbosely print estimate of time
      bool k = saveAlsoGrayToneImage;
      assert(i+1 < argc);
      int vv = atoi(argv[i+1]);
      saveAlsoGrayToneImage = (vv == 0) ? false: true;
      i+=2; // parsed, the next param different
      cout << "Saving also gray tone images = " << saveAlsoGrayToneImage << endl;
      continue;
    }
    if ((strcmp(argv[i],"-saveNormAlbedo") == 0)||
	(strcmp(argv[i],"-saveNormalizedAlbedo") == 0)) {
      // if to verbosely print estimate of time
      bool k = saveAlsoNormalizedAlbedo;
      assert(i+1 < argc);
      int vv = atoi(argv[i+1]);
      saveAlsoNormalizedAlbedo = (vv == 0) ? false: true;
      i+=2; // parsed, the next param different
      cout << "For Saving normalizing also albedo set to " << saveAlsoNormalizedAlbedo << endl;
      continue;
    }
    if (strcmp(argv[i],"-saveAndExit") == 0) {
      // if to verbosely print estimate of time
      bool k = saveAndExit;
      assert(i+1 < argc);
      int vv = atoi(argv[i+1]);
      saveAndExit = (vv == 0) ? false: true;
      i+=2; // parsed, the next param different
      cout << "Setting saveAndExit = " << saveAndExit << endl;
      continue;
    }
    if (strcmp(argv[i],"-method") == 0) {
      // rendering algorithm given currently
      int k = modeCompute;
      assert(i+1 < argc);
      modeCompute = atoi(argv[i+1]);
      assert(modeCompute >= 0);
      assert(modeCompute < 10000);
      i+=2; // parsed, the next param different
      cout << "Changing modeCompute from " << k
	   << " to " << modeCompute << " disabling processing of more methods" << endl;
      currModeComputeIndex = -1;
      if (modeCompute == 10) {
	readReferenceImageFlag = true;
      }
      continue;
    }
    if (strcmp(argv[i],"-atlas") == 0) {
      // if to verbosely print estimate of time
      bool k = saveImageBTFatlas;
      assert(i+1 < argc);
      int vv = atoi(argv[i+1]);
      saveImageBTFatlas = (vv == 0) ? false: true;
      i+=2; // parsed, the next param different
      cout << "Changing savingAtlas from " << k
	   << " to " << saveImageBTFatlas << endl;
      continue;
    }
    if (strcmp(argv[i],"-upperboundvis") == 0) {
      // The RSME Visualization image maximum
      float ff = upperBoundVisualization;
      assert(i+1 < argc);
      upperBoundVisualization = atof(argv[i+1]);
      assert(upperBoundVisualization >= 0.f); assert(pilotStageSamplesRatio <= 1000.f);
      i+=2; // parsed, the next param different
      cout << "Changing maximum value for visualization or RMSE error from " << ff
	   << " to " << upperBoundVisualization << endl;
      continue;
    }
    if (strcmp(argv[i],"-modenewsamples") == 0) {
      // This applies to computeMode6 and computeMode 7
      // either 1 .. compute minimum, 2 .. 50/50 percent, 3 .. take a single sampling 
      int k = modeNewSamples;
      assert(i+1 < argc);
      modeNewSamples = atoi(argv[i+1]);
      assert(modeNewSamples > 0);
      assert(modeNewSamples < 4);
      i+=2; // parsed, the next param different
      cout << "Changing modeNewMode for method 6 and 7 from " << k
	   << " to " << modeNewSamples << endl;
      continue;
    }

    if (strcmp(argv[i],"-pseudocolor") == 0) {
      // if to make some visualization how many samples are taken from EM and
      // how many from BRDF from all samples, etc.
      // setting it to 0 switches off the visualization pseudocolor
      int k = pseudoColorTechnique;
      assert(i+1 < argc);
      pseudoColorTechnique = atoi(argv[i+1]);
      assert(pseudoColorTechnique >= 0); // 0 is to do nothing
      assert(pseudoColorTechnique < 7);
      i+=2; // parsed, the next param different
      cout << "Changing pseudocolor mode for visualization from " << k
	   << " to " << pseudoColorTechnique << endl;
      continue;
    }
    if (strcmp(argv[i],"-o") == 0) {
      // where to save the images
      char oldpath[200]; strcpy(oldpath, FILEPREFIX);
      assert(i+1 < argc);
      strcpy(FILEPREFIX,argv[i+1]); // should be a path always
      if (FILEPREFIX[strlen(FILEPREFIX)-1] != '/') {
	strcat(FILEPREFIX,"/");
	cout << "Warning: adding / to the path, making new: " << FILEPREFIX << endl;
      }
      i+=2; // parsed, the next param different
      cout << "Images and data will be saved now to " << FILEPREFIX << endl;
      continue;
    }
    if (strcmp(argv[i],"-refimagefname") == 0) {
      // Read the reference image
      char oldfname[200]; strcpy(oldfname, referenceImageFilename);
      assert(i+1 < argc);
      strcpy(referenceImageFilename, argv[i+1]); // should be a path always
      readReferenceImageFlag = true;
      i+=2; // parsed, the next param different
      cout << "Changing refimage name from " << oldfname << " to " << referenceImageFilename << endl;
      continue;
    }
    if (strcmp(argv[i],"-diffimagefname") == 0) {
      // Difference image filename
      char oldfname[200]; strcpy(oldfname, diffImageFilename);
      assert(i+1 < argc);
      strcpy(diffImageFilename, argv[i+1]); // should be a path always
      readReferenceImageFlag = true;
      i+=2; // parsed, the next param different
      cout << "Changing diffimage name from " << oldfname << " to " << diffImageFilename << endl;
      continue;
    }
    if (strcmp(argv[i],"-diffvisimagefname") == 0) {
      // Difference image filename
      char oldfname[200]; strcpy(oldfname, diffvisImageFilename);
      assert(i+1 < argc);
      strcpy(diffvisImageFilename, argv[i+1]); // should be a path always
      readReferenceImageFlag = true; // we have to read the reference then
      i+=2; // parsed, the next param different
      cout << "Changing diffVISimage name from " << oldfname << " to " << diffvisImageFilename << endl;
      continue;
    }
    if (strcmp(argv[i],"-readrefimage") == 0) {
      assert(i+1 < argc);
      bool k = readReferenceImageFlag;
      int vv = atoi(argv[i+1]);
      readReferenceImageFlag = (vv == 0) ? false: true;
      i+=2; // parsed, the next param different
      cout << "We should read reference image " << readReferenceImageFlag << endl;
      continue;
    }

    if (strcmp(argv[i],"-vv") == 0) {
      // which variable will be visualized
      int k = visModeVariable;
      assert(i+1 < argc);
      visModeVariable = atoi(argv[i+1]);
      visualizationMode = 1; // set to make a visualization to a variable
      assert(visModeVariable >= 0);
      assert(visModeVariable < 100);
      i+=2; // parsed, the next param different
      cout << "Visualization of variable = " << visModeVariable << endl;
      continue;
    }
    if ((strcmp(argv[i],"phong") == 0)||
	(strcmp(argv[i],"phong3") == 0)
       ) {
      // command recognized;
      // this option was not recognized - go to the processing
      argc -= i+1; // the number of parameters to process
      return i; // we start with this parameter as the next, we do not change argv
    }
    else {
      cout << "ERROR: commandline option = " << argv[i] << " not recognized, exiting!!!" << endl;
      exit(0);
    }
  } // for i

  argc -= i+1; // the number of parameters to process
  return i; // we start with this parameter as the next, we do not change argv
}

// ----------------------------------------------------------------------------------
void
initOutput()
{
  glDisable(GL_CULL_FACE);
  glEnable(GL_DEPTH_TEST);
  // opengl setup for render Buffer
  glClearColor(backgroundColor[0], backgroundColor[1], backgroundColor[2], backgroundColor[3]);

  // Generate color floating point texture object to be attached
  glGenTextures(1, &fbotexture); // color buffer
  cout << "fbotexture = " << fbotexture << endl;
  glBindTexture(GL_TEXTURE_2D, fbotexture);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  //glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
  //glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
  //glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
  //glTexParameteri(GL_TEXTURE_2D, GL_GENERATE_MIPMAP, GL_TRUE);
  CHECK_GL_ERROR();
  //glTexParameteri(GL_TEXTURE_2D, GL_GENERATE_MIPMAP, GL_TRUE); // automatic mipmap
  // setting up texture as floating one
  glTexImage2D(GL_TEXTURE_2D,
	       0,
	       GL_RGBA32F,  // GL_RGBA32F, GL_RGBA
	       win_fbo_width,
	       win_fbo_height,
	       0,
	       GL_RGBA,
	       GL_FLOAT,
	       0);
  glBindTexture(GL_TEXTURE_2D, 0);
  CHECK_GL_ERROR();

  // ---------------------------------------------------------------------------
  // Create frame buffer object
  glGenFramebuffers(1, &fbo); // frame buffer
  glBindFramebuffer(GL_FRAMEBUFFER, fbo);
  CHECK_GL_ERROR();

  glGenRenderbuffers(1, &fbodepth); // depth buffer
  glBindRenderbuffer(GL_RENDERBUFFER, fbodepth);
  CHECK_GL_ERROR();
  glRenderbufferStorage(GL_RENDERBUFFER, GL_DEPTH_COMPONENT32, win_fbo_width, win_fbo_height);
  glBindRenderbuffer(GL_RENDERBUFFER, 0);
  CHECK_GL_ERROR();

  // attach a texture to FBO color attachement point
  glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, fbotexture, 0);
  // attach a renderbuffer to depth attachment point
  glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_RENDERBUFFER, fbodepth);
  CHECK_GL_ERROR();
  checkFramebufferStatus();
#if 0
  if (glCheckFramebufferStatus(fbo) != GL_FRAMEBUFFER_COMPLETE) {
    cout << "1 Wrong frame buffer status = " << glCheckFramebufferStatus(fbo) << endl;
    exit(0);
  }
#endif  
  CHECK_GL_ERROR();
  
  glBindRenderbuffer(GL_RENDERBUFFER, 0);
  glClearColor(backgroundColor[0], backgroundColor[1], backgroundColor[2], backgroundColor[3]);
  CHECK_GL_ERROR();
  return;
}

// This implements routine for comparison of the two images
void
compareTwoImages(int i, int argc, char **argv)
{
  float mult = 1.0f;
  float offset = 0.0f;
  const float BIGNUMBER = 1e20;
  float minval = BIGNUMBER;
  float maxval = -BIGNUMBER;
  bool fabsflag = false;
  for(;;) {
    if (strcmp(argv[i],"-mult") == 0) {
      assert(i+1 < argc);
      mult = atof(argv[i+1]);
      i+=2; // parsed, the next param different
      cout << "Changing multiplication from 1.0 to " << mult << endl;
      continue;
    }
    if (strcmp(argv[i],"-offset") == 0) {
      assert(i+1 < argc);
      offset = atof(argv[i+1]);
      i+=2; // parsed, the next param different
      cout << "Changing offset from 1.0 to " << offset << endl;
      continue;
    }
    if (strcmp(argv[i],"-min") == 0) {
      assert(i+1 < argc); float t = minval;
      minval = atof(argv[i+1]);
      i+=2; // parsed, the next param different
      cout << "Changing minvalue from " << t << " to " << minval << endl;
      continue;
    }
    if (strcmp(argv[i],"-max") == 0) {
      assert(i+1 < argc); float t = minval;
      maxval = atof(argv[i+1]);
      i+=2; // parsed, the next param different
      cout << "Changing maxvalue from " << t << " to " << maxval << endl;
      continue;
    }
    if (strcmp(argv[i],"-fabs") == 0) {
      assert(i+1 < argc);
      fabsflag = true;
      i+=1; // parsed, the next param different
      cout << "Making fabs at the end running "<< endl;
      continue;
    }
    break;
  }
  if ((i + 2) > argc) {
    cout << "Awaiting two images for difference" << endl;
  }
  
  CHDRImage *img1 = new CHDRImage();
  if (!img1->loadHDR(argv[i])) {
    cout << "The image from file " << argv[i] << " cannot be read - exiting" << endl;
    exit(1);
  }
  CHDRImage *img2 = new CHDRImage();
  if (!img2->loadHDR(argv[i+1])) {
    cout << "The image from file " << argv[i+1] << " cannot be read - exiting" << endl;
    exit(1);
  }
  
  // Reference HDR atlas is if file 'referenceImage'
  img1->CompareTo(*img2, true); // this prints RMSE only
  CHDRImage * diffImage = img1->Substruct(*img2, 0, false, false); // this creates difference image
  if (mult != 1.0f) {
    cout << "OPERATION Multiplying diffimage by " << mult << endl;
    diffImage->Multiply(mult);
  }
  if (offset != 0.0f) {
    cout << "OPERATION Adding offset to diffimage " << offset << endl;
    diffImage->Multiply(mult);
  }
  if (minval != BIGNUMBER) {
    cout << "OPERATION Computing Min(" << minval << ", image)" << endl;
    diffImage->Minimize(minval);
  }
  if (maxval != -BIGNUMBER) {
    cout << "OPERATION Computing Max(" << maxval << ", image)" << endl;
    diffImage->Maximize(maxval);
  }
  if (fabsflag) {
    cout << "OPERATION Computing Image=fabs(image)" << endl;
    diffImage->Fabs();
  }
      
  char diffnamefile[200];
  strcpy(diffnamefile, "diffimage2.hdr");
  if (i+2 < argc) {
    strcpy(diffnamefile, argv[i+2]);
  }
  char diffvisnamefile[200];
  strcpy(diffvisnamefile, "diffimagevis2.hdr");
  if (i+3 < argc) {
    strcpy(diffvisnamefile, argv[i+3]);
  }  
  diffImage->saveHDR(diffnamefile);
  // Visualization
  CHDRImage *diffvisImage = new CHDRImage(*diffImage, pseudoColorTechnique, true);; // it is populated below
  // and this populates it with color difference
  //cout << "1 Image diffimagevis size = " << diffvisImage->getWidth() << " " << diffvisImage->getHeight() << endl;
  //diffImage->setsPseudocolor(*diffvisImage, pseudoColorTechnique, 0, upperBoundVisualization);
  diffvisImage->saveHDR(diffvisnamefile); // saving to a file
  
  //cout << "Image difffimage size = " << diffImage->getWidth() << " " << diffImage->getHeight() << endl;
  //cout << "2 Image diffimagevis size = " << diffvisImage->getWidth() << " " << diffvisImage->getHeight() << endl;
  delete diffImage;
  delete diffvisImage;
}

void
TEST001()
{
  SVAR sv;
  const int CNT = 1000;
  double array[CNT];
  long double sum = 0;

  for (int i = 0; i < CNT; i++) {
    double X = (double)(i+10);
    array[i] = X; // store
    sv.update((float)X);
    sum += X;
  }
  double mean = sum/double(CNT);
  double sum2 = 0;
  for (int i = 0; i < CNT; i++) {
    sum2 += (array[i] - mean)*(array[i] - mean);
  }
  double variance = (double)sum2/((double)CNT-1);
  cout << "Knuth's result variance = " << sv.evaluate() << " mean = " << sv.mean << endl;
  cout << "Standard way eq. 71 and eq.75 = " << variance << " mean = " << mean << endl;
}

// -------------------------------------------------------------------------------------------------
// The main function
int 
main(int argc, char **argv) 
{
  SetRandomSeed(1); // reinitializing Mersenne Twister to some seed  
  //TEST001(); return 0;
  //  TestBatches(); exit(0);
  // Take arguments from argv that belong to glut
  
  if (argc < 2) {
    printHelp();
    printUsage(); // without parameters
    exit(0);
  }
  if (strcmp(argv[1], "h") == 0) {
    printHelp();
    printUsage();
    exit(0);
  }
  if (strcmp(argv[1], "h1") == 0) {
    printHelp();
    exit(0);
  }
  if (strcmp(argv[1], "h2") == 0) {
    printUsage();
    exit(0);
  }
  cout << "#Compiled= " << __DATE__ << " " __TIME__ << "\n";
  time_t t = time(NULL);
  struct tm * ttt = localtime(&t);
  cout << "#DATE= " << asctime(ttt);
  cout << "#Running program with arguments: ";
  for (int i=0; i < argc; i++) {
    cout << argv[i] << " ";
  }
  cout << endl << flush;
  // possibly decrease argc so removing processed switches
  int offsetargc = ProcessSwitches(argc, 1, argv);

  // separate utility
  if (whatToDo == 4) {
    compareTwoImages(offsetargc, argc, argv);
    return 0;
  }
  
  SetEnvironmentMapsNames();
  Set3DObjectsNames(true);
  if (argc < 0) {
    cout << "Exiting: no data to process, you must specify BRDF model" << endl;
    exit(0);
  }

  // now decide which BRDF model to take
  if (strcmp(argv[offsetargc], "phong") == 0) {
    printf("Running Phong model ");
    modeBRDF = 1;
    if (argc != 3) {
      printf("Number of parameters = %d", argc);
      printf("For Phong model you have to provide 3 parameters: rho_diff,"
	     " rho_spec, spec_index\n");
      exit(0);
    }
    float rho_diffuse, rho_specular, spec_index;
    rho_diffuse = atof(argv[offsetargc+1]);
    rho_specular = atof(argv[offsetargc+2]);
    spec_index = atof(argv[offsetargc+3]);
    printf("rhod=%3.2f rhos=%3.2f specindex=%3.3f\n", rho_diffuse,
	   rho_specular, spec_index);
    brdf = new BRDF_Phong(rho_diffuse, rho_specular, spec_index);
    offsetargc += 4;
  }
  else
  if (strcmp(argv[offsetargc], "phong3") == 0) {
    printf("Running Phong3 model ");
    modeBRDF = 1;
    if (argc != 7) {
      printf("Number of parameters = %d", argc);
      printf("For Phong3 model you have to provide 7 parameters: rho_diffR, rho_diffG, rhoDiffB"
	     " rho_specR, rho_specG, rho_specB, spec_index\n");
      exit(0);
    }
    float rho_diffuse[3], rho_specular[3], spec_index;
    rho_diffuse[0] = atof(argv[offsetargc+1]);
    rho_diffuse[1] = atof(argv[offsetargc+2]);
    rho_diffuse[2] = atof(argv[offsetargc+3]);
    rho_specular[0] = atof(argv[offsetargc+4]);
    rho_specular[1] = atof(argv[offsetargc+5]);
    rho_specular[2] = atof(argv[offsetargc+6]);
    spec_index = atof(argv[offsetargc+7]);
    printf("rhod=%3.2f,%3.2f,%3.2f rhos=%3.2f,%3.2f,%3.2f specindex=%3.3f\n",
	   rho_diffuse[0], rho_diffuse[1], rho_diffuse[2],
	   rho_specular[0], rho_specular[1], rho_specular[2], spec_index);
    brdf = new BRDF_Phong3(rho_diffuse, rho_specular, spec_index);
    offsetargc += 7;
  }
  int offsetargc2 = ProcessSwitches(argc, offsetargc, argv);

  if (saveAlsoNormalizedAlbedo) {
    float meanAlbedo, meanVar, stats;
    int cntEvals; // how many evaluations of albedo were used
    // compute total albedo for further normalization
    // mean albedo, meanVar0, cntEvals used.
    computeBRDFstats(brdf, meanAlbedo, meanVar, stats, cntEvals);
    assert(meanAlbedo > 0.f);
    assert(meanAlbedo <= 1.01f);
    // this is the multiplicator to make image brighter if albedo is < 1.0
    if (saveAlsoNormalizedAlbedo)
      normalizeTotalAlbedoCoeff = 1.0f/meanAlbedo;
  } // (saveAlsoNormalizedAlbedo)
  
//#define TEST_BRDF
#ifdef TEST_BRDF
  assert(brdf);
  Test_BRDF_Random(brdf);
  exit(0);
#endif
  
#ifdef RENDERING_MIRIKINA
  SetupForMirikima(argc, argv)
#endif // MIRIKIMA

  if (argc==2)
    printInitParameters();
  if (whatToDo == 1) {
    init(0);
    computeBRDFstats(brdf);
    return 0;
  }
  if (whatToDo == 2) {
    init(0);
    computeEMstats(actualEnvMap);
    return 0;
  }
  if (whatToDo == 3) {
    init(0);
    computeEMHALFstats(actualEnvMap);
    return 0;
  }

  if (whatToDo == 0) {
    //=== glut functions assignment  ===============================
    //=== initialisation ===============================
    cout << "=====================" << endl << flush;
    glutInit(&argc, argv);
    
    glutInitContextVersion(pgr::OGL_VER_MAJOR, pgr::OGL_VER_MINOR);
    glutInitContextFlags(GLUT_DEBUG | GLUT_FORWARD_COMPATIBLE);
    glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH);
    glutInitWindowPosition(0, 0);
    glutInitWindowSize(win_fbo_width, win_fbo_height);
    glutCreateWindow("Renderer Havran 2013");
    if(!pgr::initialize(pgr::OGL_VER_MAJOR, pgr::OGL_VER_MINOR))
      pgr::dieWithError("pgr init failed, required OpenGL not supported?");
    glViewport(0, 0, win_fbo_width, win_fbo_height);
    initProjectionMatrix(win_fbo_width, win_fbo_height);
    CHECK_GL_ERROR();
    init();

    //meshenvmap.loadTexture("ENV/stpeters.hdr");
    initOutput();
    glutReshapeFunc(changeSize);
    glutKeyboardFunc(processKeys);
    glutMouseFunc(processMouse);
    glutMotionFunc(motion);

    glClampColor(GL_CLAMP_READ_COLOR, GL_FALSE);
    initViewMatrices(win_fbo_width, win_fbo_height);  

    glutDisplayFunc(renderScene);
    glutIdleFunc(renderScene);

    //=== RENDERING ================================================
    glutMainLoop();
    cout << "END" << endl;
  }

  //=== OpenGL exit ================================================
  //cout << "AA" << endl;
  glutDestroyWindow(windowID);
  //cout << "BB" << endl;
  glutExit();
  cout << "CC" << endl;
  deallocations();
  
  return 0;
}//--- main ---------------------------------------------------------
