// ================================================================
// $Id: $
//
// render4.cpp
//
// This is different implementation of MIS, combining EM and BRDF*cos(theta) sampling, with storing the samples
// MIS = Multiple Importance Sampling (Veach and Guibas 1995)
//
// Initial coding by Vlastimil Havran, 2013.

#include <cassert>
#include <cmath>

#include "configg.h"
#include "AuxFuncs.h"
#include "ibtfbase.h"
#include "lightenv.h"
#include "brdfmodel.h"
#include "vrmlsup.h"
#include "tangles.h"
#include "render.h"

static SEntry *rbrdf = 0; // the records about sampling from BRDF
static SEntry *rem = 0; // the records about sampling from EM
static int raCnt = 0; // number of allocated samples for both arrays

void
computeMISpart(SEntry *ra, int cntSamplesFirst, int cntSamplesSecond,
	       int offset, int cntSamplesForSum, double resRGB[3])
{
  assert(cntSamplesFirst >= 0);
  assert(cntSamplesSecond >= 0);

  // the sampling according to sampling technique
  const int technique = ra[offset].technique;
  for (int i = 0; i < cntSamplesForSum; i++) {
    assert(ra[i+offset].technique == technique); // consistent technique during summing
    double weight = evaluateMISweight(ra[offset+i].probThis, cntSamplesFirst,
				      ra[offset+i].probOther, cntSamplesSecond);
    for (int isp = 0; isp<3; isp++)
      resRGB[isp] += weight * (double)ra[offset+i].RGB[isp];
  } // for i
  // finally, divided by the number of samples taken, when all the samples
  // are summed up, which is more numerically robust than doing it directly above

  // we are summing the only valid samples, but are dividing by all
  // the number of samples taken
  float mult = 1.0f/(float)cntSamplesFirst; 
  for (int isp = 0; isp<3; isp++)
    resRGB[isp] *= mult;
  return;
}

// using precomputed array of samples stored and reweighted at the end
void         // interpolated vertices, tangents, normals and binormals
computeMode5(float Vint[3], float Tint[3], float Nint[3], float Bint[3], 
	     float t_v, float p_v, int irow, int jcol,
	     int BTFindex, float normF)
{
  // use the importance sampling to get the value
  resetCountersST();
  const bool fixNumberOfSamples = false;// either fix it or not

  if ((rbrdf == 0)||(rem == 0) || (cntSamplesPerPixel > raCnt)) {
    delete []rbrdf; delete[]rem;
    rbrdf = new SEntry[cntSamplesPerPixel];
    rem = new SEntry[cntSamplesPerPixel];
    assert(rbrdf); assert(rem);
    raCnt = cntSamplesPerPixel;
  }
  
  int cntSamplesPerPixelBRDF = (1.0f-alphaMIS) * (float)cntSamplesPerPixel;
  int cntSamplesPerPixelEM = cntSamplesPerPixel - cntSamplesPerPixelBRDF;
  // the computed alpha can be different due to the discretization
  float alphaMISreal = (float)cntSamplesPerPixelBRDF/(float)cntSamplesPerPixel;

  // compute the samples from BRDF
  int cntValidSamplesBRDF = 0; // the number of samples really taken from BRDF
  int detectFailsBRDF = 0;
  int cntAllSamplesBRDF = 0;
  for(int l=0; l < cntSamplesPerPixelBRDF; l++) {
    // now we compute a ray generated by BRDF model according to BRDF*cos(theta)
    // or some other probability density function that is evaluated to 'probBRDF'
    float rnd1 = RandomValue(); // generate two random numbers
    float rnd2 = RandomValue();
    cntAllSamplesBRDF++;
    // compute the stochastic sampling for the two random values from BRDF*cos(theta)
    float probBRDF; // probability of sampling at the direction XYZL
    float brdfRGB[3]; // color=energy of environment map at the direction
    float XYZL[3]; // the direction on the surface of an object
    if (brdf->spawnStochasticRay(irow, jcol, rnd1, rnd2, // input
				 t_v, p_v, XYZL, brdfRGB, probBRDF)) { // output
      detectFailsBRDF++; // the number of failures
      if (fixNumberOfSamples) {
	if (detectFailsBRDF > cntSamplesPerPixelBRDF*4) {
	  cout << "Error, we cannot sample BRDF with really many samples = " << detectFailsBRDF << endl;
	  BTFimageF[BTFindex] = 0;
	  BTFimageF[BTFindex+1] = MARK_PROBLEM; // make it really GREEN
	  BTFimageF[BTFindex+2] = 0;
	  cout << "GREEN" << flush;
	  return;
	}
	l--; // we have to really taken specified number of samples
      }
      // so we repeat the sampling
      continue; // the ray for rnd1,rnd2 does not exist for the sampling
    }
#if 0
    assert(!isnan_local(brdfRGB[0]));
    assert(!isnan_local(brdfRGB[1]));
    assert(!isnan_local(brdfRGB[2]));
    //    assert(probBRDF > 0.f);
    if (probBRDF <= 0.f) {
      if (brdf->spawnStochasticRay(irow, jcol, rnd1, rnd2, // input
				   t_v, p_v, XYZL, brdfRGB, probBRDF)) { // output
	cout << "prob = " << probBRDF << endl;
	abort();
      }
    }
#endif    
    // The result ray in XYZL[3] is in the coordinate system (1,0,0),(0,1,0),(0,0,1)
    // at vertex V, given by vector T,B,N. We convert it to the world-space coordinate
    // system by using T,B,N using standard multiplication of vector by matrix
    float XYZ[3];
    // Normal  0, 0, 1 is converted to Nint
    // Tangent 1, 0, 0 is converted to Tint
    // Bint    0, 1, 0 is converted to Bint
    //
    // This is transpose matrix from local coordinate system to world-space system
    // it is possibly only since vectors T,B,N create orthogonal matrix, then inverse
    // matrix is equal to the transposed matrix
    XYZ[0] = Tint[0]*XYZL[0] + Bint[0]*XYZL[1] + Nint[0]*XYZL[2];
    XYZ[1] = Tint[1]*XYZL[0] + Bint[1]*XYZL[1] + Nint[1]*XYZL[2];
    XYZ[2] = Tint[2]*XYZL[0] + Bint[2]*XYZL[1] + Nint[2]*XYZL[2];
    // Set for the conversion to the global coordinate system
    float t_i, p_i;
    ConvertXYZtoThetaPhi(XYZ, t_i, p_i);
    // the light colour from environment map
    float envRGB[3]; // and the probability of sampling
    float probEM = environMap[actualEnvMap].getRGBandProb_ThetaPhi(t_i, p_i, envRGB);    
#if 0
    if (probEM > 1.0f) {
      cout << "PEM = " << probEM << " " << flush;
    }
#endif	
    // we store the result here
    rbrdf[cntValidSamplesBRDF].setZero();
    for(int isp=0; isp<=2; isp++) {// for all 3 color channels
      // we multiply BRDF.cos(theta) from BRDF sampling and RGB colour from EM
      // and divide it by probability from BRDF sampling
      // and store the partial result to the array 'rbrdf'
      rbrdf[cntValidSamplesBRDF].RGB[isp] = brdfRGB[isp] * envRGB[isp] / probBRDF; 
    }
    rbrdf[cntValidSamplesBRDF].probThis = probBRDF; // probability of sampling from BRDF
    rbrdf[cntValidSamplesBRDF].probOther = probEM; // probability of sampling from EM
    rbrdf[cntValidSamplesBRDF].technique = 0; // sampling according to BRDF.cos(theta)
    // the number of samples taken to compute the integrand
    cntValidSamplesBRDF++;
    if (cntValidSamplesBRDF > cntSamplesPerPixel) {
      cout << "ERROR A" << endl;
      abort();
    }
  } // for all samples according to BRDF
  // statistics update of global counters
  failBRDFcnt +=detectFailsBRDF;
  evalBRDFcnt +=cntAllSamplesBRDF;

  // BRDF: independent of setting variable 'fixNumberOfSamples'
  // the number of failures (samples below the plane) is in 'detectFailsBRDF'
  // the number of valid samples (samples above the plane given by normal) is in 'cntValidSamplesBRDF'
  // the number of all samples taken from BRDF is in 'cntAllSamplesBRDF'
  if (fixNumberOfSamples) {
    if (cntValidSamplesBRDF != cntSamplesPerPixelBRDF) {
      cout << "Error with sampling, BRDF samples not taken correctly" << endl;
      abort();
    }
  }
  // we now that overhead of sampling was 'detectFailsBRDF' variable

  // -----------------------------------------------------------------------------
  // Sampling according to EM
  int cntValidSamplesEM = 0; // the number of samples really taken from BRDF
  int cntAllSamplesEM = 0;
  int detectFailsEM = 0;
  int samplesMax = cntSamplesPerPixelEM;
  for(int l = 0; l < samplesMax; l++) {
    // copy the positions of the directional light for environment map here
    // this direction of light was precomputed.

    // possibly randomize the access to light
    environMap[actualEnvMap].generateNewRandomLightPrecomputed();
    TA->setIlluDir(environMap[actualEnvMap].getDirectionPrecomputedLight(l)); 
    float t_i, p_i;  // direction (theta, phi) from the environment map
    TA->computeIlluAngles(Vint, Tint, Bint, Nint, &t_i, &p_i);
    cntAllSamplesEM++; // we have to really take specified number of samples
    if (cntAllSamplesEM == maxCntSamplesPerPixel) {
      cout << "Error, the number of precomputed lights were exhausted without"
	   << " finding the required number of samples = " << cntSamplesPerPixelEM << endl;
      BTFimageF[BTFindex+0] = MARK_PROBLEM; // make pixel RED
      BTFimageF[BTFindex+1] = 0;
      BTFimageF[BTFindex+2] = 0;
      cout << "RED" << flush;
      return; // and return result
    } 
    // theta and phi are in radians
    if (t_i >= M_PI/2.f) {
      detectFailsEM++; // the number of failures for EM sampling
      if (fixNumberOfSamples)
	samplesMax++;
      // nothing to do with the illumination from this light direction, continue sampling
      continue;
    }

    // the value given by BRDF model
    float brdfRGB[3];
    // the cos(theta) factor is already included in the call below
    // including the probability of sampling from BRDF*cos(theta)
    bool tt = brdf->getBRDFValue(irow, jcol, t_v, p_v, t_i, p_i, brdfRGB);
    // compute probability of sampling by current algorithm, particular function used for sampling here
    // not sure if to use this or this below
    float probBRDF = brdf->getProbOfSampling(irow, jcol, t_v, p_v, t_i, p_i, brdfRGB);
#if 0
    if (probBRDF > 1.f) {
      cout << "PBRDF = " << probBRDF << " " << flush;
    }
#endif
        
    // this value has been precomputed with this probability
    float probEM = environMap[actualEnvMap].getProbPrecomputedLight(l);
    assert(probEM > 0.f);
    if (probEM > 1.f) {
      //  cout << "UNLIK EM = " << probEM << endl;
    }
#if 0
    if (isnan_local(brdfRGB[2])) {      
      cout << "UUU " << brdfRGB[2] << endl;
      abort();
    }
    if (isnan_local(environMap[actualEnvMap].getColorPrecomputedLight(l, 0))) {      
      cout << "UUU " << environMap[actualEnvMap].getColorPrecomputedLight(l, 0) << endl;
      abort();
    }
#endif    

    // probability of sampling to some directions can be zero, since if we have
    // only specular function or some parts with zero values, this will exactly happen
    // in physical reality impossible, in mathemathical model easily possible
    if (probBRDF <= 0.f) {
      probBRDF = luminance(brdfRGB) * 1e10f+1e-20; // so the result will be zero as value/prob is the result
    }
    assert(probBRDF > 0.f);
    // we store the result here
    rem[cntValidSamplesEM].setZero();
    for(int isp=0;isp<=2;isp++) {// for all 3 color channels
      rem[cntValidSamplesEM].RGB[isp] =
	// the colour of light from EM was precomputed - multiply BRDF*cos(theta) * lightEM/probEM
	brdfRGB[isp] * environMap[actualEnvMap].getColorPrecomputedLight(l, isp)/probEM;
    } // for all channels
    rem[cntValidSamplesEM].probThis = probEM; // probability of sampling from EM
    rem[cntValidSamplesEM].probOther = probBRDF; // probability of sampling from BRDF
    rem[cntValidSamplesEM].technique = 1; // sampling according to EM is marked
    cntValidSamplesEM++; // the number of samples really taken from EM 
    if (cntValidSamplesEM > cntSamplesPerPixel) {
      cout << "ERROR B " << endl;
      abort();
    }   
  } // for all lights
  evalEMcnt +=cntAllSamplesEM; // global counter
  failEMcnt +=detectFailsEM; // global counter

  if (fixNumberOfSamples) {
    if (cntValidSamplesEM != cntSamplesPerPixelEM) {
      cout << "Error with sampling, EM samples not taken correctly." << endl;
      abort();
    }
  }

  // BRDF: independent of setting variable 'fixNumberOfSamples'
  // the number of failures (samples below the plane) is in 'detectFailsBRDF'
  // the number of valid samples (samples above the plane given by normal) is in 'cntValidSamplesBRDF'
  // the number of all samples taken from BRDF is in 'cntAllSamplesBRDF'

  // EM: independent of setting variable 'fixNumberOfSamples'
  // the number of failures (samples below the plane) is in 'detectFailsEM'
  // the number of valid samples (in direction of surface normal) is in 'cntValidSamplesEM'
  // the number of all samples taken from EM is in 'cntAllSamplesEM'
  // It does not need to hold that cntSamplesEM == cntSamplesPerPixelEM
  
  // Note that in general cntSamplesPerPixel != cntSamplesPerPixelBRDF + cntSamplesPerPixelEM

  // ----------------------------------------------------------------------------------
  // now let us compute the results for two different sampling strategies

  // First evaluate the BRDF sampling
  double resBRDF[3];
  resBRDF[0]=resBRDF[1]=resBRDF[2]=0;
  if (cntValidSamplesBRDF) {
    assert(rbrdf[0].technique == 0);
    computeMISpart(rbrdf, cntAllSamplesBRDF, cntAllSamplesEM, // for weighting
		   0, cntValidSamplesBRDF, resBRDF); // how many samples to sum up
#if 0
    float RGBs[3]; RGBs[0]=RGBs[1]=RGBs[2] = 0.f;
    for (int i = 0; i < cntValidSamplesBRDF; i++) {
      for (int isp = 0; isp < 3; isp++) {
	if (isnan_local(rbrdf[i].RGB[2])) {
	  cout << "AAA " << endl << flush;
	  abort();
	}
	RGBs[isp] += rbrdf[i].RGB[isp];
      }
    }
    cout << "BRDF = " << RGBs[0] << " " << RGBs[1] << " " << RGBs[2] << endl;
#endif
  }

  //  cout << "=============" << endl;
  
  // Second evaluate the EM sampling
  double resEM[3];
  resEM[0]=resEM[1]=resEM[2]=0;
  if (cntValidSamplesEM) {
    assert(rem[0].technique == 1);
    computeMISpart(rem, cntAllSamplesEM, cntAllSamplesBRDF,
		   0, cntValidSamplesEM, resEM); // how many samples to sum up
#if 0
    float RGBs[3]; RGBs[0]=RGBs[1]=RGBs[2] = 0.f;
    for (int i = 0; i < cntValidSamplesEM; i++) {
      for (int isp = 0; isp < 3; isp++) {
	if (isnan_local(rem[i].RGB[2])) {
	  cout << "BBB " << endl << flush;
	  abort();
	}
	RGBs[isp] += rem[i].RGB[isp];
      }	
    }
    cout << "EM = " << RGBs[0] << " " << RGBs[1] << " " << RGBs[2] << endl;
    cout << "validSamples = " << cntValidSamplesEM << endl;
    cout << "allSamples = " << cntAllSamplesEM << endl;
#endif    
  }

  // now sum together two contributions evaluated
  float result[3];
  for (int isp = 0; isp<3; isp++)
    result[isp] = resBRDF[isp] + resEM[isp];

  // and multiply it by given constant and copy result to required output array
  for(int isp=0;isp<=2;isp++) {
    // here we copy the result to the final destination
    BTFimageF[BTFindex+isp] = normF * result[isp];
  }
  
  setCountersST(cntAllSamplesBRDF, cntAllSamplesEM);
  return;
}

